/*
 * $Id: XmlDocument.java,v 1.5 2002/11/29 12:44:32 edankert Exp $
 *
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License.
 *
 * The Original Code is eXchaNGeR browser code. (org.xngr.browser.*)
 *
 * The Initial Developer of the Original Code is Cladonia Ltd. Portions created 
 * by the Initial Developer are Copyright (C) 2002 the Initial Developer. 
 * All Rights Reserved. 
 *
 * Contributor(s): Edwin Dankert <edankert@cladonia.com>
 */
package org.xngr.browser.editor;

import javax.swing.text.GapContent;
import javax.swing.text.PlainDocument;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Element;

/**
 * The XML Document is responsible for handling the user insertions and 
 * deletions, for changing the tab characters to spaces and to automatically 
 * indent the text correctly.
 *
 * <p>
 * <b>Note:</b> The XML Editor package is based on the JavaEditorKit example as 
 * described in the article <i>'Customizing a Text Editor'</i> by <b>Timothy Prinzing</b>.
 * See: http://java.sun.com/products/jfc/tsc/articles/text/editor_kit/
 * </p>
 *
 * @version	$Revision: 1.5 $, $Date: 2002/11/29 12:44:32 $
 * @author Edwin Dankert <edankert@cladonia.com>
 */
public class XmlDocument extends PlainDocument {
	private XmlEditorPane editor = null;
	
	private boolean loading = false;

	/**
	 * Constructs the XML Document with a new GapContent buffer.
	 */
    public XmlDocument( XmlEditorPane editor) {
		super( new GapContent( 1024));
		
		this.editor = editor;
    }
	
	/**
	 * Lets the document know that it is being loaded, don't attempt to 
	 * do any text conversion.
	 */
	public void setLoading( boolean loading) {
		this.loading = loading;
	}

    /**
     * Gets the location where the last significant tag ended, this
	 * location can be used as a start for the scanner. 
	 * If the current element is a comment, the last significant tag 
	 * is before the begin of the comment. 
	 *
	 * @param p the preferred start position.
	 *
	 * @return the position where the last significant tag ended or 0. 
     */
    public int getTagEnd( int p) {
		int elementEnd = 0;

		if ( p > 0) {
			try {
				int index = 0;
				
				String s = getText( 0, p);
				int commentStart = s.lastIndexOf( "<!--");
				int commentEnd = s.lastIndexOf( "-->");
				
				if ( commentStart > 0 && commentStart > commentEnd) {
					index = s.lastIndexOf( ">", commentStart);
				} else {
					index = s.lastIndexOf( ">");
				}

			    if ( index != -1) {
					elementEnd = index;
		    	}
			} catch ( BadLocationException bl) {}
		}

		return elementEnd;
    }

	/**
	 * Inserts some content into the document.
	 * When the content is a tab character, the character will be replaced 
	 * by spaces. When the content is a new line character, an indentation will
	 * be added to the content.
	 */ 
	public void insertString( int off, String str, AttributeSet set) throws BadLocationException {
//		System.out.println("[1]insertString( "+str+")");
		if ( !loading && str.equals( ">")) {
			if ( editor.isTagCompletion()) {
				int caretPosition = editor.getCaretPosition();
				StringBuffer endTag = new StringBuffer( str);

				String text = getText( 0, off);
				int startTag = text.lastIndexOf( '<', off);
				int prefEndTag = text.lastIndexOf( '>', off);

				// If there was a start tag and if the start tag is not empty and 
				// if the start-tag has not got an end-tag already.
				if ( (startTag > 0) && (startTag > prefEndTag) && (startTag < text.length()-1) ) {
					String tag = text.substring( startTag, text.length());
					char first = tag.charAt( 1);
					
					if ( first != '/' && first != '!' && first != '?' && !Character.isWhitespace( first)) {
						boolean finished = false;
						char previous = tag.charAt( tag.length() - 1);

						if ( previous != '/' && previous != '-') {

							endTag.append( "</");
							
							for ( int i = 1; (i < tag.length()) && !finished; i++) {
								char ch = tag.charAt( i);
								
								if ( !Character.isWhitespace( ch)) {
									endTag.append( ch);
								} else {
									finished = true;
								}
							}

							endTag.append( ">");
						}
					}
				}
				
				str = endTag.toString();
			
				super.insertString( off, str, set);

				editor.setCaretPosition( caretPosition+1);
			} else {
				super.insertString( off, str, set);
			}
		} else if ( !loading && str.equals( "\n")) {
			StringBuffer newStr = new StringBuffer( str);
			Element elem = getDefaultRootElement().getElement( getDefaultRootElement().getElementIndex( off));
			int start = elem.getStartOffset();
			int end = elem.getEndOffset();
		    String line = getText( start, off - start);
			
			boolean finished = false;
			
			for ( int i = 0; (i < line.length()) && !finished; i++) {
				char ch = line.charAt( i);
				
				if ( ((ch != '\n') && (ch != '\f') && (ch != '\r')) && Character.isWhitespace( ch)) {
					newStr.append( ch);
				} else {
					finished = true;
				}
			}
			
			if ( isStartElement( line)) {
				newStr.append( getTabString());
			}
			
			str = newStr.toString();

			super.insertString( off, str, set);

		} else {
			super.insertString( off, str, set);
		}
	}
	
	private String getTabString() {
		byte[] tab = new byte[editor.getSpaces()];
		
		for ( int i = 0; i < tab.length; i++) {
			tab[i] = ' ';
		}

		return new String( tab);
	}

	// Tries to find out if the line finishes with an element start
	private boolean isStartElement( String line) {
		boolean result = false;
		
		int first = line.lastIndexOf( "<");
		int last = line.lastIndexOf( ">");
		
		if ( last < first) { // In the Tag
			result = true;
		} else {
			int firstEnd = line.lastIndexOf( "</");
			int lastEnd = line.lastIndexOf( "/>");

			// Last Tag is not an End Tag
			if ( (firstEnd != first) && ((lastEnd + 1) != last)) {
				result = true;
			}
		}
		
		return result;
	}
}

