package sk.robotics.rlcs;

import sk.robotics.rlcs.agentspace.*;
import sk.robotics.rlcs.security.*;

import com.microstepmis.xplatform.X2O;
import com.microstepmis.xplatform.XException;

import java.io.*;
import java.util.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.security.*;
import java.lang.reflect.ReflectPermission;

/**
 *  RLCS appilcation entry point.
 *  This class extends Agent, just to reuse its logging methods.
 * <p>
 *  Command line parameters (all optional):<br>
 *  -c initial-scheduler-cfg<br>
 *  -p tcp-listening-port<br>
 *  -s security-cfg<br>
 *  -l preload-classes-cfg
 */
public class Main extends Agent
{
	private static final int DEFAULT_PORT = 35759;

	// the main thread
	private static Thread mainThread;

	// the system thread group
	private static ThreadGroup systemThreadGroup;

	/** Get the main thread. */
	public static Thread getMainThread() {
		return mainThread;
	}

	/** Get the system thread group. */
	public static ThreadGroup getSystemThreadGroup() {
		return systemThreadGroup;
	}

	/**
	 *  Program entry point.
	 */
	public static void main(String[] args) {
		// enable headless mode
		System.setProperty("java.awt.headless", "true");

		// set the main thread, and system thread group references
		mainThread = Thread.currentThread();
		systemThreadGroup = mainThread.getThreadGroup().getParent();

		// disable (for all agents) ImageIO file caching
		javax.imageio.ImageIO.setUseCache(false);

		// this needs special permissions first time it is called, and is needed for any
		// image manipulation, so call it here because agents won't have the permissions
		java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment();

		// run the app
		Main app = new Main();
		app.setLogLevel(1);
		try {
			app.run(args);
		}
		catch (Exception e) {
			app.log(0, e, "Exception in main thread, exiting.");
			System.exit(1);
		}
	}

	/**
	 *  Class for shutdown-hook threads.
	 *  All permissions are granted to threads of this class.
	 */
	public static class ShutdownHook extends Thread {
		// privte ctor - prevents abusing of this class
		private ShutdownHook(Runnable target) {
			super(target, "ShutdownHook");
		}
		// thread func - catches exceptions
		@Override public void run() {
			try { super.run(); }
			catch (Throwable e) {
				synchronized (System.err) {
					System.err.println("Exception in ShutdownHook:");
					e.printStackTrace(System.err);
				}
			}
		}
	}

	/**
	 *  Class for client-connection-handler threads.
	 *  Scheduler.restart() call is allowed for threads of this class,
	 *  and also a fixed set of permission (see below) is granted for them.
	 */
	public static class ClientHandler extends Thread {
		// private ctor - prevents abusing of this class
		private ClientHandler(Runnable target) {
			super(target, "ClientHandler");
		}

		// permissions granted to client handler threads
		public static final Permissions permissions = new Permissions();
		static {
			permissions.add(new PropertyPermission("*", "read"));
			permissions.add(new ReflectPermission("suppressAccessChecks")); // needed by XML mapping
			permissions.add(new RuntimePermission("accessDeclaredMembers"));
			permissions.setReadOnly();
		}
	}

	private void run(String[] args) throws IOException, XException {
		// parse command line paramaters
		final String paramTypes = "cspl"; // accepted command line parameters (one char per parameter)
		Map<Character, String> params = new HashMap<Character, String>();
		for (int i = 0; i < args.length; ++i) {
			if (i == args.length - 1) {
				log(0, "unpaired parameter: " + args[i]);
				break;
			}
			if (args[i].length() == 2 && args[i].charAt(0) == '-' && paramTypes.indexOf(args[i].charAt(1)) >= 0) {
				params.put(args[i].charAt(1), args[++i]);
			} else {
				log(0, "invalid argument: " + args[i]);
			}
		}

		// preload configured classes
		if (params.containsKey('l')) {
			// load list of classes to preload form xml config, and get Class object for all specified classes
			String[] preload = (String[]) X2O.mapFromXML(new File(params.get('l')));
			for (String className: preload) {
				try { Class.forName(className); }
				catch (ClassNotFoundException e) {
					throw new RuntimeException(e);
				}
			}
			log(0, "Preloaded " + preload.length + " classes");
		}

		// init security manager
		if (params.containsKey('s')) {
			// load permissions from XML, and setup security manager
			PermissionDesc[] perms = (PermissionDesc[]) X2O.mapFromXML(new File(params.get('s')));
			System.setSecurityManager(new sk.robotics.rlcs.security.SecurityManager(perms));
		}

		// load config from file, if one was specified
		Scheduler.Cfg initialConfig = null;
		if (params.containsKey('c')) {
			initialConfig = (Scheduler.Cfg) X2O.mapFromXML(new File(params.get('c')));
		}

		// init socket server
		Integer port = null;
		String customPort = params.get('p');
		if (customPort != null) {
			try {
				port = new Integer(customPort);
			} catch (NumberFormatException e) {
				log(0, "invalid TCP port: " + customPort);
			}
			if (port != null && (port < 0 || port > 65535)) {
				log(0, "invalid TCP port: " + port);
				port = null;
			}
		}
		log(0, "opening socket server channel");
		final ServerSocketChannel server = ServerSocketChannel.open();
		server.socket().setReuseAddress(true);
		SocketAddress bindAddr = new InetSocketAddress((port == null) ? DEFAULT_PORT : port);
		log(0, "binding to " + bindAddr);
		server.socket().bind(bindAddr);
		log(0, "listening on port " + server.socket().getLocalPort());

		// start scheduler
		// (and if initial configuration was specified, start configured agents)
		final Scheduler scheduler = Scheduler.getInstance();
		scheduler.start();
		if (initialConfig != null) {
			scheduler.restart(initialConfig);
		}

		// register shutdown hook, which will close the server socket, and quit all agents
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(new Runnable() {
			public void run() {
				log(1, "shutdown hook");
				mainThread.interrupt();
				synchronized (clientThreads) {
					for (Thread t: clientThreads) {
						t.interrupt();
					}
				}
				scheduler.exit();
				scheduler.quitAll();
			}
		}));

		// start handling connections
		Socket client = null;
		try {
			for (;;) {
				// at this point, the previous client socket (if any) is either already closed,
				// or owned by another thread, so reset the reference to avoid trying to close it
				// if accept throws an exception
				client = null;

				// wait for connection
				client = server.accept().socket();
				log(0, "accepted connection from " + client.getRemoteSocketAddress());

				// we need this to access the socket from inner classes
				final Socket fclient = client;

				// initialize input reader
				final BufferedReader reader = new BufferedReader(new InputStreamReader(client.getInputStream())) {
					// we don't want to close the socket, when the reader gets closed,
					// because we may still want to send something through the socket
					@Override public void close() {}
				};

				// read command/query (first line of input), but with a timeout
				String cmd;
				client.setSoTimeout(3000);
				try { cmd = reader.readLine(); }
				catch (SocketTimeoutException e) {
					log(0, "client timed out");
					client.close();
					continue;
				}
				client.setSoTimeout(0);

				// GETCFG command - send current scheduler config
				if (cmd.equals("GETCFG")) {
					log(0, "GETCFG command from client");
					startClientThread(new Runnable() { public void run() { sendConfig(fclient); } });
					continue;
				}

				// PUTCFG command - read new config and restart scheduler-managed agents
				if (cmd.equals("PUTCFG")) {
					log(0, "PUTCFG command from client");
					startClientThread(new Runnable() { public void run() { receiveConfig(fclient, reader); } });
					continue;
				}

				// READSPACE command - send value from space
				if (cmd.startsWith("READSPACE ")) {
					final String blockName = cmd.substring(10);
					log(0, "READSPACE command from client, block: " + blockName);
					startClientThread(new Runnable() { public void run() { sendValueFromSpace(fclient, blockName); } });
					continue;
				}

				// unknown command, or timeout
				log(0, "unknown command, dropping connection");
				client.close();
			}
		} catch (ClosedByInterruptException e) {
		} catch (InterruptedException e) {
		} finally {
			if (client != null) {
				try {
					client.close();
					log(0, "client socket closed");
				} catch (Exception ee) {
					log(0, ee, "failed to close client socket");
				}
			}
			try {
				server.close();
				log(0, "socket server closed");
			} catch (Exception ee) {
				log(0, ee, "failed to close socket server");
			}
		}
	}

	// set of client threads
	private Collection<Thread> clientThreads = Collections.synchronizedSet(new HashSet<Thread>());

	// safely starts a client threads, storing it into the client threads set
	private void startClientThread(Runnable target) throws InterruptedException {
		synchronized (clientThreads) {
			if (Thread.interrupted()) throw new InterruptedException();
			Thread thread = new ClientHandler(target);
			thread.start();
			clientThreads.add(thread);
		}
	}

	// send current configuration to a client
	private void sendConfig(Socket client) {
		try {
			String config = X2O.mapToXML(Scheduler.getInstance().getCurrentConfig());
			client.getOutputStream().write(config.getBytes());
			client.getOutputStream().flush();
		}
		catch (Exception e) {
			log(0, e, "failed to send configuration to client at " + client.getRemoteSocketAddress());
		}
		finally {
			try { client.close(); }
			catch (IOException e) {
				log(0, e, "failed to close client socket");
			}
		}
	}

	// receive new configuration from a client
	private void receiveConfig(Socket client, BufferedReader reader) {
		BufferedWriter writer = null; // writer to send reply through
		Scheduler.Cfg config; // received new configuration
		try {
			try {	// init the writer
				writer = new BufferedWriter(new OutputStreamWriter(client.getOutputStream()));
			} catch (Exception e) {
				throw new Exception("failed to initialize socket output", e);
			}

			try {	// receive the new config
				config = (Scheduler.Cfg) X2O.mapFromXML(reader);
			} catch (Exception e) {
				throw new Exception("failed to receive/parse the config", e);
			}

			try {	// restart RLCS
				Scheduler.getInstance().restart(config);
			} catch (Exception e) {
				throw new Exception("failed to restart RLCS", e);
			}

			try {	// send ACK
				writer.write("OK\n");
				writer.flush();
			} catch (Exception e) {
				throw new Exception("failed to send ACK", e);
			}
		}
		catch (Exception e) {
			log(0, e, "failed to receive and launch new configuration from client at " + client.getRemoteSocketAddress());

			// try to send NACK
			if (writer != null) {
				try {
					writer.write("ERR\n");
					writer.flush();
				} catch (IOException ee) {
					log(0, "failed to send NACK to client at " + client.getRemoteSocketAddress());
				}
			}
		}

		// close the socket
		try { client.close(); }
		catch (IOException e) {
			log(0, e, "failed to close client socket");
		}
	}

	// send a value from space to a client
	private void sendValueFromSpace(Socket client, String blockName) {
		try {
			try {
				client.getOutputStream().write(X2O.mapToXML(space.getBlock(blockName)).getBytes());
				client.getOutputStream().flush();
			} catch (XException e) {
				log(0, e, "failed to map value from space to XML");
				client.getOutputStream().write(("error: " + e).getBytes());
				client.getOutputStream().flush();
			}
		} catch (Exception e) {
			log(0, e, "failed to send value from space to client at " + client.getRemoteSocketAddress());
		} finally {
			try { client.close(); }
			catch (IOException e) {
				log(0, e, "failed to close client socket");
			}
		}
	}
}
