package sk.robotics.rlcs.agents.beh;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;
import sk.robotics.rlcs.data.Movement;
import static sk.robotics.rlcs.data.Movement.Direction;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.PositionQueue;
import sk.robotics.rlcs.data.Collision;
import sk.robotics.rlcs.constants.TurtleConstants;

import java.awt.geom.Point2D;

public class ImitateAgent extends Agent {
	/**
	 *  Agent which moves one robot to imitate movement of another robot.
	 */
	public static class Cfg extends Agent.Cfg {
		/** Imitation delay, in milliseconds. */
		public @Required long delayMs = 500;

		/** Name of block with position history of the imitated agent.
		 *  @block.type sk.robotics.rlcs.data.PositionQueue
		 *  @block.access read */
		public @Required String blockTargetHistory = "posHistory/robot?";

		/** Name of block with actual position of the imitating robot.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required String blockActualPos = "actualPos/robot?";

		/** Name of block where to store the desired movement command.
		 *  @block.type sk.robotics.rlcs.data.Movement
		 *  @block.access read/write */
		public @Required String blockMovement = "movement/robot?";

		// default activation
		public Cfg() {
			timerPeriodMs = 100;
		}
	}
	protected final Cfg cfg;

	private Movement lastMove = Movement.NO_MOVE;

	public ImitateAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		// read needed data from space
		Position pos = (Position) space.read(cfg.blockActualPos);
		PositionQueue hist = (PositionQueue) space.read(cfg.blockTargetHistory);
		if (pos == null || hist == null) {
			if (pos == null) log(5, "no position!");
			if (hist == null) log(5, "no history!");
			return; // something is wrong, do nothing
		}

		// get current position of the imitated robot, and older position
		long now = System.currentTimeMillis();
		Position p1 = hist.getPositionByTime(now - cfg.delayMs);
		Position p2 = hist.getPositionByTime(now);

		// determine move needed to reach position p2 from position p1 in given delay
		double dist = p1.distance(p2);
		double dir = -pos.headingDiff(p1.directionTo(p2));
		double hdif = p2.headingDiff(p1);
		Movement move = Movement.NO_MOVE;
		if (dist < 10 && Math.abs(hdif) > 0.1) {
			int speed = (int)((Math.abs(hdif) / cfg.delayMs) * 99.0 / TurtleConstants.SPEED_RAD_PER_MS);
			speed = Math.min(speed, 99);
			move = new Movement(hdif > 0 ? Direction.RT : Direction.LT, speed, speed);
		} else if (dist >= 10) {
			// FIXME: don't move straight here, but towards p2!
			int speed = (int)((dist / cfg.delayMs) * 99.0 / TurtleConstants.SPEED_MM_PER_MS);
			speed = Math.min(speed, 99);
			move = new Movement(Math.abs(dir) > Math.PI/2 ? Direction.BK : Direction.FD, speed, speed);
		}

		// write desired movement command to space
		space.write(cfg.blockMovement, move);

		// logging
		if (!move.equals(lastMove)) {
			log(2, "movement %s (was %s)", move, lastMove);
			lastMove = move;
		}
	}
}
