package sk.robotics.rlcs.agents.beh;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;

import java.awt.geom.Point2D;

/**
 *  Agent which moves a robot along a predefined path. Once the robot get close enough
 *  to its current destination, the destination point is changed to next one on the path.
 */
public class PathAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Tolerance for checking whether we are at destination point, in millimeters. */
		public Double destinationTolerance = 10.0;

		/** The predefined path. Points delimited by spaces, commas, semicolons, or combinations.
		 *  Points are number coordinates delimited by colons. Exapmle: 0:0 0:100 100:100 100:0. */
		public @Required String path;

		/** Priority of writing destination points into space. */
		public Float priority = 0.0f;

		/** Name of block with actual position of the robot.
		 *  Used to check whether the robot is already at its current destination place.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		/** Name of block with current destination point of the robot.
		 *  @block.type java.awt.geom.Point2D
		 *  @block.access read/write */
		public @Required @NoDefault String blockDestination = "destination/robot?";

		public Cfg() {
			// default activation
			timerPeriodMs = 1000;
		}

		/** Parses and validates the path, in addition to inherited validation. */
		@Override public void validate() {
			// inherited validation
			super.validate();

			// parse the path
			String[] points = path.split("[ ,;]+");
			if (points.length < 2) {
				throw new IllegalArgumentException("Path must consist of at least 2 points!");
			}
			pathPoints = new Point2D[points.length];
			for (int i = 0; i < points.length; ++i) {
				String[] coords = points[i].split(":");
				if (coords.length != 2) {
					throw new IllegalArgumentException("Path point must consist of two colon delimited numbers!");
				}
				pathPoints[i] = new Point2D.Double(Double.valueOf(coords[0]), Double.valueOf(coords[1]));
			}
		}
		private transient Point2D[] pathPoints;
	}
	private final Cfg cfg;

	// index of current destination point int the configured path
	private int currentPathIndex = 0;

	public PathAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		// update the destination in space
		Point2D dest = cfg.pathPoints[currentPathIndex];
		space.write(cfg.blockDestination, dest, 0, cfg.priority);

		// check whether we are already at the destination point already
		Position pos = (Position)space.read(cfg.blockActualPos);
		if (pos == null || pos.distance(dest) > cfg.destinationTolerance) {
			return; // we are not yet at current destinaion point
		}

		// change destination to next path point
		currentPathIndex = (currentPathIndex + 1) % cfg.pathPoints.length;
		dest = cfg.pathPoints[currentPathIndex];
		space.write(cfg.blockDestination, dest, 0, cfg.priority);

		log(1, "new destination [%s, %s] (path point %s)", dest.getX(), dest.getY(), currentPathIndex+1);
	}
}
