package sk.robotics.rlcs.agents.beh;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.constants.TurtleConstants;
import sk.robotics.rlcs.constants.VlabConstants;

import java.awt.geom.Point2D;

/**
 *  Roaming agent, moves the robot randomly. Once the robot get close enough to its
 *  current destination point, this agent generates randomly a new destination point.
 */
public class RoamAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Tolerance for checking whether we are at destination point, in millimeters. */
		public Double destinationTolerance = 10.0;

		/** Restricted zone around around walls, in millimeters.
		 *  New random destination points will never be inside this restricted zone.
		 *  This is without robot radius, which is also taken into account
		 *  when generating the destination point. */
		public Double wallBorderMm = 10.0;

		/** Name of block with actual position of the robot.
		 *  Used to check whether the robot is already at its current destination place.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		/** Name of block with current destination point of the robot.
		 *  @block.type java.awt.geom.Point2D
		 *  @block.access read/write */
		public @Required @NoDefault String blockDestination = "destination/robot?";

		public Cfg() {
			// default activation
			timerPeriodMs = 1000;
		}
	}
	private final Cfg cfg;

	public RoamAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}
	
	// latest next destination change
	private long latestNextChange = 0;

	@Override protected void senseSelectAct() {
		// check whether we are already at current destination point
		Position pos = (Position)space.read(cfg.blockActualPos);
		if (pos == null) return; // no position -> no destination
		Point2D dest = (Point2D)space.read(cfg.blockDestination);
		if (dest != null && pos.distance(dest) > cfg.destinationTolerance && System.currentTimeMillis() < latestNextChange)
			return; // we are not yet at current destinaion point, and not yet timed out

		// choose a new destination point randomly
		double B = TurtleConstants.RADIUS_MM + cfg.wallBorderMm;
		Point2D destination = new Point2D.Double(
			B + Math.random() * (VlabConstants.WIDTH_MM - 2*B),
			B + Math.random() * (VlabConstants.HEIGHT_MM - 2*B));

		// write the new destination into space
		space.write(cfg.blockDestination, destination);
		log(1, "new destination [%s, %s]", destination.getX(), destination.getY());

		// compute timeout from distance of the new destination
		latestNextChange = System.currentTimeMillis() + (long)(pos.distance(destination) * 30);
	}
}
