package sk.robotics.rlcs.agents.camera;

import sk.robotics.rlcs.agents.image.TransformAgent;
import sk.robotics.rlcs.data.DetectionData;
import sk.robotics.rlcs.constants.VlabConstants;

import java.awt.image.BufferedImage;
import java.awt.image.AffineTransformOp;
import java.awt.geom.AffineTransform;

import java.util.Set;

/**
 *  Agent for trasforming calibrated image from camera to our coordinates.
 */
public class ImageTransformAgent extends TransformAgent {

	public static class Cfg extends TransformAgent.Cfg {
		/** Name of block with detection data.
		 *  Detection data includes the AffineTransform object for transforming the
		 *  calibrated image to desk coordinates. It cuts off the borders, and rotates
		 *  the image as needed. The transformation is determined from detection data,
		 *  which should be stored in space by another agent.
		 *  @block.type sk.robotics.rlcs.data.DetectionData
		 *  @block.access read */
		public @Required @NoDefault String blockDetectionData = "detectionData";

		// default block names, and activation
		public Cfg() {
			blockSourceImage = "calibratedImage";
			blockTargetImage = "transformedImage";
			triggers = "$blockSourceImage, $blockDetectionData";
		}

		// hide the transform matrix fields in web editor
		public Set<String> getHiddenFields() {
			return hideFields(super.getHiddenFields(), "scaleX", "scaleY", "shearX", "shearY", "translateX", "translateY");
		}
	}
	private Cfg cfg;

	public ImageTransformAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	// transformation from calibrated image to lab coordinates
	private AffineTransform transform;

	@Override protected void senseSelectAct(int triggeredIds) {
		// get the transformation
		DetectionData detectionData = (DetectionData) space.read(cfg.blockDetectionData);
		if (detectionData == null) {
			// transformation is not available, no nothing
			return;
		}
		transform = detectionData.getTransform();

		// inherited senseSelectAct will take care of manipulating the images in blocks,
		// and call the filter() method, which we also override
		super.senseSelectAct(triggeredIds);
	}

	@Override protected void filter(BufferedImage source, BufferedImage target) {
		// we have transformation from calibrated image to lab coordinates,
		// but we still need to transform it to new image coordinates
		AffineTransform tx = new AffineTransform(
			target.getWidth() / (double)VlabConstants.WIDTH_MM, 0,
			0, -target.getHeight() / (double)VlabConstants.HEIGHT_MM,
			0, target.getHeight());
		transform.preConcatenate(tx);

		// update the operator
		transformOp = new AffineTransformOp(transform, cfg.interpolation.getValue());

		// transform the image
		super.filter(source, target);
	}
}
