package sk.robotics.rlcs.agents.camera;

import java.net.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.awt.geom.*;
import java.awt.*;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.DetectionData;
import sk.robotics.rlcs.constants.VlabConstants;

import sk.robotics.vlabcfg.VlabCfg;

/**
 *  Agent for parsing text file with detected positions of robots.
 *  This agent reads positions of reference points from vlabcfg.
 */
public class ReadDetectionAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** URL of the detection file. */
		public @Required @NoDefault String detectionFileURL = "file:///var/www/imgcalib/robots.txt";

		/** Only detect single robot. For debugging purposes. */
		public boolean singleRobot = false;

		/** Name of block where to store the detection data.
		 *  Output of this agent, for use by other agents.
		 *  @block.type sk.robotics.rlcs.data.DetectionData
		 *  @block.access write */
		public @Required @NoDefault String blockDetectionData = "detectionData";

		public Cfg() {
			// default activation
			timerPeriodMs = 300;
		}
	}
	private Cfg cfg;

	// the URL, constructed from config
	private URL url;

	// patterns for matching data in the file
	private static final Pattern patternRefPoint = Pattern.compile("\\[([0-9]+) ([0-9]+)\\]");
	private static final Pattern patternRobot = Pattern.compile("\\[([0-9]+) ([0-9]+) (-?[0-9]\\.[0-9]+)\\]");
	private static final Pattern patternDetectTime = Pattern.compile("[0-9]+ [0-9]+ [0-9]+ [0-9]+ [0-9]+ [0-9]+ [0-9]+\\.[0-9]+");
	private static final Pattern patternFrameTimes = Pattern.compile("[0-9]+ ([0-9]+) ([0-9]+)");

	// we'll remember the reference point, so that if one ofe them is zero sometimes, it doesn't hurt
	private Point2D[] rememberedRefPoints = new Point2D[3];

	// transformation matrix from base [r1-r0,r2-r0] in the calibrated image
	// to lab coordinates in millimetres (r0,r1,r2 are reference points)
	private AffineTransform ref2lab;

	// raw detection data
	public static class RawData {
		public Collection<Point2D> refPoints;
		public Collection<Point2D> rememberedRefPoints;
		public Collection<Position> robots;
		public long timestamp;

		public RawData(Collection<Point2D> refPoints, Collection<Point2D> rememberedRefs, Collection<Position> robots, long timestamp) {
			this.refPoints = Collections.unmodifiableCollection(refPoints);
			this.rememberedRefPoints = Collections.unmodifiableCollection(rememberedRefs);
			this.robots = Collections.unmodifiableCollection(robots);
			this.timestamp = timestamp;
		}
	}

	public ReadDetectionAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		// construct URL of detection file
		try { url = new URL(cfg.detectionFileURL); }
		catch (MalformedURLException e) {
			throw new RuntimeException(e.toString(), e);
		}

		// load refpoint locations from vlabcfg
		if (!VlabCfg.ready()) {
			throw new RuntimeException("VlabCfg not accessible.");
		}
		int p1x = getIntFromVlabCfg("refpoint1_pos_x");
		int p1y = getIntFromVlabCfg("refpoint1_pos_y");
		int p2x = getIntFromVlabCfg("refpoint2_pos_x");
		int p2y = getIntFromVlabCfg("refpoint2_pos_y");
		int p3x = getIntFromVlabCfg("refpoint3_pos_x");
		int p3y = getIntFromVlabCfg("refpoint3_pos_y");

		// compute transpormation matrix from refpoints base to lab coordinates
		ref2lab = new AffineTransform(p2x - p1x, p2y - p1y, p3x - p1x, p3y - p1y, p1x, p2x);

		/*rememberedRefPoints[0] = new Point2D.Double(83, 276);
		rememberedRefPoints[1] = new Point2D.Double(346, 46);
		rememberedRefPoints[2] = new Point2D.Double(603, 284);*/
	}
	private static int getIntFromVlabCfg(String name) {
		int val = VlabCfg.getInt(name, 1);
		if (val == -1) {
			throw new RuntimeException("Failed to load value '" + name + "' from vlabcfg.");
		}
		return val;
	}

	@Override protected void senseSelectAct() {
		BufferedReader reader = null;
		try {
			// we'll simulate a state machine while reading the file
			final int STATE_INIT = 0;
			final int STATE_REFPOINTS = 1;
			final int STATE_ROBOTS = 2;
			final int STATE_DETECTTIME = 3;
			final int STATE_FRAMETIME = 4;
			final int STATE_ACCEPT = 5;
			final int STATE_REJECT = 6;
			int state = STATE_INIT;

			// we'll parse these from the file
			Collection<Point2D> refPoints = new ArrayList<Point2D>(3);
			Collection<Position> robots = new ArrayList<Position>(3);
			Long detectionTime = null;
			Long frameTime = null;

			// read the file
			reader = new BufferedReader(new InputStreamReader(url.openStream()));
			while (state != STATE_ACCEPT && state != STATE_REJECT) {
				String line = reader.readLine();
				if (line == null) {
					// EOF
					if (state == STATE_FRAMETIME && frameTime != null) {
						state = STATE_ACCEPT;
					} else {
						log(0, "parse error: unexpected EOF");
						state = STATE_REJECT;
					}
					break;
				}
				if (line.length() == 0) continue; // skip empty lines

				switch (state) {
					case STATE_INIT:
						if (line.equals("Reference points")) {
							state = STATE_REFPOINTS;
						} else {
							state = STATE_REJECT;
							log(0, "parse error: expected reference points header");
						}
						break;
					case STATE_REFPOINTS:
						if (refPoints.size() < 3) {
							// we are expecting a ref point line
							Matcher matcher = patternRefPoint.matcher(line);
							if (matcher.matches()) {
								int x = Integer.parseInt(matcher.group(1));
								int y = Integer.parseInt(matcher.group(2));
								refPoints.add(new Point(x, y));
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected reference point coordinates");
							}
						} else {
							// we are expecting transition to robots state
							if (line.equals("Positions of robots")) {
								state = STATE_ROBOTS;
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected robots header");
							}
						}
						break;
					case STATE_ROBOTS:
						if (line.equals("Detect time")) {
							state = STATE_DETECTTIME;
						} else {
							Matcher matcher = patternRobot.matcher(line);
							if (matcher.matches()) {
								int x = Integer.parseInt(matcher.group(1));
								int y = Integer.parseInt(matcher.group(2));
								double h = Double.parseDouble(matcher.group(3));
								if (!cfg.singleRobot || robots.isEmpty())
									robots.add(new Position(x, y, h));
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected robot position or detection time header");
							}
						}
						break;
					case STATE_DETECTTIME:
						if (detectionTime == null) {
							// we are expecting the detection time
							Matcher matcher = patternDetectTime.matcher(line);
							if (matcher.matches()) {
								detectionTime = Long.valueOf(0); // we don't actually use this
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected detection time");
							}
						} else {
							// we are expecting transition to frame-time state
							if (line.equals("Frame download time")) {
								state = STATE_FRAMETIME;
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected frame download time header");
							}
						}
						break;
					case STATE_FRAMETIME:
						if (frameTime == null) {
							// we are expecting frame times
							Matcher matcher = patternFrameTimes.matcher(line);
							if (matcher.matches()) {
								long t1 = Long.parseLong(matcher.group(1));
								//long t2 = Long.parseLong(matcher.group(2));
								frameTime = Long.valueOf(t1 / 1000);
							} else {
								state = STATE_REJECT;
								log(0, "parse error: expected frame download times");
							}
						} else {
							// we are expecting end of file
							state = STATE_REJECT;
							log(0, "parse error: expected end of file");
						}
						break;
				}
			}

			// if we didn't finish in the accept state, quit
			if (state != STATE_ACCEPT) return;

			// store the raw detection data in space for debugging purposes
			RawData rawData = new RawData(refPoints, Arrays.asList(rememberedRefPoints), robots, frameTime);
			space.write(cfg.blockDetectionData + "/raw", rawData);

			// We'll create an AffineTransform, which will transform the detected robot positions
			// to our coordinates:
			//  1) the reference points r0,r1,r2 are on positions [0,h/2], [w/2,h], [w,h/2] in our coordinates,
			//     where w is width of the table, and h is height of the table
			//  2) we can easily compute coordinates in calibrated image of points [0,0], [1,0], [0,1]
			//     in our coordinates, lets call them p0,p1,p2
			//  3) from these points, we can easily construct matrix of affine transformation, inverse
			//     to the one we want
			//  4) we'll compute the inverse matrix, which defines the transformation we want

			// the reference points in calibrated image coordinates
			Point2D[] r = refPoints.toArray(new Point2D[0]);

			// the reference points have to be non-zero (otherwise, they were not detected)
			// (we also try to use rembered values from last time)
			for (int i = 0; i < 3; ++i) {
				if (r[i].getX() == 0 && r[i].getY() == 0) {
					r[i] = rememberedRefPoints[i];
					if (r[i] == null) {
						// ref point not detected, and we also dont have any remembered position for it
						log(0, "not all reference points detected!");
						return;
					}
				} else {
					rememberedRefPoints[i] = r[i];
				}
			}

			// coordinates transformation matrix from calibrated image in pixels to lab in millimeters
			AffineTransform transform = new AffineTransform(
				r[1].getX() - r[0].getX(), r[1].getY() - r[0].getY(),
				r[2].getX() - r[0].getX(), r[2].getY() - r[0].getY(),
				r[0].getX(), r[0].getY()).createInverse();
			transform.preConcatenate(ref2lab);

			// transform the detected robot positions
			Collection<Position> transformedRobots = new ArrayList<Position>(robots.size());
			double headingDiff = 0;//Math.atan2(-hx, -hy);
			for (Position pos: robots) {
				// transform position
				Point2D tp = transform.transform(pos.asPoint2D(), null);

				// transform heading (we'll transform a point one millimeter from pos,
				// in heading direction from it, and from the result compute transformed heading)
				double dir = pos.getHeading(); // 0 means heading east
				Point2D hp = new Point2D.Double(pos.getX() + Math.cos(dir), pos.getY() + Math.sin(dir));
				transform.transform(hp, hp);

				transformedRobots.add(new Position(tp.getX(), tp.getY(), Position.directionTo(tp, hp)));
			}

			// construct the detection data object
			DetectionData detectionData = new DetectionData(frameTime, transformedRobots, transform);

			// store it in space
			space.write(cfg.blockDetectionData, detectionData);
			log(4, "detection data written to space");
		}
		catch (Exception e) {
			log(0, e);
		}
		finally {
			try { if (reader != null) reader.close(); }
			catch (Exception e) {}
		}
	}
}
