package sk.robotics.rlcs.agents.camera;

import sk.robotics.rlcs.agents.image.ImageFilterAgent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.constants.VlabConstants;
import sk.robotics.rlcs.constants.TurtleConstants;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 *  Agent for drawing detected position and destinations of robots into an image.
 *  To be used in image processing pipelines. Draws given set of positions,
 *  and given set of destinations. The positions and destinations are fully independent,
 *  they do not have to correspond to the same set of robots.
 * <p>
 *  This agent is by default triggered by given position and destination blocks,
 *  as well as by the source image block. Note that it uses custom trigger ID
 *  for the source image block, which allows to distinguish between trigger sources,
 *  which in turn allows to ignore triggers caused by the position and destnation block,
 *  if there was no real change. This prevents unneccessary drawing, and thus saves
 *  processor time.
 */
public class ViewPosAgent extends ImageFilterAgent {

	public static class Cfg extends ImageFilterAgent.Cfg {
		/** Names of blocks with actual positions of robots.
		 *  Delimited by semicolons, commas, spaces, or combinations.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required @NoDefault String blocksPosition = "actualPos/robot?";

		/** Names of blocks with destination points of robots.
		 *  Delimited by semicolons, commas, spaces, or combinations.
		 *  @block.type java.awt.geom.Point2D
		 *  @block.access read */
		public @Required @NoDefault String blocksDestination = "destination/robot?";

		public Cfg() {
			// default activation
			triggers = "$blockSourceImage=0x02, $blocksPosition, $blocksDestination";

			// blit source image to target before filter()
			copyBeforeFilter = true;
		}
	}
	private final Cfg cfg;

	// names of blocks with positions and destinations
	private String[] posBlocks, destBlocks;

	// positions & destinations read in senseSelectAct(), stored to be used in filter()
	private Collection<Position> positions = new ArrayList<Position>();
	private Collection<Point2D> destinations = new ArrayList<Point2D>();

	// we remember last positions and destinations, to avoid unneccessary drawing
	private Collection<Position> lastPositions = new ArrayList<Position>();
	private Collection<Point2D> lastDestinations = new ArrayList<Point2D>();

	public ViewPosAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		posBlocks = cfg.blocksPosition.split("[;,\\s]+");
		destBlocks = cfg.blocksDestination.split("[;,\\s]+");
	}

	@Override protected void senseSelectAct(int triggeredIds) {
		// read current positions and destinations
		positions.clear();
		destinations.clear();
		for (String block: posBlocks) {
			Position pos = (Position) space.read(block);
			if (pos != null) {
				positions.add(pos);
			}
		}
		for (String block: destBlocks) {
			Point2D dest = (Point2D) space.read(block);
			if (dest != null) {
				destinations.add(dest);
			}
		}

		// if we were NOT trigerred by the source image, check whether the values
		// changed since last time, and if not, do nothing
		if ((triggeredIds & 0x02) == 0) {
			if (positions.equals(lastPositions) && destinations.equals(lastDestinations)) {
				log(5, "no change");
				return;
			}
		}

		// call the senseSelectAct() inherited from ImageFilterAgent
		// (it prepares the target image anc calls filter())
		super.senseSelectAct();

		// remember current positions and destinations as last ones
		Collection<Position> tmp1 = lastPositions;
		lastPositions = positions;
		positions = tmp1;
		Collection<Point2D> tmp2 = lastDestinations;
		lastDestinations = destinations;
		destinations = tmp2;
	}

	@Override protected void filter(BufferedImage image) {
		// if there is no position nor destination to draw, do nothing
		if (positions.isEmpty() && destinations.isEmpty()) return;

		// create a graphics context, and backup transformation
		Graphics2D g = image.createGraphics();
		AffineTransform backupTransform = g.getTransform();

		// setup coordinates transformation
		AffineTransform t = new AffineTransform();
		t.scale(image.getWidth() / (double)VlabConstants.WIDTH_MM,
			-image.getHeight() / (double)VlabConstants.HEIGHT_MM);
		t.translate(0, -VlabConstants.HEIGHT_MM);

		// draw position indicators
		for (Position pos: positions) {
			// setup transformation
			g.setTransform(t);
			g.translate(pos.getPosX(), pos.getPosY());
			g.rotate(-pos.getHeading());

			// draw arrow
			double w = TurtleConstants.DIAMETER_MM / 20;
			double l = w * 4;
			g.setPaint(Color.RED);
			g.setStroke(new BasicStroke((float)w));
			g.draw(new Line2D.Double(-l, -l, 0, 0));
			g.draw(new Line2D.Double( l, -l, 0, 0));
		}

		// draw destination indicators
		for (Point2D dest: destinations) {
			// setup transformation
			g.setTransform(t);
			g.translate(dest.getX(), dest.getY());

			// draw point
			double l = TurtleConstants.DIAMETER_MM / 5;
			double w = l / 3;
			g.setPaint(Color.RED);
			g.setStroke(new BasicStroke((float)w));
			g.draw(new Line2D.Double(-l, -l, l,  l));
			g.draw(new Line2D.Double(-l,  l, l, -l));
		}

		// restore original transformation
		g.setTransform(backupTransform);
	}
}
