package sk.robotics.rlcs.agents.image;

import java.awt.image.*;

/**
 *  Color inversion image filter.
 *  Does not work with indexed color models (which include black and white images).
 */
public class ConvolutionAgent extends ImageFilterAgent {

	public static class Cfg extends ImageFilterAgent.Cfg {
		/** Width of the convolution matrix. */
		public @Required @Limited(min=1,max=10) Integer matrixWidth;

		/** Height of the convolution matrix. */
		public @Required @Limited(min=1,max=10) Integer matrixHeight;

		/** The convolution matrix data. Line separated by semicolons,
		 *  and in each line values separated by spaces. */
		public @Required String matrixData;

		/** Copy edge pixels. If this is set, pixel on edges of the image are copied from
		 *  source to destination, otherwise the edges in destination are filled with zeroes. */
		public Boolean copyEdgePixels = false;

		// the matrix data, converted from string
		private transient float[] matrix;

		public void validate() {
			// parse and check the matrix data
			matrix = new float[matrixWidth * matrixHeight];
			String[] lines = matrixData.split(";");
			if (lines.length != matrixHeight) {
				throw new IllegalArgumentException("Bad number of lines in the matrix data!");
			}
			for (int y = 0, p = 0; y < matrixHeight; ++y) {
				String[] values = lines[y].trim().split(" +");
				if (values.length != matrixWidth) {
					throw new IllegalArgumentException("Bad number of values in line " + y + " of the matrix data!");
				}
				for (int x = 0; x < matrixWidth; ++x, ++p) {
					matrix[p] = Float.parseFloat(values[x]);
				}
			}
		}
	}

	// lookup opearator which does the inversion
	private ConvolveOp convolveOp;

	public ConvolutionAgent(Cfg cfg) {
		super(cfg);

		// construct the convolution operator
		Kernel kernel = new Kernel(cfg.matrixWidth, cfg.matrixHeight, cfg.matrix);
		convolveOp = new ConvolveOp(kernel, cfg.copyEdgePixels ? ConvolveOp.EDGE_NO_OP : ConvolveOp.EDGE_ZERO_FILL, null);
	}

	@Override protected void filter(BufferedImage source, BufferedImage target) {
		convolveOp.filter(source, target);
	}
}
