package sk.robotics.rlcs.agents.image;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import static java.awt.image.BufferedImage.*;

/**
 *  Base class for image processing agents.
 * <p>
 *  This agent alone (without deriving from it) can be used for converting
 *  images to different size and/or type, by setting non-zero width/height
 *  (to change size), outputImageType other than SRC (to change type),
 *  together with copyBeforeFilter. Note that changing the image size does
 *  not scale the image, you have to use TransformAgent to do that.
 * <p>
 *  It there is already an image in the target block, and is compatible
 *  (is of corrent size and type), this agent modifies <b>that</b> image
 *  (synchronizing on it), rather than creating a new one, to save memory.
 */
public class ImageFilterAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Name of block with the source image to draw to.
		 *  @block.type java.awt.BufferedImage
		 *  @block.access read */
		public @Required String blockSourceImage;

		/** Name of block where to store processed image.
		 *  @block.type java.awt.image.BufferedImage
		 *  @block.access write */
		public @Required String blockTargetImage;

		/** Priority of writing into target block.
		 *  Not set means copy priority of input block. */
		public Float priority;

		/** Validity of writing into target block.
		 *  Not set means copy validity of input block. */
		public Long validity;

		/** Type of input image.
		 *  If set, and the input image does not match it, the agent does not process it.
		 *  Possible values are GRAY (gray scale), BW (black and white), RGB (3-bands),
		 *  RGBA (4-bands). */
		public ImageType inputImageType;

		/** Type of output image.
		 *  Allowed values are RGB (24-bit), RGBA (32-bit), GRAY (gray scale),
		 *  BW (black and white). Not set means same as input. */
		public ImageType outputImageType;

		/** Width of output image. Not set means same as input. */
		public Integer outputWidth;

		/** Height of output image. Not set means same as output. */
		public Integer outputHeight;

		/** Blit source to target before filter()?.
		 *  Set to true for filters which copy the image, and draw something over it.
		 *  Or to use ImageFilterAgent directly as image type converter. */
		public Boolean copyBeforeFilter = false;

		public Cfg() {
			// default activation
			triggers = "$blockSourceImage";
		}

		// image types enum
		public static enum ImageType {
			RGB(TYPE_INT_RGB),
			RGBA(TYPE_INT_ARGB),
			GRAY(TYPE_BYTE_GRAY),
			BW(TYPE_BYTE_BINARY);

			private int value;
			private ImageType(int val) { value = val; }

			public int getValue() { return value; }
		}
	}
	private final Cfg cfg;

	private ColorModel outputColorModel;

	public ImageFilterAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		// init the output color model
		if (cfg.outputImageType != null) {
			outputColorModel = new BufferedImage(1, 1, cfg.outputImageType.getValue()).getColorModel();
		}

		// initial trigger
		trigger();
	}

	/** Filtering method. One of these two should be overriden.
	 *  Overriding the second overload makes sense if copyBeforeFilter is set. */
	protected void filter(BufferedImage source, BufferedImage target) { filter(target); }
	protected void filter(BufferedImage target) {}

	@Override protected void senseSelectAct() {
		// load the source image (with block meta data)
		Space.Block sourceBlock = space.getBlock(cfg.blockSourceImage);
		BufferedImage sourceImage = (sourceBlock == null) ? null : (BufferedImage)sourceBlock.getValue();
		if (sourceImage == null) {
			// if there is no source image, do nothing
			log(5, "no source image in block '" + cfg.blockSourceImage + "'");
			return;
		}

		// check the input image type
		if (cfg.inputImageType != null) {
			boolean ok = false;
			switch (cfg.inputImageType) {
				case RGB:
					ok = sourceImage.getColorModel().getNumComponents() >= 3;
					break;
				case RGBA:
					ok = sourceImage.getColorModel().getNumComponents() == 4;
					break;
				case GRAY:
					ok = sourceImage.getType() == TYPE_BYTE_GRAY;
					break;
				case BW:
					ok = sourceImage.getType() == TYPE_BYTE_BINARY && sourceImage.getColorModel().getPixelSize() == 1;
					break;
			}
			if (!ok) {
				log(0, "unsupported input image type");
				return;
			}
		}

		// determine desired output size
		int width = (cfg.outputWidth == null) ? sourceImage.getWidth() : cfg.outputWidth;
		int height = (cfg.outputHeight == null) ? sourceImage.getHeight() : cfg.outputHeight;

		// for SRC output type (same as input), query desired color model from the input image
		if (cfg.outputImageType == null) {
			outputColorModel = sourceImage.getColorModel();
		}

		// load target image
		BufferedImage targetImage = (BufferedImage) space.read(cfg.blockTargetImage);

		// if the target image already exists, and is compatible, we'll reuse it
		boolean reuse = targetImage != null && targetImage != sourceImage &&
		                targetImage.getWidth() == width && targetImage.getHeight() == height &&
		                targetImage.getColorModel().equals(outputColorModel);
		if (cfg.outputImageType == null) {
			// for SRC output (same as input), we'll also chech the BufferedImage type constant
			reuse = reuse && targetImage.getType() == sourceImage.getType();
		}

		if (!reuse) {
			// we need to create a new image
			WritableRaster raster;
			boolean alphaPremultiplied;
			if (cfg.outputImageType == null) {
				raster = sourceImage.getRaster().createCompatibleWritableRaster();
				alphaPremultiplied = sourceImage.isAlphaPremultiplied();
			} else {
				raster = outputColorModel.createCompatibleWritableRaster(sourceImage.getWidth(), sourceImage.getHeight());
				alphaPremultiplied = false;
			}
			targetImage = new BufferedImage(outputColorModel, raster, alphaPremultiplied, null);
		}

		// synchronize on both images while reading/changing their data
		synchronized (sourceImage) {
			synchronized (targetImage) {
				if (cfg.copyBeforeFilter) {
					if (sourceImage.getColorModel().isCompatibleSampleModel(targetImage.getRaster().getSampleModel())) {
						// raw copy
						sourceImage.copyData(targetImage.getRaster());
					} else {
						// the color models are not compatible, we have to draw
						Graphics g = targetImage.getGraphics();
						g.drawImage(sourceImage, 0, 0, null);
					}
				}

				// do the filtering
				filter(sourceImage, targetImage);
			}
		}

		// setup priority and validity
		float priority = (cfg.priority == null) ? sourceBlock.getPriority() : cfg.priority;
		long validFrom = (cfg.validity == null) ? sourceBlock.getValidFrom() : System.currentTimeMillis();
		long validTo = (cfg.validity == null) ? sourceBlock.getValidTo() : validFrom + cfg.validity;

		// [re]write the image into space
		space.write(cfg.blockTargetImage, targetImage, validFrom, validTo, priority);
		log(4, "image written to block '" + cfg.blockTargetImage + "'");
	}
}
