package sk.robotics.rlcs.agents.image;

import java.io.File;
import java.io.IOException;
import java.io.EOFException;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import javax.imageio.stream.ImageInputStreamImpl;

import sk.robotics.rlcs.agentspace.Agent;

/**
 *  Agent for loading images from files.
 *  Constructs a BufferedImage from raw image data, loaded from space.
 *  The raw image data can be loaded from a file using FileLoaderAgent.
 */
public class ImageLoaderAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Name of block where to read the image data from.
		 *  The block should contain loaded contents of an image file.
		 *  @block.type byte[]
		 *  @block.access read */
		public @Required String blockRawData;

		/** Name of block where to store the constructed image.
		 *  @block.type java.awt.image.BufferedImage
		 *  @block.access write */
		public @Required String blockImage;

		// default activation
		public Cfg() {
			triggers = "$blockRawData";
		};
	}
	private Cfg cfg;

	// ImageInputStream implementation backed by byte array
	private static class ImgInput extends ImageInputStreamImpl {
		private byte[] data;
		private ImgInput(byte[] data) { this.data = data; }

		public int read() throws IOException {
			if (streamPos >= data.length) throw new EOFException();
			int val = data[(int)streamPos++];
			return val < 0 ? val + 256 : val;
		}
		public int read(byte[] buff, int off, int len) throws IOException {
			if (streamPos >= data.length) throw new EOFException();
			if (streamPos + len > data.length) len = (int)(data.length - streamPos);
			System.arraycopy(data, (int)streamPos, buff, off, len);
			streamPos += len;
			return len;
		}
	}

	// error flag, avoids flooding the log on erroneous input
	private boolean error;

	public ImageLoaderAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		// load the raw data from space
		byte[] raw = (byte[]) space.read(cfg.blockRawData);

		// construct the image
		BufferedImage image;
		try {
			image = ImageIO.read(new ImgInput(raw));
		} catch (IOException e) {
			// this shouldn't happen
			throw new IllegalStateException();
		}
		if (image == null) {
			if (!error) {
				log(0, "The image file is in unknown format.");
			}
			error = true;
			return;
		}
		error = false;

		// store the image in space
		space.write(cfg.blockImage, image);
	}
}
