package sk.robotics.rlcs.agents.image;

import java.net.*;
import java.io.*;
import java.nio.channels.*;
import java.awt.image.RenderedImage;
import javax.imageio.ImageIO;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Scheduler;

/**
 *  Agent which listens for clients, who want to receive images.
 *  This agent listens for connections, and for each new connection,
 *  starts through Scheduler a new instance of ImageStreamAgent,
 *  which then sends to the client images from space, whenever
 *  given image containing block is triggered.
 * <p>
 *  This can be also used in cooperation with PHP to stream MJPEG through HTTP.
 * <p>
 *  This agent should be triggered periodically. It does not listen all the time,
 *  but just periodically checks for pending connections, and accepts them.
 */
public class ImageServerAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** IP address to listen on. Not set means all interfaces/addresses. */
		public String listenAddress;

		/** TCP port to listen on. */
		public @Limited(min=1,max=65535) int listenPort = 35755;

		/** Prefix to send before each image.
		 *  This will be used as formatting string in call to String.format(),
		 *  using single formatting argument, the length of the image in bytes.
		 *  So %d can be used in this string, and will expand to the image length. */
		public String imagePrefix = "\r\n===boundary===\r\nContent-Type: image/jpeg\r\nContent-Length: %d\r\n\r\n";

		/** Suffix to send after each image.
		 *  This string is just sent, no formatting is performed. */
		public String imageSuffix = "\r\n";

		/** Format of sent image.
		 *  Allowed are values accepted by default by javax.imageio.ImageIO. */
		public String imageFormat = "jpg";

		/** Name of block with the image to generate stream from.
		 *  @block.type java.awt.image.RenderedImage
		 *  @block.access read */
		public @Required @NoDefault String blockImage = "processedImage";

		public Cfg() {
			// default activation
			timerPeriodMs = 100;
		}

		@Override public void validate() {
			// check image format
			if (!ImageIO.getImageReadersByFormatName(imageFormat).hasNext()) {
				throw new IllegalArgumentException("Invalid image format!");
			}
		}
	}
	private Cfg cfg;

	// server socket channel
	private ServerSocketChannel server;

	public ImageServerAgent(Cfg cfg)
	{
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct()
	{
		try {
			// recreate the server socket channel if needed
			if (server == null) {
				// create channel
				server = ServerSocketChannel.open();

				// set to non-blocking mode
				server.configureBlocking(false);
			}

			// bind the server socket, if not yet done
			if (server != null && !server.socket().isBound()) {
				// bind
				SocketAddress bindAddr = (cfg.listenAddress == null)
					? new InetSocketAddress(cfg.listenPort)
					: new InetSocketAddress(cfg.listenAddress, cfg.listenPort);
				server.socket().setReuseAddress(true);
				server.socket().bind(bindAddr);
			}

			// accept a connection
			SocketChannel client = server.accept();
			if (client == null) {
				return; // no pending connections
			}

			// launch a streaming agent for this connection
			ImageStreamAgent.Cfg streamCfg = new ImageStreamAgent.Cfg();
			streamCfg.channel = client;
			streamCfg.imagePrefix = cfg.imagePrefix;
			streamCfg.imageSuffix = cfg.imageSuffix;
			streamCfg.imageFormat = cfg.imageFormat;
			streamCfg.blockImage = cfg.blockImage;
			Scheduler.getInstance().launch(streamCfg);
		}
		catch (IOException e) {
			log(0, e, "I/O error, closing server socket.");

			// close the socket, we'll try to listen again
			try { server.close(); }
			catch (IOException ee) {}
			server = null;

			// sleep a while before listening again
			sleep(3000);
		}
	}

	@Override protected void cleanup() {
		if (server != null) {
			try { server.close(); }
			catch (IOException e) {}
			server = null;
		}
	}
}
