package sk.robotics.rlcs.agents.image;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.awt.image.RenderedImage;
import javax.imageio.ImageIO;

import sk.robotics.rlcs.agentspace.Agent;

/**
 *  Agent for streaming images to clients.
 *  This agent is not meant to be used directly. It is launched by ImageServerAgent.
 *  It generates a image stream from a (changing) image stored in space,
 *  and streams it through a socket channel.
 */
public class ImageStreamAgent extends Agent {

	public static @Hidden class Cfg extends Agent.Cfg {
		/** The channel to stream the video through. */
		public SocketChannel channel;

		/** Prefix to send before each image.
		 *  See ImageServerAgent.Cfg for details. */
		public String imagePrefix;

		/** Suffix to send after each image. */
		public String imageSuffix;

		/** Format of image. */
		public String imageFormat;

		/** Name of block with the image to generate stream from.
		 *  @block.type java.awt.image.RenderedImage
		 *  @block.access read */
		public String blockImage;

		public Cfg() {
			// default activation
			triggers = "$blockImage";
		}
	}
	private Cfg cfg;

	// buffer for encoding image into
	// (we derive from ByteArrayOutputStream to access its buffer directly)
	private static class Buffer extends ByteArrayOutputStream {
		public ByteBuffer asByteBuffer() {
			return ByteBuffer.wrap(buf);
		}
	}
	private Buffer imageBuffer = new Buffer();

	// a selector & key used to detect socket's writability
	private Selector selector;
	private SelectionKey writable;

	public ImageStreamAgent(Cfg cfg)
	{
		super(cfg);
		this.cfg = cfg;

		try {
			// create a selector
			selector = Selector.open();

			// set the socket to non-blocking mode (required by Selector)
			cfg.channel.configureBlocking(false);

			// register the socket's write operation with the selector
			writable = cfg.channel.register(selector, SelectionKey.OP_WRITE, null);
		}
		catch (IOException e) {
			log(0, e, "I/O error, exiting agent.");
			exit(); // quit the agent
		}

		// send initial image immediately
		trigger();
	}

	@Override protected void senseSelectAct()
	{
		try {
			// wait until the socket gets writable
			for (;;) {
				selector.select();

				// if the select() call was interrupted, quit this sense-select-act cycle
				if (thread.interrupted()) {
					return;
				}

				// is the socket already writeble?
				if (selector.selectedKeys().contains(writable)) {
					selector.selectedKeys().remove(writable);
					break;
				}
			}

			// get the image from space
			RenderedImage image = (RenderedImage) space.read(cfg.blockImage);
			if (image == null) {
				log(1, "no image in block '" + cfg.blockImage + "'"); // shouldn't happen
				return;
			}

			// encode the image into JPEG (synchronizing on the image while reading it)
			imageBuffer.reset();
			synchronized (image) {
				if (!ImageIO.write(image, cfg.imageFormat, imageBuffer)) {
					// this really shouldn't happen
					throw new RuntimeException("image writer for JPEG not found!");
				}
			}

			// construct array of buffers to send through the channel
			ByteBuffer[] buffers = new ByteBuffer[3];
			int usedBuffers = 0;

			// prefix buffer
			if (cfg.imagePrefix != null && cfg.imagePrefix.length() > 0) {
				String prefix = String.format(cfg.imagePrefix, imageBuffer.size());
				buffers[usedBuffers++] = ByteBuffer.wrap(prefix.getBytes());
			}

			// image buffer
			buffers[usedBuffers++] = imageBuffer.asByteBuffer();

			// suffix buffer
			if (cfg.imageSuffix != null && cfg.imageSuffix.length() > 0) {
				buffers[usedBuffers++] = ByteBuffer.wrap(cfg.imageSuffix.getBytes());
			}

			// send everything through the channel
			cfg.channel.write(buffers, 0, usedBuffers);
		}
		catch (ClosedByInterruptException e) {}
		catch (IOException e) {
			if (e.getMessage().indexOf("forcibly closed") >= 0 ||
			    e.getMessage().indexOf("established connection was aborted") >= 0) {
				log(0, "connection closed");
			} else {
				log(0, e, "I/O error, closing connection.");
			}

			// close the socket
			try { cfg.channel.close(); }
			catch (IOException ee) {}

			// exit the agent, it is not needed anymore
			exit();
		}
	}

	@Override protected void cleanup() {
		// on exit, close the socket
		try { cfg.channel.close(); }
		catch (IOException e) {}
	}
}
