package sk.robotics.rlcs.agents.os;

import sk.robotics.rlcs.agentspace.Agent;

import java.util.Map;
import java.util.LinkedHashMap;
import java.io.IOException;

/**
 *  Agent for executing a command.
 *  This agent lauches new processes. It supports not only regularly
 *  executing fixed command, but also using values from space as
 *  arguments to the command, or even as the command itself.
 */
public class ExecuterAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** The command to execute. This is a format string, which will be passed
		 *  to String.format() before each execution. Format arguments are loaded
		 *  from space, see parameters field. */
		public @Required String command;

		/** Blocks whose values will be used as format arguments to construct the
		 *  command. Delimited by spaces, commas, semicolons, or combinations of them.
		 *  Block with default value (used when there is nothing in the block) may be
		 *  specified as block:default. If no default value is specified for a block,
		 *  empty string is used. */
		public String parameters;

		/** Launch next command, only if previous execution has already finished. */
		public Boolean onlyIfFinished = true;
	}
	private Cfg cfg;

	// parameter blocks, and default values
	// (must be a LinkedHashMap, and not any other Map type, to preserve order of entries)
	private Map<String, String> blocks = new LinkedHashMap<String, String>();

	// the last launched process
	private Process process;

	public ExecuterAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		// construct the parameters map
		if (cfg.parameters != null) {
			String[] blockSpecs = cfg.parameters.split("[ ,;]+");
			for (String blockSpec: blockSpecs) {
				if (blockSpec.length() == 0) continue; // ignore leading & trailing space
				String[] parts = blockSpec.split(":");
				if (parts.length > 2) {
					throw new IllegalArgumentException("More than one colon in block specification!");
				}
				blocks.put(parts[0], (parts.length == 2) ? parts[1] : "");
			}
		}
	}

	@Override protected void senseSelectAct() {
		// if the last command didn't finish yet, do nothing
		if (process != null) {
			try { process.exitValue(); }
			catch (IllegalThreadStateException e) {
				return;
			}
			process = null; // not needed anymore
		}

		// load parameters from space
		Object[] params = null;
		if (!blocks.isEmpty()) {
			params = new Object[blocks.size()];
			int paramIndex = 0;
			for (Map.Entry<String, String> entry: blocks.entrySet()) {
				Object val = space.read(entry.getKey());
				params[paramIndex++] = (val == null) ? entry.getValue() : val;
			}
		}

		// construct the command
		String command = String.format(cfg.command, params);

		// execute the command
		try { process = Runtime.getRuntime().exec(command); }
		catch (IOException e) {
			throw new RuntimeException("Failed to execute command", e);
		}
	}

	@Override protected void cleanup() {
		// if a command we launched is still running, kill it
		if (process != null) {
			try { process.exitValue(); }
			catch (IllegalThreadStateException e) {
				log(0, "forcibly terminating command process");
				process.destroy();
			}
		}
	}
}
