package sk.robotics.rlcs.agents.os;

import sk.robotics.rlcs.agentspace.Agent;

import java.io.*;
import java.nio.charset.Charset;

/**
 *  Agent for loading content of a file into space.
 *  It supports two modes: binary and text. In binary mode,
 *  contents of the file are stored in space as raw array of
 *  bytes, while in text mode, the file is read into a string,
 *  possibly using a custom charset, and stored in space as string.
 */
public class FileLoaderAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Path to the file to load. */
		public @Required String path;

		/** File type, Binary or text.
		 *  Specifies type of data to store in space: string or byte array. */
		public FileType fileType = FileType.BINARY;
		public enum FileType { BINARY, TEXT }

		/** Character encoding used in the file.
		 *  Applicable obly for text files, ignored for binary files.
		 *  If not specified for a text file, a system-default charset is used. */
		public String charset;

		/** Reload the file only if it was modified since last time. */
		public Boolean loadOnlyModified = true;

		/** Name of block where to store content of the file.
		 *  A string is stored there for text files, an array of bytes for for binary files.
		 *  @block.type String/byte[]
		 *  @block.access write */
		public @Required String blockContent;

		// validate the charset
		@Override public void validate() {
			if (fileType == FileType.TEXT) {
				Charset.forName(charset); // throws if charset not found
			}
		}

		// default activation
		public Cfg() {
			timerPeriodMs = 100;
		}
	}
	private Cfg cfg;

	// maximal file size to allow to load
	private static int MAX_FILE_SIZE_MB = 1;

	// the file (it is more convenient to have it as File, rather than String)
	private File file;

	// timestamp of last loaded version of the file
	private long lastTimeStamp = 0;

	// error codes, and remembered last error code, to prevent repeated error logs
	private int lastError = 0;
	private static final int ERROR_EXIST  = 1; // file does not exist
	private static final int ERROR_READ   = 2; // file read error

	public FileLoaderAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		file = new File(cfg.path);
	}

	@Override protected void senseSelectAct() {
		// get current timestamp of the file
		long timeStamp = file.lastModified();
		if (timeStamp == 0) {
			// the file does not exist, or an I/O error occured
			if (lastError != ERROR_EXIST) {
				log(0, "The file does not exist.");
			}
			lastError = ERROR_EXIST;
			return;
		}

		// check whether the file changed since last load
		if (cfg.loadOnlyModified && timeStamp == lastTimeStamp) {
			return;
		}

		// read the file
		long length = file.length();
		if (length > MAX_FILE_SIZE_MB * 1024 * 1024) {
			throw new RuntimeException("File too big!");
		}
		byte[] buffer = new byte[(int)length];
		try {
			InputStream in = new FileInputStream(file);
			if (in.read(buffer) != length) {
				throw new IOException();
			}
		} catch (IOException e) {
			if (lastError != ERROR_READ) {
				log(0, "File read error.");
			}
			lastError = ERROR_READ;
			return;
		}

		// remember last timestamp
		lastTimeStamp = timeStamp;

		// store the contents in space
		if (cfg.fileType == Cfg.FileType.BINARY) {
			space.write(cfg.blockContent, buffer);
		} else {
			try { space.write(cfg.blockContent, new String(buffer, cfg.charset)); }
			catch (UnsupportedEncodingException e) { throw new RuntimeException(e); }
		}
	}
}
