package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;
import sk.robotics.rlcs.data.Movement;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.Collision;

import java.awt.geom.Point2D;

/**
 *  Agent for avoiding an obstacle.
 *  This agent detects situations when robot is not moving, because of an imminent collision.
 *  It then tries to avoid the collision, so that the robot can continue to it's destination.
 *  It should be configured with greater priority of writing movement commands than MoveAgent,
 *  otherwise they could interfere in an unwanted way.
 * <p> Once this agent detects an avoid situation, it starts moving the robot. By writing first
 *  value to the desired movement block, it effectively disables associated MoveAgent, because
 *  if higher write priority. The MoveAgent starts working again only once this agent deletes
 *  the block, and it does that only once there is again free path to destination.
 */
public class AvoidAgent extends MoveAgentBase {

	public static class Cfg extends MoveAgentBase.Cfg {
		/** Obstacle distance considered as collision. */
		public Double collisionDistanceMm = 10.0;

		/** Name of block with current Collision object.
		 *  The collision object is used to detect imminent collisions.
		 *  @block.type sk.robotics.rlcs.data.Collision
		 *  @block.access read */
		public @Required @NoDefault String blockCollision = "collision/robot?";

		public Cfg() {
			// higher priority by default, to suppress ApproachAgent
			priority = 1.0f;
		}
	}
	private Cfg cfg;

	public AvoidAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	// current avoiding state
	private boolean avoiding = false;

	@Override protected Movement selectMove(Position pos, Point2D dest) {
		// get current collision-computing object from space
		Collision collision = (Collision) space.read(cfg.blockCollision);
		if (collision == null) {
			// it there is no collision object, we have nothing to do...
			return null;
		}

		// if we are not yet in avoid mode, enter it if there is an obstacle before the robot
		if (!avoiding) {
			double dir = pos.directionTo(dest);
			Collision.CheckResult collres = collision.checkMove(dir, cfg.collisionDistanceMm);
			if (collres.getCollisionDistance() > 0) {
				// no imminent collision
				return null;
			}

			log(1, "avoiding obstacle");
			avoiding = true;
		}

		// by default, we'll move forward
		Movement.Direction move = Movement.Direction.FD;

		// check for collision right before us
		Collision.CheckResult collRes = collision.check(move, 10);

		// determine movement direction
		if (collRes.getCollisionDistance() <= 0) {
			// if we can't move forward, turn, and choose direction according
			// to where is the obstacle which would cause the collision
			double collDir = pos.headingDiff(collRes.getObstacleDirection());
			if (collDir > 0) {
				move = Movement.Direction.RT;
			} else if (collDir < 0) {
				move = Movement.Direction.LT;
			} else {
				// choose a side randomly
				move = (Math.random() < 0.5) ? Movement.Direction.LT : Movement.Direction.RT;
			}
		} else {
			// we can move forward without collision
			double dir = pos.directionTo(dest);
			double dirDiff = -pos.headingDiff(dir);
			if (Math.abs(dirDiff) < cfg.directionTolerance) {
				// we have free path to the destination, leave the avoid mode
				// (returning NO_MOVE from here will delete the movement block,
				//  which in turn will allow MoveAgent to start working again)
				log(1, "obstacle avoided");
				avoiding = false;
				return Movement.NO_MOVE;
			} else {
				// check whether it is possible to turn to the destination, and then move forward
				Collision.CheckResult turnCollRes = collision.checkMoveRel(Math.signum(dirDiff) * 0.1, 10);
				if (turnCollRes.getCollisionDistance() > Math.min(100, pos.distance(dest))) {
					// we can turn to destination, so do it
					move = (dirDiff > 0) ? Movement.Direction.RT : Movement.Direction.LT;
				}
				// otherwise continue moving forward
			}
		}

		return Movement.valueOf(move, 99, 99);
	}
}
