package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Collision;
import sk.robotics.rlcs.data.Movement;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.DetectionData;
import sk.robotics.rlcs.constants.TurtleConstants;
import sk.robotics.rlcs.constants.VlabConstants;

import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.util.Collection;
import java.util.ArrayList;

/**
 *  Agent for generating and storing in space a collision computing object.
 *  The collision object is bound to a single robot. It is then used by other
 *  agents to determine obstacle distances in all directions from the robot.
 */
public class CollisionAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Identifier, associated with robot recognized from camera image.
		 *  This is used to exclude the robot from detected obstacles.
		 *  The association of the identifier with the detected robot is done
 		 *  by another agent, so this field must match between these agents. */
		public @Required Integer robotNum;

		/** Name of block with data from image recognition.
		 *  This is used to determine the set of obstacles. 
		 *  @block.type sk.robotics.rlcs.data.DetectionData
		 *  @block.access read */
		public @Required @NoDefault String blockDetectionData = "detectionData";

		/** Name of block with actual position of the robot.
		 *  Actual position of the robot is used by the Collision object
		 *  to compute obstacle distances.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read  */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		/** Name of block with current Collision object.
		 *  For use by other agents to detect collisions.
		 *  @block.type sk.robotics.rlcs.data.Collision
		 *  @block.access write */
		public @Required @NoDefault String blockCollision = "collision/robot?";

		public Cfg() {
			// default activation
			timerPeriodMs = 10;
		}
	}
	private final Cfg cfg;

	public CollisionAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	private static final Collection<Line2D> walls = new ArrayList<Line2D>();
	static {
		Point2D[] p = {
			new Point2D.Double(0, 0),
			new Point2D.Double(0, VlabConstants.HEIGHT_MM),
			new Point2D.Double(VlabConstants.WIDTH_MM, VlabConstants.HEIGHT_MM),
			new Point2D.Double(VlabConstants.WIDTH_MM, 0)
		};
		walls.add(new Line2D.Double(p[0], p[1]));
		walls.add(new Line2D.Double(p[1], p[2]));
		walls.add(new Line2D.Double(p[2], p[3]));
		walls.add(new Line2D.Double(p[3], p[0]));
	}

	@Override protected void senseSelectAct() {
		// get current position
		Position pos = (Position)space.read(cfg.blockActualPos);

		// construct collision object
		Collision collision;
		if (pos == null) {
			// if position is not given, simulate collision in all directions (to not move anywhere)
			collision = Collision.ALL;
		} else {
			// find other robots
			DetectionData detectionData = (DetectionData)space.read(cfg.blockDetectionData);
			Collection<Position> robots = new ArrayList<Position>(detectionData == null ? 0 : detectionData.getRobots().size());
			if (detectionData != null) {
				for (DetectionData.Robot r: detectionData.getRobots()) {
					Position p;
					// check whether this position is associated to a robot
					Integer robotNum = r.robotNum;
					if (robotNum != null) {
						// yes, it is
						if (robotNum.intValue() == cfg.robotNum)
							continue; // that's us, no collision
						// use actual position of the robot, instead of the detected one
						p = (Position)space.read(r.blockActualPos);
						if (p == null) p = r.position; // no actual position, use the detected one
					}
					else {
						// no, this position is not associated to any robot, use it directly
						p = r.position;

						if (p.distance(pos) < 80 && p.headingDiff(pos) < Math.PI/2)
							continue; // that's us, ignore
					}

					robots.add(p);
				}
			}

			// the collision object
			collision = new Collision(pos, walls, robots);
		}

		// store the collision object in space
		space.write(cfg.blockCollision, collision);
	}
}
