package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.robot.Robot;

import java.net.InetAddress;

/**
 *  Agent for managing connection to the robot.
 *  This agent stores the robot control object in space once it establishes
 *  connection to the robot, and deletes it from there if the connection gets down.
 *  Once the Robot object is in space, all access to it by this agent is synchronized
 *  on the object itself, and so must be by other agents as well.
 */
public class ConnectionAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Host name or IP address where the robot daemon runs. Not set means localhost. */
		public String robotdHost;

		/** TCP port on which the robot daemon is listening.
		 *  Can be not-set, if robot is set (see below). */
		public @Limited(min=1,max=65535) Integer robotdPort;

		/** User name to use for connection to the robot.
		 *  Does not have to be set, there is a working default username, which will
		 *  be used if this is not set. However, it is possible to register in the
		 *  robotics laboratory for exclusive access robots. */
		public String username;

		/** Password for use with username.
		 *  Set this only if you also set username, and if you have registered in
		 *  the robotics laboratory for exclusive access to robots. */
		public String password;

		/** Predefined robot to connect to. Alternative to setting robotdPort directly. */
		public PredefinedRobot robot;
		public static enum PredefinedRobot {
			REAL1(7123), REAL2(3331), REAL3(3332), SIM1(3333), SIM2(3334);

			PredefinedRobot(int port) { this.port = port; }

			private int port;
			public int getPort() { return port; }
		}

		/** Request exclusive access to the robot?
		 *  True if we want to control the robot, false if we only want to track its position. */
		public boolean exclusive = true;

		/** Name of block with the robot control object, made available for other agents.
		 *  However, they must synchronize on it to use it.
		 *  @block.type sk.robotics.robot.Robot
		 *  @block.access read/write */
		public @Required @NoDefault String blockComm = "comm/robot?";

		// validation
		@Override public void validate() {
			// either robot, or robotdPort, but not both, must be set
			if (!((robot != null) ^ (robotdPort != null))) {
				throw new IllegalArgumentException("Either robot or robotdPort (but not both) must be set.");
			}
			super.validate();
		}

		public Cfg() {
			// default activation
			timerPeriodMs = 2000;
		}
	}
	private final Cfg cfg;

	// resolved IP address, and port
	private String robotdIP;
	private int robotdPort;

	// default user name and password for connection
	private static final String DEFAULT_USERNAME = "ui";
	private static final String DEFAULT_PASSWORD = "turtle";

	// this helps to avoid duplicate fail messages
	private int lastError = 0;
	private static final int ERROR_CONNECT = 1;
	private static final int ERROR_EXCLUSIVE = 2;
	private static final int ERROR_ISALIVE = 3;

	public ConnectionAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		try { robotdIP = InetAddress.getByName(cfg.robotdHost).getHostAddress(); }
		catch (Exception e) { throw new RuntimeException(e); }

		robotdPort = (cfg.robotdPort == null) ? cfg.robot.getPort() : cfg.robotdPort;
	}

	@Override protected void senseSelectAct() {
		// get the robot object
		Robot robot = (Robot)space.read(cfg.blockComm);

		// if connection was already established...
		if (robot != null) {
			// ...check whether it is still ok (and do nothing if it is)
			synchronized (robot) {
				if (robot.isalive() == 1) {
					return; // OK
				}

				// connection is down, close the robot object, will create a new one
				robot.destructor();
			}

			// connection is down
			space.delete(cfg.blockComm);
			log(0, "disconnected from the robot.");
			sleep(1000);
			return;
		}

		int error = 0;

		// try to [re]connect
		try { robot = new Robot(0); } // 0 - no message boxes
		catch (Exception e) {
			log(0, e, "failed to create Robot object");
			return;
		}
		robot.user(cfg.username == null ? DEFAULT_USERNAME : cfg.username,
		           cfg.password == null ? DEFAULT_PASSWORD : cfg.password);
		robot.portnumber(robotdPort);

		if (robot.connect(robotdIP) == 1) {
			// connection went ok
			log(0, "connected to robot.");
		} else {
			// connection failed
			error = ERROR_CONNECT;
			//if (lastError != error) {
				log(0, "failed to connect to the robot.");
			//}
		}

		// request exclusivity, if configured
		if (error == 0 && cfg.exclusive) {
			String exclRes = robot.exclusive("on");
			if (exclRes.equals("ok")) {
				// OK, we have excluive access to the robot
				log(1, "exclusive robot access granted.");
			} else {
				// exclusive access has been rejected
				error = ERROR_EXCLUSIVE;
				//if (lastError != error) {
					log(0, "exclusive robot access rejected: '" + exclRes + "'");
				//}
			}
		}

		// check whether the robot is alive
		if (error == 0) {
			if (robot.isalive() == 1) {
				// OK, the robot is alive
				log(1, "the robot is alive");
			} else {
				// the robot si dead
				error = ERROR_ISALIVE;
				log((lastError != error) ? 0 : 4, "the robot is dead");
			}
		}

		// if everything went ok, store the robot control object in space
		if (error == 0) {
			space.write(cfg.blockComm, robot);
		} else { // otherwise, close and drop it
			log(2, "there was an error, closing the Robot object");
			robot.destructor();
		}

		// remember last error, to avoid logging one error more than once
		lastError = error;
	}

	// on exit, close the robot object (if connected)
	@Override protected void cleanup()
	{
		Robot robot = (Robot)space.delete(cfg.blockComm);
		if (robot != null) {
			synchronized (robot) {
				robot.destructor();
			}
		}
	}
}
