package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.TimedPosition;
import sk.robotics.rlcs.data.PositionQueue;
import sk.robotics.rlcs.data.DetectionData;
import sk.robotics.rlcs.constants.TurtleConstants;
import sk.robotics.robot.Robot;

import java.util.Collection;
import java.util.Date;

/**
 *  Agent for identifying a robot in the detection data acquired from
 *  image recognition.
 * <p> If we have exclusive connection to the robot,
 *  this agent turns the robot to recognize it in the detection data.
 *  If we don't have exclusive access to the robot, it is still possible
 *  to identify it, but it has to be the only one remaining unidentified
 *  robot on the table.
 * <p> Once the robot is identified, this agent tries to track its
 *  position by comparing actual position with positions of detected robots.
 *  For this to work, OdometryAgent is used in cooperation.
 */
public class IdentifyAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Identifier, associated upon identification with robot
		 *  recognized from camera image. This identifier is then
		 *  stored in the detection, so can be used by other agents. */
		public @Required Integer robotNum;

		/** Exclusive access to the robot. True if we can control the robot, false otherwise. */
		public boolean exclusive = true;

		/** Distance tolerance for comparing detected positions with the actual one, in millimeters.
		 *  This is used when tracking position of already identified robot. */
		public Double moveToleranceDistance = 30.0;
		/** Heading tolerance for comparing detected positions with the actual one, in radians.
		 *  This is used when tracking position of already identified robot. */
		public Double moveToleranceHeading = Math.PI / 4;

		/** Distance tolerance for comparing detected positions after turning the robot, in millimeters.
		 *  This is used when identifying robot by turning him. */
		public Double turnToleranceDistance = 2.0;
		/** Heading tolerance for comparing detected positions after turning the robot, in radians.
		 *  This is used when identifying robot by turning him. */
		public Double turnToleranceHeading = Math.PI / 20;

		/** Name of block with the robot control object.
		 *  The robot control object is used to turn the robot, in order to identify it.
		 *  Access to the object is synchronized (on the object itself).
		 *  @block.type robot.Robot
		 *  @block.access read */
		public @Required @NoDefault String blockComm = "comm/robot?";

		/** Name of block with data from image recognition.
		 *  @block.type sk.robotics.rlcs.data.DetectionData
		 *  @block.access read */
		public @Required @NoDefault String blockDetectionData = "detectionData";

		/** Name of block with history of actual positions.
		 *  This is used for position tracking purposes (by determining position of the robot from
		 *  time of detection), and also updated with actual position on sucessfull robot identification.
		 *  @block.type sk.robotics.rlcs.data.PositionQueue
		 *  @block.access read/write */
		public @Required @NoDefault String blockPosHistory = "posHistory/robot?";

		/** Name of block with current desired movement.
		 *  By deleting this block with high priority and non-zero validity,
		 *  this agent disables other agents trying to move the robot.
		 *  @block.type sk.robotics.rlcs.data.Movement
		 *  @block.access write */
		public @Required @NoDefault String blockMovement = "movement/robot?";

		/** Name of block with last detected position of the robot.
		 *  For use by other agents.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access write */
		public @Required @NoDefault String blockDetectedPos = "detectedPos/robot?";

		/** Name of block with actual position of the robot.
		 *  This agent updates the actual position upon successfull identification.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access write */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		public Cfg() {
			// default activation
			triggers = "$blockDetectionData";
		}
	}
	private final Cfg cfg;

	// position history size, in milliseconds
	private static final long POSITION_HISTORY_SIZE = 2000;

	public IdentifyAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		// get detected robot positions from space
		DetectionData detectionData = (DetectionData)space.read(cfg.blockDetectionData);
		if (detectionData == null) {
			log(5, "no detection data");
			return; // no detected robots yet
		}

		// we'll try to find the right one in the detected positions
		Position detected = null;

		// scan the detection data, and determine whether we already did find ourselves in this data,
		// and also check for the possibility that there is only one remaining not-yet-identified detected robot
		// (this allows us to identify a robot we cannot control, if there is only one such)
		DetectionData.Robot unidentified = null;
		boolean multipleUnidentified = false; // are there more than one unidentified robots?
		for (DetectionData.Robot r: detectionData.getRobots()) {
			if (r.robotNum == null) {
				// an unidentified robot, if it is the only one, we'll asume it's us
				if (unidentified == null)
					unidentified = r;
				else multipleUnidentified = true;
			}
			else if (r.robotNum.intValue() == cfg.robotNum) {
				// we have already identified ourselves in this detection data!
				log(4, "already identified");
				return;
			}
		}
		if (detected == null && unidentified != null && !multipleUnidentified && unidentified.robotNum == null) {
			// there is already one not yet identified robot left in the detection data
			detected = unidentified.position;
			unidentified.identify(cfg.robotNum, cfg.blockActualPos);
			space.trigger(cfg.blockDetectionData); // we modified the object stored in space

			// we identified the robot, but the detected position is not the actual one,
			// but an older (delayed) one, so we need to do transform the position to
			// present time, using position history
			PositionQueue history = (PositionQueue) space.read(cfg.blockPosHistory);
			if (history != null) {
				Position currPos, histPos;
				synchronized (history) {
					currPos = history.getNewest();
					histPos = history.getPositionByTime(detectionData.getTimeStamp());
				}
				if (histPos != null) {
					detected = new Position(
						detected.getPosX() + currPos.getPosX() - histPos.getPosX(),
						detected.getPosY() + currPos.getPosY() - histPos.getPosY(),
						detected.getHeading() + currPos.getHeading() - histPos.getHeading());
				}
			}
		}

		// if we haven't identified ourselves in the detected data yet,
		// we'll determine our historical position, at time of detection,
		// to compare with the detected positions
		if (detected == null)
		{
			PositionQueue history = (PositionQueue)space.read(cfg.blockPosHistory);
			if (history != null) {
				Position histPos, currPos;
				synchronized (history) {
					histPos = history.getPositionByTime(detectionData.getTimeStamp());
					currPos = history.getNewest();
				}
				if (histPos != null) {
					// compare the historical position with the detected ones
					for (DetectionData.Robot robot: detectionData.getRobots()) {
						Position pos = robot.position;
						if (pos.distance(histPos) < cfg.moveToleranceDistance &&
							pos.headingDiff(histPos) < cfg.moveToleranceHeading)
						{
							// this is the right position, update it to present-time
							detected = new Position(
								pos.getPosX() + currPos.getPosX() - histPos.getPosX(),
								pos.getPosY() + currPos.getPosY() - histPos.getPosY(),
								pos.getHeading() + currPos.getHeading() - histPos.getHeading());

							// associate this robot with the position in the detection data object
							robot.identify(cfg.robotNum, cfg.blockActualPos);
							space.trigger("detectedRobots"); // we change the object in space
							break;
						}
					}
				}
			}
		}

		// if we didn't find our position in the detection data, we'll try to turn
		// the robot, and compare the detected positions to find the robot which turned
		// (but we do this only if we have exclusive connection to the robot)
		if (detected == null && cfg.exclusive)
		{
			// get the robot control object
			Robot robot = (Robot)space.read(cfg.blockComm);
			if (robot == null) {
				log(5, "not connected");
				return; // robot not connected, do nothing
			}

			// stop the robot, and disable other agents trying to move it
			space.delete(cfg.blockMovement, 20000, Float.POSITIVE_INFINITY);

			log(1, "identifying robot...");

			// now wait some time, until the position gets detected again
			try { Thread.sleep(3000); }
			catch (InterruptedException e) {
				log(2, "interrupted");
				return;
			}
			detectionData = (DetectionData)space.read(cfg.blockDetectionData);
			if (detectionData == null) {
				log(0, "no detection data before turn");
				return;
			}

			// turn the robot 90 degrees to the left
			log(1, "rt 90");
			synchronized (robot)
			{
				if (!robot.connected()) {
					log(1, "robot disconnected");
					return;
				}
				robot.rt((short)(TurtleConstants.TURN360_STEPS / 4));
				for (int x = 1;; ++x) {
					if (robot.isdone() == 1) break;
					if (x % 10 == 0 && robot.isalive() != 1) {
						log(1, "robot disconnected");
						return;
					}
					try { Thread.sleep(100); }
					catch (InterruptedException e) {
						log(2, "interrupted");
						return;
					}
				}
			}

			// now wait some time, until the new position gets detected
			try { Thread.sleep(3000); }
			catch (InterruptedException e) {
				log(2, "interrupted");
				return;
			}
			DetectionData newDetectionData = (DetectionData)space.read(cfg.blockDetectionData);
			if (newDetectionData == null) {
				log(0, "no detection data after turn");
				return;
			}

			// find a robot in both detected sets, which didn't move and just turned
			for (DetectionData.Robot r1: detectionData.getRobots()) {
				Position p1 = r1.position;
				for (DetectionData.Robot r2: newDetectionData.getRobots()) {
					Position p2 = r2.position;
					if (p1.distance(p2) < cfg.turnToleranceDistance &&
						Math.abs(p2.headingDiff(p1) - Math.PI/2) < cfg.turnToleranceHeading)
					{
						detected = p2;
						r2.identify(cfg.robotNum, cfg.blockActualPos); // associate detected position with robot
						space.trigger(cfg.blockDetectionData); // we change the object in space
						break;
					}
				}
				if (detected != null) break;
			}
			if (detected == null) {
				log(1, "identification failed");
				return;
			} else {
				log(1, "identification successful");

				// reenable the movement agent (by deleting movement block with minimal validity)
				space.delete(cfg.blockMovement, 0, Float.POSITIVE_INFINITY);
			}
		}

		// if we identified the robot successfully...
		if (detected != null) {
			// update the detected position in space
			space.write(cfg.blockDetectedPos, detected);
			updateRobotActualPosition(cfg.blockActualPos, cfg.blockPosHistory, detected, true);
		}
	}

	/**
	 *  Update actual position of a robot in space.
	 *  This method is accessible for other agents as well (notably OdometryAgent).
	 */
	static void updateRobotActualPosition(String blockActualPos, String blockPosHistory, Position position) {
		updateRobotActualPosition(blockActualPos, blockPosHistory, position, false);
	}
	private static void updateRobotActualPosition(String blockActualPos, String blockPosHistory,
	                                              Position position, boolean shiftHistory)
	{
		// update actual position
		Position oldPos = (Position)space.write(blockActualPos, position);

		// get (mutable) position history queue from the space
		PositionQueue history = (PositionQueue)space.read(blockPosHistory);
		if (history == null)
			// the history is not yet created, do it now
			space.write(blockPosHistory, history = new PositionQueue());

		// times
		long now = System.currentTimeMillis(),
		     old = now - POSITION_HISTORY_SIZE;

		// we are going to change a mutable object stored in space,
		// so we have to synchronize on it
		synchronized (history) {
			// remove too old items
			while (!history.isEmpty() && history.getOldest().getTimeStamp() < old)
				history.removeOldest();

			// shift history
			if (shiftHistory && oldPos != null) {
				double dx = position.getPosX() - oldPos.getPosX(),
				       dy = position.getPosY() - oldPos.getPosY(),
				       dh = position.headingDiff(oldPos);
				history.shiftBy(dx, dy, dh, now);
			}

			// insert current position
			history.add(new TimedPosition(position, now));
		}
	}
}
