package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;
import sk.robotics.rlcs.data.Movement;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.Collision;
import sk.robotics.rlcs.constants.TurtleConstants;

import java.awt.geom.Point2D;

/**
 *  Base class for agents which move with a robot towards destination.
 *  Agents deriving from this class just have to override abstract method
 *  <code>selectMove(actualPos, destination)</code>, which is called to
 *  determine movement command to send to the robot. Its parameters
 *  are set to values read from corresponding space blocks, and it returns
 *  an object of class Movement.
 * <p> Both ApproachAgent and AvoidAgent derive from this class.
 */
public abstract class MoveAgentBase extends Agent {

	public static @Hidden class Cfg extends Agent.Cfg {
		/** Tolerance for stopping at destination, in millimeters. */
		public Double destinationTolerance = 1.0;

		/** Tolerance for setting right direction to destination, in radians. */
		public Double directionTolerance = 0.1;

		/** Name of block with actual position of the robot.
		 *  The actual position is used, together with destination point,
		 *  to compute desired movement direction.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		/** Name of block with current destination of the robot.
		 *  The destination point is used to compute desired movement direction.
		 *  @block.type java.awt.geom.Point2D
		 *  @block.access read */
		public @Required @NoDefault String blockDestination = "destination/robot?";

		/** Name of block where to store the desired movement command.
		 *  @block.type sk.robotics.rlcs.data.Movement
		 *  @block.access read/write */
		public @Required @NoDefault String blockMovement = "movement/robot?";

		/** Priority of writing the desired movement command to space. */
		public Float priority = 0.0f;

		// default activation
		public Cfg() {
			timerPeriodMs = 10;
		}
	}
	protected final Cfg cfg;

	private Movement lastMove = Movement.NO_MOVE;

	protected MoveAgentBase(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	/**
	 *  Abstract method for determining next movement command to send to the robot.
	 *  It should return Movement object representing the desired command.<br>
	 *  Returning null will quit current sense-select-act iteration immediately,
	 *  without sending anything to the robot.<br>
	 *  Returning Movement.NO_MOVE will delete the movement block from space,
	 *  instead of storing it there. The effect on the robot is the same, it
	 *  will stop, but the difference is that another agents with lower priority
	 *  can reclaim control over the block.
	 */
	protected abstract Movement selectMove(Position pos, Point2D dest);

	/**
	 *  Sense-select-act reads relevant data from space (position, destination, collision),
	 *  lets derived class decide about movement command, and stores this command in space,
	 *  where RobotMovementAgent reads it and sends it to the robot.
	 */
	@Override protected void senseSelectAct() {
		// read actual position, and desired destination from space
		Point2D dest = (Point2D) space.read(cfg.blockDestination);
		Position pos = (Position) space.read(cfg.blockActualPos);
		if (pos == null || dest == null) {
			if (pos == null) log(5, "no position!");
			if (dest == null) log(5, "no destination!");
			return; // something is wrong, do nothing
		}

		// if we are at the right place already, stop
		double dist = pos.distance(dest);
		if (dist < cfg.destinationTolerance) {
			space.delete(cfg.blockMovement, 0, cfg.priority);
			log(5, "we are at the destination point!");
			return;
		}

		// let the derived class determine next movement
		Movement move = selectMove(pos, dest);
		if (move == null) return; // break signal

		// write desired movement command to space
		if (move.equals(Movement.NO_MOVE)) {
			// use delete instead of write in this case, because
			// we want minimal validity instead of infinite validity
			space.delete(cfg.blockMovement, 0, cfg.priority);
		} else {
			space.write(cfg.blockMovement, move, 0, cfg.priority);
		}

		// logging
		if (!move.equals(lastMove)) {
			log(2, "movement %s (was %s)", move, lastMove);
			lastMove = move;
		}
	}
}
