package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Movement;
import sk.robotics.rlcs.data.Collision;
import sk.robotics.robot.Robot;

/**
 *  Agent for sending movement commands to robots.
 *  There should be one instance of this agent for each robot controlled by RLCS.
 *  It reads desired movement from space, and sends corresponding command to the robot.
 *  This agent also checks for collisions, and stops the robot, if commend read from
 *  space would cause one.
 */
public class MovementAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Name of block with desired movement.
		 *  @block.type sk.robotics.data.Movement
		 *  @block.access read */
		public @Required @NoDefault String blockMovement = "movement/robot?";

		/** Name of block with current collision object.
		 *  The block should be filled by CollisionAgent.
		 *  @block.type sk.robotics.data.Collision
		 *  @block.access read */
		public @Required @NoDefault String blockCollision = "collision/robot?";

		/** Name of block with robot control object.
		 *  The object is used to send movement commands to the robot.
		 *  Any access to the object is synchronized (on the object itself).
		 *  @block.type sk.robotics.robot.Robot
		 *  @block.access read */
		public @Required @NoDefault String blockComm = "comm/robot?";

		/** Border to add to turtle radius when checking for collisions. */
		public Double collisionDistanceMm = 10.0;

		/** Time after a movement command is resent again to the robot, if the movement
		 *  in space doesn't change, to avoid timing out and stopping the robot. */
		public Long commandRepeatPeriodMs = 10000L;

		// default activation
		public Cfg() {
			timerPeriodMs = 10;
		}
	}
	protected final Cfg cfg;

	public MovementAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	// last movement command sent to the robot, and timestamp of the send
	private Movement lastSentMovement;
	private long lastSentTimestamp;

	@Override protected void senseSelectAct() {
		// get the robot object
		Robot robot = (Robot)space.read(cfg.blockComm);
		if (robot == null) {
			log(5, "no robot connection");
			return; // not connected, do nothing
		}

		// read desired movement from space
		Movement move = (Movement)space.read(cfg.blockMovement);

		// if desired movement is set, check for collision
		boolean collision = false;
		if (move != null) {
			Collision coll = (Collision)space.read(cfg.blockCollision, Collision.ALL);
			collision = coll.check(move, cfg.collisionDistanceMm).getCollisionDistance() < 0;
		}

		// if desired movement is not set, or would cause a collision, stop the robot
		if (move == null || collision) {
			if (lastSentMovement != null) {
				if (collision) {
					log(2, "collision, stopping robot");
				} else {
					log(2, "stopping robot");
				}
				synchronized (robot) {
					if (robot.connected()) {
						robot.fdspeed((short)0, (short)0);
					}
				}
				lastSentMovement = null;
			}
			return;
		}

		// if the movement didn't change, and was sent not too long ago, do nothing
		if (move.equals(lastSentMovement) && (System.currentTimeMillis() - lastSentTimestamp) < cfg.commandRepeatPeriodMs) {
			return;
		}

		// send movement command to the robot
		short leftSpeed = (short)move.getLeftSpeed(),
		      rightSpeed = (short)move.getRightSpeed();
		synchronized (robot) {
			if (!robot.connected()) {
				log(1, "robot disconnected");
				return;
			}
			switch (move.getDirection()) {
				case FD: robot.fdspeed(leftSpeed, rightSpeed); break;
				case BK: robot.bkspeed(leftSpeed, rightSpeed); break;
				case LT: robot.ltspeed(leftSpeed, rightSpeed); break;
				case RT: robot.rtspeed(leftSpeed, rightSpeed); break;
			}
		}
		log(2, "movement " + move);

		// update the last sent movement and timestamp
		lastSentMovement = move;
		lastSentTimestamp = System.currentTimeMillis();
	}
}
