package sk.robotics.rlcs.agents.robot;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Space;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.Collision;
import sk.robotics.robot.Robot;

/**
 *  Agent for estimating robots position from odometry.
 *  Cooperates with IdentifyAgent to always have actual position of the robot
 *  as accurate as possible. Never resets odometry of the robot control object,
 *  so it is still usable by other agents as well.
 * <p> Actual position is modified according to odometry only if no collision
 *  is detected, otherwise odometry is ignored.
 */
public class OdometryAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Global odometry flag.
		    Local odometry is used with exclusively connected robots,
		    and the global one with non-exclusively connected ones. */
		public Boolean globalOdometry = false;

		/** Name of block with the robot control object.
		 *  Used to query odometry from the robot.
		 *  Access to it is synchronized (on the object itself).
		 *  @block.type sk.robotics.robot.Robot
		 *  @block.access read */
		public @Required @NoDefault String blockComm = "comm/robot?";

		/** Name of block with last detected position of the robot.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access read */
		public @Required @NoDefault String blockDetectedPos = "detectedPos/robot?";

		/** Name of block with actual position of the robot.
		 *  @block.type sk.robotics.rlcs.data.Position
		 *  @block.access write */
		public @Required @NoDefault String blockActualPos = "actualPos/robot?";

		/** Name of block with history of actual positions of the robot.
		 *  @block.type sk.robotics.rlcs.data.PositionQueue
		 *  @block.access read/write */
		public @Required @NoDefault String blockPosHistory = "posHistory/robot?";

		/** Name of block with current Collision object.
		 *  Used to detect collisions.
		 *  @block.type sk.robotics.rlcs.data.Collision
		 *  @block.access read */
		public @Required @NoDefault String blockCollision = "collision/robot?";

		/** Name of block to store raw odometry string to.
		 *  @block.type String
		 *  @block.access write */
		public String blockRawOdometry;

		public Cfg() {
			// default activation
			timerPeriodMs = 10;
		}
	}
	private final Cfg cfg;

	// time of last odometry reset
	private long lastReset = Long.MIN_VALUE;

	// starting position (from time of last reset)
	private Position startPosLocal; // in our coordinates
	private Position startPosOdometry; // as returned by odometry

	public OdometryAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		// get the detected-position block from space
		Space.Block detectedPosBlock = space.getBlock(cfg.blockDetectedPos);

		// if no position was detected yet, do nothing
		if (detectedPosBlock == null) return;

		// get the robot control object
		Robot robot = (Robot)space.read(cfg.blockComm);
		if (robot == null) return; // not connected, do nothing

		// get robots relative position based on odometry
		String rawOdometry;
		synchronized (robot) {
			if (!robot.connected()) {
				log(1, "robot disconnected");
				return;
			}
			rawOdometry = cfg.globalOdometry ? robot.globalpos((short)0) : robot.pos((short)0);
		}
		Position odometry = new Position(rawOdometry);

		// store raw odometry string in space
		if (cfg.blockRawOdometry != null) {
			space.write(cfg.blockRawOdometry, rawOdometry);
		}

		// if there is new detected position, reset odometry
		// (reset here means not actually resetting the odometry,
		//  but instead just remembering new starting positions,
		//  both in our coordinates and as returned by odometry)
		if (detectedPosBlock.getValidFrom() > lastReset && detectedPosBlock.getValue() != null) {
			startPosLocal = (Position)detectedPosBlock.getValue();
			startPosOdometry = odometry;
			lastReset = detectedPosBlock.getValidFrom();
			return;
		}

		// do nothing if position has not changed
		if (odometry.equals(startPosOdometry)) {
			return;
		}

		// compute actual position, based on odometry
		double delta = startPosOdometry.distance(odometry);
		double angle = startPosLocal.getHeading() + startPosOdometry.directionTo(odometry) - startPosOdometry.getHeading();
		Position newPos = new Position(
			startPosLocal.getPosX() + delta * Math.sin(angle),
			startPosLocal.getPosY() + delta * Math.cos(angle),
			startPosLocal.getHeading() + odometry.getHeading() - startPosOdometry.getHeading());

		// check for collision (if there is a colision, we'll ignore the odometry)
		Collision collision = (Collision) space.read(cfg.blockCollision);
		if (collision != null) {
			Position oldPos = collision.getActualPos();
			if (oldPos == null) return; // probably Collision.ALL

			double dist = oldPos.distance(newPos);
			if (dist > 0.01) {
				// robot is moving (not turning only)
				double dir = oldPos.directionTo(newPos);
				if (collision.checkMove(dir, 0).getCollisionDistance() < dist) {
					log(4, "collision");
					return; // collision!
				}
			}
		}

		// update actual position in space
		IdentifyAgent.updateRobotActualPosition(cfg.blockActualPos, cfg.blockPosHistory, newPos);
	}
}
