package sk.robotics.rlcs.agents.sim;

import java.net.*;
import java.io.*;
import java.awt.image.*;
import javax.imageio.ImageIO;

import sk.robotics.rlcs.agentspace.Agent;

/**
 *  Agent for loading current calibrated image from ImgGen.
 *  ImgGen is a simulator client utility, which communicates with the
 *  virtuallab simulator, and generates image of the simulated arena.
 */
public class ImageLoaderImgGenAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Host name or IP address where ImgGen runs. Not set means localhost. */
		public String imgGenHost;

		/** TCP port on which ImgGen is listening. */
		public @Limited(min=1,max=65535) int imgGenPort = 35754;

		/** Set this to receive raw image data. See ImgGen for details about the raw format.
		 *  If receiveing raw image, and there already is a compatible image in the image
		 *  block, <b>that</b> image is reused, rather than creating a new instance.
		 *  If raw is not set, a new instance of image is created for each received frame. */
		public Boolean rawImage = false;

		/** Name of block where to store the loaded image.
		 *  @block.type java.awt.image.BufferedImage
		 *  @block.access write */
		public @Required @NoDefault String blockImage = "calibratedImage";

		public Cfg() {
			// default activation
			timerPeriodMs = 1000;
		}
	}
	private Cfg cfg;

	// connection to ImgGen
	private Socket sock;
	private InputStream input; // binary input
	private BufferedReader reader; // string input
	private PrintWriter writer; // string output

	// connection failure flag, preventing repetitive error logs
	private boolean connectionFailure = false;

	// timestamp of last received image
	private long lastTimeStamp = 0;

	// buffer for downloading the image
	private byte[] buffer;

	// image size
	private int width, height;

	public ImageLoaderImgGenAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		try {
			// try to connect if needed
			if (sock == null) {
				try {
					sock = new Socket(cfg.imgGenHost, cfg.imgGenPort);
					input = sock.getInputStream(); // no need for buffering here
					reader = new BufferedReader(new InputStreamReader(sock.getInputStream()));
					writer = new PrintWriter(new BufferedWriter(new OutputStreamWriter(sock.getOutputStream())), true);
					log(0, "connected to ImgGen");
				}
				catch (Exception e) {
					if (!connectionFailure) {
						log(0, "failed to connect to ImgGen");
						log(1, e);
						connectionFailure = true;
					}
					// after a connection failure, sleep a while, don't hammer the target
					sleep(3000);
					return;
				}

				// read image bounds
				writer.println("BOUNDS");
				String[] bounds = readLine(reader).split(":");
				if (bounds.length != 2) {
					throw new Exception("Bad reply to BOUNDS request!");
				}
				width = Integer.parseInt(bounds[0]);
				height = Integer.parseInt(bounds[1]);
			}

			// get timestamp of last change in the simulated arena
			writer.println("TIME");
			long timeStamp = Long.parseLong(readLine(reader));

			// check whether there were any changes since last load
			if (timeStamp == lastTimeStamp) {
				return; // no changes
			}

			// reset ImgGen to receive current image
			writer.println("RESET");
			if (!readLine(reader).equals("ok")) {
				throw new IOException("unexpected reply to RESET");
			}

			// read image data length
			writer.println("LENGTH");
			int length = Integer.parseInt(readLine(reader));

			// realloc buffer if needed
			if (buffer == null || buffer.length < length) {
				buffer = new byte[length];
			}

			// read the image data
			writer.println("DATA");
			int read = 0;
			while (read < length) {
				read += input.read(buffer, read, length - read);
			}

			// construct the image
			BufferedImage image;
			if (cfg.rawImage) {
				// check that the image data is of correct size
				if (length != width * height * 3) {
					throw new Exception("Wrong image data length!");
				}

				// load image from the destination block
				image = (BufferedImage) space.read(cfg.blockImage);

				// if there was no image, or is not compatible, create a new one
				if (image == null || image.getType() != BufferedImage.TYPE_INT_RGB ||
				    image.getWidth() != width || image.getHeight() != height)
				{
					image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
				}

				// synchronize on the image while changing it
				synchronized (image) {
					WritableRaster raster = image.getRaster();
					SampleModel sampleModel = raster.getSampleModel();
					DataBuffer dataBuffer = raster.getDataBuffer();
					int[] pixel = new int[1];
					for (int x = 0, p = 0; x < width; ++x) {
						for (int y = 0; y < height; ++y) {
							pixel[0] = (buffer[p++] & 0xFF) << 16;
							pixel[0] |= (buffer[p++] & 0xFF) << 8;
							pixel[0] |= (buffer[p++] & 0xFF);
							sampleModel.setDataElements(x, y, pixel, dataBuffer);
						}
					}
				}
			} else {
				// not raw, read the image, create a new image instance
				image = ImageIO.read(new ByteArrayInputStream(buffer, 0, length));
				if (image == null) {
					throw new Exception("image is in unknown format!");
				}
			}

			// remember last timestamp
			lastTimeStamp = timeStamp;

			// store the image in space
			space.write(cfg.blockImage, image);
			log(5, "image written to block '" + cfg.blockImage + "'");

			// ensure at least some sleep
			sleep(100);
		}
		catch (Exception e) {
			if (e instanceof EOFException || e.getMessage().indexOf("connection abort") >= 0) {
				log(0, "connection closed");
			} else {
				log(0, e, "Error, closing socket.");
			}

			// close the socket, we'll try to connect again
			try { sock.close(); }
			catch (IOException ee) {}
			sock = null;

			// sleep a while, don't hammer the target
			sleep(3000);
		}
	}

	// read a line from reader, throwing IOException on EOF
	// (this is just a convenience wrapper, the underlying readLine
	//  method returns null, instead of throwing an exception)
	private String readLine(BufferedReader reader) throws IOException {
		String line = reader.readLine();
		if (line == null) {
			throw new EOFException("EOF");
		}
		return line;
	}
}
