package sk.robotics.rlcs.agents.sim;

import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;
import javax.imageio.ImageIO;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.constants.VlabConstants;
import sk.robotics.simulator.Animator;

/**
 *  Agent for drawing current calibrated image for simulated arena.
 *  Uses the simulator's Animator class.
 */
public class ImageLoaderSimAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Host name or IP address where ImgGen runs. Not set means localhost. */
		public String simdHost;

		/** TCP port on which ImgGen is listening. */
		public @Limited(min=1,max=65535) int simdPort = 35753;

		/**
		 *  Width of the image to draw.
		 *  Either this or height will be decreased as needed
		 *  to match the simated desk aspect ratio.
		 */
		public Integer imageWidth = 640;

		/**
		 *  Height of the image to draw.
		 *  Either this or width will be decreased as needed
		 *  to match the simated desk aspect ratio.
		 */
		public Integer imageHeight = 480;

		/** Name of block where to store the image.
		 *  @block.type java.awt.image.BufferedImage
		 *  @block.access write */
		public @Required @NoDefault String blockImage = "calibratedImage";

		/** Name of block where to store collection of turtle positions.
		 *  @block.type java.util.Collection&lt;sk.robotics.rlcs.data.Position&gt;
		 *  @block.access write */
		public @Required @NoDefault String blockPositions = "simPositions";

		/** Period for second timer, in milliseconds.
		 *  On the standard timer, with id 0x01, the image is generated.
		 *  On this timer, with id 0x02, the positions are generated. */
		public Integer timer2PeriodMs = 10;

		// default activation
		public Cfg() {
			timerPeriodMs = 1000;
		}
	}
	private Cfg cfg;

	// the Animator object
	private Animator animator;

	// real image size (adjusted to keep desk aspect ratio)
	private int width, height;

	// was there a change on the simulated desk?
	private volatile boolean changeOnDesk = false;

	// logger for animator
	private class AnimatorLogger extends Animator.Logger {
		@Override public void log(Object msg) { ImageLoaderSimAgent.this.log(1, msg); }
		@Override public void log(Exception e) { ImageLoaderSimAgent.this.log(0, e); }
		@Override public void log(Exception e, Object msg) { ImageLoaderSimAgent.this.log(0, msg); }
	}

	public ImageLoaderSimAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;

		// register second timer (for positions generation)
		if (cfg.timer2PeriodMs != null && cfg.timer2PeriodMs > 0) {
			setPeriodicTrigger(cfg.timer2PeriodMs, 0x02);
		}

		// compute real image size (adjusted to keed desk aspect ratio)
		double ratio = cfg.imageWidth / (double)cfg.imageHeight;
		if (ratio > VlabConstants.ASPECT_RATIO) {
			height = cfg.imageHeight;
			width = (int)(height * VlabConstants.ASPECT_RATIO);
		} else if (ratio < VlabConstants.ASPECT_RATIO) {
			width = cfg.imageWidth;
			height = (int)(width / VlabConstants.ASPECT_RATIO);
		} else {
			width = cfg.imageWidth;
			height = cfg.imageHeight;
		}
	}

	@Override protected void startup() {
		animator = new Animator((cfg.simdHost == null ? "localhost" : cfg.simdHost) + ":" + cfg.simdPort,
		                        new Runnable() { public void run() { changeOnDesk = true; } },
		                        width, height, null, new AnimatorLogger());
		animator.start();
	}

	@Override protected void cleanup() {
		animator.destroy();
	}

	@Override protected void senseSelectAct(int triggeredIds) {
		// if there was no change on the desk, do nothing
		if (!changeOnDesk) return;

		// generate the image
		if ((triggeredIds & 0x01) != 0) {
			// load the target image from space
			BufferedImage image = (BufferedImage) space.read(cfg.blockImage);

			// if there was no image, or is not compatible, create a new one
			if (image == null || image.getType() != BufferedImage.TYPE_INT_RGB ||
			    image.getWidth() != width || image.getHeight() != height)
			{
				image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
			}

			// synchronize on the image while changing it
			synchronized (image) {
				// draw current situation on the desk
				animator.paint(image.createGraphics(), Animator.PAINT_ANTIALIAS);
			}

			// store the image in space
			space.write(cfg.blockImage, image);
		}

		// generate positions
		if ((triggeredIds & 0x02) != 0) {
			// get current positions
			double[][] turtles = animator.getTurtlePositions();

			// generate the collection
			Collection<Position> positions = new ArrayList<Position>(turtles.length);
			for (double[] pos: turtles) {
				positions.add(new Position(pos[0], pos[1], pos[2]));
			}

			// store it in space
			space.write(cfg.blockPositions, Collections.unmodifiableCollection(positions));
		}
	}
}

