package sk.robotics.rlcs.agents.sim;

import java.net.*;
import java.io.*;
import java.util.*;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.DetectionData;

/**
 *  Agent for loading list of detected objects from the simulator.
 *  For now, only robots are detected, as with the real lab.
 */
public class SimDetectionAgent extends Agent {

	public static class Cfg extends Agent.Cfg {
		/** Host name or IP address where the simulator runs. Not set means localhost. */
		public String simdHost;

		/** TCP port on which the simulator is listening. */
		public @Limited(min=1,max=65535) int simdPort = 35753;

		/** Name of block with data from image recognition.
		 *  Output of this agent, for use by other agents.
		 *  @block.type sk.robotics.rlcs.data.DetectionData
		 *  @block.access write */
		public @Required @NoDefault String blockDetectionData = "detectionData";

		public Cfg() {
			// default activation
			timerPeriodMs = 1000;
		}
	}
	private Cfg cfg;

	public SimDetectionAgent(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@Override protected void senseSelectAct() {
		try {
			// connect to simulator
			Socket sock;
			try { sock = new Socket(cfg.simdHost, cfg.simdPort); }
			catch (IOException e) { return; } // simd not running - no error

			// send position query
			OutputStream out = sock.getOutputStream();
			out.write("posi".getBytes());
			out.flush();
			sock.shutdownOutput();

			// the reply will come delayed, but it will contain positions from right now
			long timestamp = System.currentTimeMillis();

			// receive reply
			InputStream in = sock.getInputStream();
			StringBuilder reply = new StringBuilder(256);
			int b;
			while ((b = in.read()) != -1)
				reply.append((char)b);

			// parse the reply
			Collection<Position> positions = new ArrayList<Position>();
			for (String line: reply.toString().split("\n")) {
				String[] parts = line.split(";");
				if (parts.length != 3) continue;

				double x, y, h;
				try {
					x = Double.parseDouble(parts[0]);
					y = Double.parseDouble(parts[1]);
					h = Double.parseDouble(parts[2]);
				} catch (Exception e) { continue; }

				positions.add(new Position(x, y, h));
			}

			// save positions in space
			space.write(cfg.blockDetectionData, new DetectionData(timestamp, positions, null));
		}
		catch (IOException e) {
			if (e.getMessage().equals("Connection reset")) {
				log(0, "connection reset");
			} else {
				log(0, e);
			}
		}
	}
}
