package sk.robotics.rlcs.agents.sim;

import java.util.*;
import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.data.Position;
import sk.robotics.rlcs.data.DetectionData;

/**
 *  Alternative agent to SimDetectionAgent.
 *  This one generates the detection data from ImageLoaderSimAgent output.
 *  Therefore, ImageLoaderSimAgent must run too, for this agent to work.
 *  The original SimDetectionAgent polls the simulator using TCP to receive
 *  a position snapshot. That creates new socket on each sense-select-act
 *  cycle iteration, which may be somewhat annoying.
 */
public class SimDetectionAgent2 extends Agent {

	public static class Cfg extends SimDetectionAgent.Cfg {
		/** Artificial delay. This agent won't return current data,
		 *  but rather older data to better simulate reality. */
		public @NoDefault Integer delayMs = 1000;

		/** Name of block with current set of positions. Output of ImageLoaderSimAgent.
		 *  @block.type java.util.Collection&lt;sk.robotics.rlcs.data.Position&gt;
		 *  @block.access read */
		public @Required @NoDefault String blockPositions = "simPositions";

		public Cfg() {
			// default activation
			timerPeriodMs = 2000;
		}
	}
	private Cfg cfg;

	public SimDetectionAgent2(Cfg cfg) {
		super(cfg);
		this.cfg = cfg;
	}

	@SuppressWarnings("unchecked")
	@Override protected void senseSelectAct() {
		// read current positions from space
		Collection<Position> positions = (Collection<Position>) space.read(cfg.blockPositions);
		if (positions == null) return;

		// make a defensive copy (it's a mutable object)
		synchronized (positions) {
			positions = new ArrayList<Position>(positions);
		}

		// simulate a delay, present in reality
		long timestamp = System.currentTimeMillis();
		if (cfg.delayMs != null) {
			sleep(cfg.delayMs);
		}

		// save detection data in space
		space.write(cfg.blockDetectionData, new DetectionData(timestamp, positions, null));
	}
}
