package sk.robotics.rlcs.agentspace;

import sk.robotics.rlcs.security.Trusted;

import java.util.*;
import java.text.*;
import java.io.*;
import java.lang.reflect.*;
import java.lang.annotation.*;
import java.security.*;

/**
 *  Base class for agents.
 *
 *  Steps to implement a new agent, launchable from XML configuration:
 * <ol>
 *  <li>Derive a class from Agent (this class).
 *  <li>Derive a config class from Agent.Cfg, and make it member class of the agent class.
 *  <li>In the agent class, define a constructor taking an object of the config class,
 *      and passing it to the constructor of the base class.
 *  <li>Override senseSelectAct() or senseSelectAct(int) to do something when the agent
 *      is triggered. The version with int parameters can distinguish between different
 *      trigger events.
 *  <li>Agent activation is configured in the config object using timerPeriodMs and
 *      triggers fields. You can define default values for these field in construtor
 *      of the config class.
 * </ol>
 *
 *  To launch the agent from XML configuration, simply include it's config object
 *  in the configuration.
 */
public abstract class Agent implements Runnable {
	/**
	 *  Annotation used for hiding agents in web interface.
	 *  If agents config class is annotated with this annotation,
	 *  it is left out from the add-agent dialog in web interface.
	 */
	@Retention(RetentionPolicy.RUNTIME)
	@Target(ElementType.TYPE)
	public static @interface Hidden {}

	/**
	 *  Config base class.
	 *  Declaring it as abstract makes the Web Editor deny creating
	 *  agents of this base class (but allows derived agents of course).
	 */
	@Hidden @com.microstepmis.xplatform.AsElement
	public static class Cfg {
		/**
		 *  Annotation type for required fields. The validate() method automatically
		 *  checks fields annotated with this annotation whether they are filled.
		 */
		@Retention(RetentionPolicy.RUNTIME)
		@Target(ElementType.FIELD)
		public static @interface Required {}

		/**
		 *  Annotation for limiting range of allowed values for number fields.
		 *  The validate() method checks fields annotated with this annotation for the limits.
		 */
		@Retention(RetentionPolicy.RUNTIME)
		@Target(ElementType.FIELD)
		public static @interface Limited {
			double min();
			double max();
		}

		/**
		 *  Default values of Fields annotated with this annotation are treaten in
		 *  a special way. Normally, in the web interface, field value are nulled
		 *  after creating a new config object, and then later the default values
		 *  are used if the user does not set the field value explicitely. Fields
		 *  with this annotation, however, are skipped in both cases. They are not
		 *  nulled on creation, and their default values are ignored if user clears
		 *  the field value.
		 */
		@Retention(RetentionPolicy.RUNTIME)
		@Target(ElementType.FIELD)
		public static @interface NoDefault {}

		/**
		 *  Is the agent enabled?.
		 *  Setting this to false allows to leave an agent in configuration,
		 *  but disable it temporarily, and not launch it.
		 */
		public Boolean enabled = true;

		/**
		 *  Optional agent identification string. Shown in log, eases debugging.
		 */
		public String id;

		/**
		 *  Logging level. Higher level causes more verbose logging.
		 */
		public Integer logLevel = 0;

		/**
		 *  Automatic trigger period, in milliseconds.
		 *  If not set, the agent won't get periodically triggering. The automatic trigger
		 *  has always ID 0x01 (OR-mask of triggered IDs is passed to senseSelectAct(int)).
		 */
		public Integer timerPeriodMs;

		/**
		 *  Set of blocks to register triggers for.
		 *  Delimited by whitespaces, commas, or semicolons (or combinations).
		 *  The trigger has by default ID 0x01.If custom trigger ID is needed,
		 *  it can be done by using "blockName=triggerId" as block name (without
		 *  quotes), where triggerId is an integer (may be hex prefixed with 0x,
		 *  or octal prefixed with 0). Block names can also be specified as $xxx,
		 *  in which case they will expand to value of field xxx of the config object.
		 */
		public String triggers;

		/**
		 *  Config validation. This base versions checks that fields declared as required
		 *  (through the Required annotation) are filled. Derived agent config classes may
		 *  override this to provide custom validation function for user defined configuration.
		 *  It can be as well used for setting default values for uninitialized fields,
		 *  or any other processing. To veto launching the agent, it can throw an exception.
		 */
		public void validate() {
			// we want to check field annotations here, but that requires special priviliges,
			// which the caller generally does not have, so do it in a trusted thread
			validator.validate(this);
		}

		// private validation implementation, called by the validator, see below
		private void validateImpl() {
			try {
				for (Field field: getClass().getFields()) {
					// check required fields, whether they are filled
					if (field.isAnnotationPresent(Required.class)) {
						if (field.get(this) == null) {
							throw new IllegalArgumentException(field.getName() + " must be filled!");
						}
					}
					// check limited fields, whether they are in range
					if (field.isAnnotationPresent(Limited.class)) {
						Limited limits = field.getAnnotation(Limited.class);
						Number value = (Number) field.get(this);
						if (value != null && (value.doubleValue() < limits.min() || value.doubleValue() > limits.max())) {
							throw new IllegalArgumentException(field.getName() + " out of bounds!");
						}
					}
				}
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}

		// trusted thread for config validation
		public static class Validator extends Trusted {
			// private constructor, to prevent abusing this class
			private Validator() {
				super("AgentCfgValidator", permissions);
			}

			// permissions granted to allow config validation
			private static Permissions permissions = new Permissions();
			static {
				permissions.add(new RuntimePermission("accessDeclaredMembers"));
				permissions.add(new ReflectPermission("suppressAccessChecks"));
				permissions.add(new PropertyPermission("sun.misc.ProxyGenerator.saveGeneratedFiles", "read"));
				permissions.add(new RuntimePermission("createClassLoader"));
				permissions.setReadOnly();
			}

			// validate a config
			private void validate(final Cfg cfg) {
				exec(new Runnable() { public void run() { cfg.validateImpl(); } });
			}
		}
		private static Validator validator = new Validator();

		/**
		 *  Create the agent, described by this config.
		 *  This method works if the config class is a member class of the agent class,
		 *  and the agent class has a constructor taking a config object. If this is not
		 *  the case, this method must be overriden.
		 */
		public Agent createAgent() {
			Class<?> configClass = getClass();
			Class<?> agentClass = configClass.getEnclosingClass();
			if (agentClass == null || !Agent.class.isAssignableFrom(agentClass)) {
				throw new UnsupportedOperationException("Config class "+configClass+" not a member of Agent class.");
			}
			Constructor<?> ctor;
			try {
				ctor = agentClass.getConstructor(configClass);
			} catch (NoSuchMethodException e) {
				throw new UnsupportedOperationException("No agent constructor taking config object (agent class: "+agentClass+").");
			}
			try {
				return (Agent) ctor.newInstance(this);
			} catch (Exception e) {
				throw new RuntimeException("Failed to create agent from config (agent class: "+agentClass+").", e);
			}
		}

		/**
		 *  Config comparison.
		 *  Compares all public fields. This allows Scheduler not to restart
		 *  agents whose config didn't change.
		 */
		public boolean equals(Object o) {
			Class clazz = o.getClass();
			if (!clazz.equals(getClass())) {
				return false; // class mismatch
			}
			try {
				for (Field field: clazz.getFields()) {
					Object val1 = field.get(o),
					       val2 = field.get(this);
					if (val1 != null && val2 != null) {
						if (!field.get(o).equals(field.get(this))) {
							return false; // field mismatch (different values)
						}
					} else if (val1 != null || val2 != null) {
						return false; // field mismatch (null - nonnull)
					}
				}
			} catch (IllegalAccessException e) {
				throw new RuntimeException(e);
			}
			return true; // configs are equal
		}

		/**
		 *  Returns names of fields to hide in the web editor.
		 *  Derived config classes may override this to add fields to the set.
		 *  @see Cfg#hideFields hideFields
		 */
		public Set<String> getHiddenFields() {
			return new HashSet<String>();
		}

		/**
		 *  Helper for adding hidden fields in a derived config.
		 *  To hide fields x, y, z, the derived getHiddenField method can be implemented as follows:<br>
		 *  <code>return hideFields(super.getHiddenFields(), "x", "y", "z")</code>
		 */
		protected Set<String> hideFields(Set<String> hidden, String... fields) {
			Collections.addAll(hidden, fields);
			return hidden;
		}
	}

	/** Agent threads are of this class, to ease permission checking. */
	public class AgentThread extends Thread {
		private AgentThread() {
			super(Agent.this, Agent.this.toString());
		}
		public Agent getAgent() {
			return Agent.this;
		}
	}

	/** Thread group class for agent-launched worker threads.
	 *  The security manager automatically adds all threads created by agents
	 *  to their worker thread group. */
	public class WorkerThreadGroup extends ThreadGroup {
		private WorkerThreadGroup() {
			super(Agent.this.toString() + "/WorkerThreads");
		}
		public Agent getAgent() {
			return Agent.this;
		}
	}

	// the agent thread
	protected AgentThread thread;

	// the worker thread group
	protected WorkerThreadGroup workerThreads;

	// config
	private Cfg cfg;

	// triggered ID's (or-ed together)
	private int triggeredIDs = 0;

	// trigger monitor (for synchronizing access to triggeredIDs)
	private Object triggerMonitor = new Object();

	// timer for periodic triggering (static, shared by all agents)
	private static Timer triggerTimer = new Timer("TriggerTimer", true);

	// agent exit flag (use exit() to set it)
	private boolean exitFlag = false;

	// timestamp of last sense-select-act
	private long lastSenseSelectAct;

	// logging level
	private int logLevel = 0;

	// time format for logging
	protected static final DateFormat logTimeFormat = new SimpleDateFormat("HH:mm:ss:SSS");

	/** The space, a convenience shortcut usable in derived classes. */
	protected static Space space = Space.getInstance();

	/**
	 *  Agent constructor.
	 *  To be called by derived classes' contructors.
	 */
	protected Agent(Cfg cfg) {
		this.cfg = cfg;
		this.logLevel = cfg.logLevel;

		// validate the config
		cfg.validate();

		// register timer (periodic triggers)
		if (cfg.timerPeriodMs != null && cfg.timerPeriodMs > 0) {
			setPeriodicTrigger(cfg.timerPeriodMs, 0x01);
			log(1, "registered timer (period: " + cfg.timerPeriodMs + ")");
		}

		// register space triggers
		if (cfg.triggers != null) {
			registerTriggers(cfg.triggers, 0x01);
		}
	}
	private void registerTriggers(String triggers, int defTriggerId) {
		// split to blocks
		String[] blocks = triggers.split("[\\s,;]+");
		for (String block: blocks) {
			if (block.isEmpty()) continue;

			// split to block name and trigger id
			String[] split = block.split("=");
			if (split.length > 2) {
				throw new IllegalArgumentException("Illegal char '=' in block name (agent class: "+getClass()+").");
			}

			// determine trigger id
			int triggerId = defTriggerId;
			if (split.length == 2) {
				try { triggerId = Integer.decode(split[1]); }
				catch (NumberFormatException e) {
					throw new IllegalArgumentException("Invalid trigger ID: " + split[1]);
				}
			}

			// expand block name, if it is avariable
			String blockName = split[0];
			if (blockName.startsWith("$")) {
				try { blockName = (String) cfg.getClass().getField(blockName.substring(1)).get(cfg); }
				catch (Throwable e) {
					throw new IllegalStateException(e);
				}
				registerTriggers(blockName, triggerId);
				continue;
			}

			// register trigger
			space.attachTrigger(blockName, this, triggerId);
			log(1, "registered trigger: " + blockName + " (id:" + triggerId + ")");
		}
	}

	/** Default constructor, for use by custom agents, which don't use standard configs. */
	protected Agent() {}

	/**
	 *  Set logging level.
	 */
	public final void setLogLevel(int level) {
		this.logLevel = level;
	}

	/**
	 *  Start the agent thread.
	 */
	public final void start() {
		// NB: creating AgentThread objects (and objects of any other Thread subclasses)
		// requires RuntimeException accessDeclaredMembers. Rather than granting this
		// permission to all agents, we'll create the Thread objects in a trusted thread.
		threadCreator.createThread(this);

		// now start the thread
		thread.start();
	}

	// trusted thread for creating AgentThread objects
	public static class ThreadCreator extends Trusted {
		// private constructor, to prevent abusing this class
		private ThreadCreator() {
			super("AgentThreadCreator", permissions);
		}

		// permissions granted to allow config validation
		private static Permissions permissions = new Permissions();
		static {
			permissions.add(new RuntimePermission("accessDeclaredMembers"));
			permissions.setReadOnly();
		}

		// create an AgentThread object for an agent
		private void createThread(final Agent agent) {
			exec(new Runnable() { public void run() { agent.thread = agent.new AgentThread(); } });
		}
	}
	private static ThreadCreator threadCreator = new ThreadCreator();

	/**
	 *  Kill the agent thread forcibly.
	 *  Use this only as fallback, if exit() does not work.
	 */
	@SuppressWarnings("deprecation") // we use deprecated Thread.stop() (but only as fallback)
	public final void kill() {
		thread.stop();
	}

	/**
	 *  Thread alive check.
	 */
	public final boolean isAlive() {
		return thread != null && thread.isAlive();
	}

	/**
	 *  Initialization method called from the agent's thread.
	 *
	 *  Can be overriden to perform special thread-specific initialization,
	 *  for which constructor is not appropriate, or do anything which should
	 *  be done only once the agent's thread is started.
	 */
	protected void startup() {}

	/**
	 *  Cleanup method, called just before the agent thread finishes.
	 *
	 *  Can be overriden to cleanup any allocated resources.
	 */
	protected void cleanup() {}

	/**
	 *  Register for periodic triggers.
	 *
	 *  @param periodMs Automatic trigger period.
	 *  @param triggerId The value to pass to trigger() function.
	 */
	protected void setPeriodicTrigger(long periodMs, final int triggerId) {
		triggerTimer.schedule(new TimerTask() { public void run() { trigger(triggerId); } },
		                      0, periodMs);
	}
	protected void setPeriodicTrigger(long periodMs) {
		setPeriodicTrigger(periodMs, 0x01);
	}

	/**
	 *  Sense-select-act, called upon trigger().
	 *
	 *  One of the two overloads should be overriden. The overload without arguments
	 *  is for convenience, can be used if derived agent doesn't need to distinguish
	 *  between different types of triggers.
	 *
	 *  @param triggerIDs Or-ed combination of triggered ID's (params for trigger()).
	 */
	protected void senseSelectAct(int triggerIDs) { senseSelectAct(); }
	protected void senseSelectAct() {}

	/**
	 *  Trigger (wake up) the agent.
	 *
	 *  @param triggerId Trigger ID - will be or-ed with other ID's to pass to senseSelectAct().
	 */
	public final void trigger(int triggerId) {
		synchronized (triggerMonitor) {
			triggeredIDs |= triggerId;
			triggerMonitor.notifyAll();
		}
	}
	public final void trigger() { trigger(0xFFFFFFFF); }

	/**
	 *  Quit the agent.
	 */
	public final void exit() {
		exitFlag = true;
		trigger();
		thread.interrupt(); // if the thread is waiting for a blocking call to finish...
	}

	/**
	 *  Get time passed since last sense-select-act.
	 */
	public final long getIdleTime() {
		return System.currentTimeMillis() - lastSenseSelectAct;
	}

	/**
	 *  Get agent config.
	 */
	public final Cfg getConfig() {
		return cfg;
	}

	/**
	 *  Main loop. Call start() to launch it.
	 */
	public final void run() {
		startup();

		for (;;) {
			// check for exit flag
			if (exitFlag) break;

			// wait for trigger
			int ids = 0;
			synchronized(triggerMonitor) {
				if (triggeredIDs == 0) {
					try { triggerMonitor.wait(); }
					catch (InterruptedException e) {}
				}
				ids = triggeredIDs;
				triggeredIDs = 0;
			}

			// check for exit flag again
			if (exitFlag) break;

			// sense-select-act
			try {
				lastSenseSelectAct = System.currentTimeMillis();
				senseSelectAct(ids);
			}
			catch (SecurityException e) {
				// on security exception, quit the agent
				log(0, e, "security exception in senseSelectAct()");
				break;
			}
			catch (Exception e) {
				// just log other exceptions, and continue in execution
				// (but sleep some time to avoid flooding log)
				log(0, e, "exception in senseSelectAct()");
				sleep(5000);
			}
		}

		cleanup();
	}

	/**
	 *  A sleep convenience helper.
	 *  Removes the necessity to catch the exception thrown from Thread.sleep().
	 *  But of course, the sleep is then not guaranteed to last for given time.
	 */
	public static void sleep(long millis) {
		try { Thread.sleep(millis); }
		catch (InterruptedException e) {}
	}

	/**
	 *  Access to the worker threads group.
	 *  This method is public, so that SecurityManager has access to it,
	 *  but only the parent agent thread is allowed to call it.
	 */
	public final WorkerThreadGroup getWorkerThreadGroup() {
		if (Thread.currentThread() != thread) {
			throw new SecurityException("Access to agent worker threads from outside denied!");
		}
		if (workerThreads == null) {
			workerThreads = new WorkerThreadGroup();
		}
		return workerThreads;
	}

	/**
	 *  Debug string.
	 */
	public String toString() {
		String cl = getClass().getName();
		return (cfg != null && cfg.id != null && cfg.id.length() > 0) ? cl + "/" + cfg.id : cl;
	}

	/**
	 *  Logging.
	 */
	protected void log(int level, Object msg) {
		log(level, "%s", msg);
	}
	protected void log(int level, String fmt, Object... params) {
		if (level > logLevel) return;
		synchronized (System.err) {
			logHeader(System.err);
			System.err.printf(" - " + fmt + "\n", params);
		}
	}
	protected void log(int level, Throwable e) {
		log(level, e, "Exception");
	}
	protected void log(int level, Throwable e, Object msg) {
		log(level, e, "%s", msg);
	}
	protected void log(int level, Throwable e, String fmt, Object... params) {
		if (level > logLevel) return;
		synchronized (System.err) {
			logHeader(System.err);
			System.err.printf(": " + fmt + "\n", params);
			e.printStackTrace(System.err);
		}
	}
	private void logHeader(PrintStream out) {
		out.print(logTimeFormat.format(new Date(System.currentTimeMillis())));
		out.print(" - ");
		out.print(getClass().getSimpleName());
		out.print(" [");
		if (cfg != null && cfg.id != null && cfg.id.length() > 0) {
			out.print(cfg.id);
		} else {
			out.print("0x");
			out.print(Integer.toHexString(hashCode()));
		}
		out.print("]");
	}
}
