package sk.robotics.rlcs.agentspace;

import sk.robotics.rlcs.Main;
import sk.robotics.rlcs.security.SecurityManager;
import sk.robotics.rlcs.security.Restricted;

import java.util.*;

/**
 *  Agent scheduler. While beeing an agent itself, manages other running agents.
 */
public class Scheduler extends Agent
{
	/**
	 *  Configuration - set of currently running agents.
	 *  This is not a standard agent config class. Standard agent config objects
	 *  are only used to create and start the agents. This one, instead, is passed
	 *  to Scheduler at runtime to start given agents. Also, current config object
	 *  can be queried from Scheduler.
	 */
	public static class Cfg {
		public static class AgentGroup {
			public String name = "group";
			public Agent.Cfg[] agents = new Agent.Cfg[0];
		}
		public AgentGroup[] groups = new AgentGroup[0];
	}

	/**
	 *  Helper for executing agent object creation in restricted thread.
	 *  Just a wrapper for Restricted thread class.
	 */
	private static class RestrictedAgentCreator implements Runnable {
		private RestrictedAgentCreator(Agent.Cfg cfg) {
			this.cfg = cfg;
		}
		private static Agent exec(Agent.Cfg cfg) {
			// get agent class, corresponding to the config
			Class agentClass = cfg.getClass().getEnclosingClass();
			if (agentClass == null || !Agent.class.isAssignableFrom(agentClass)) {
				throw new RuntimeException("Not an agent config class: " + cfg.getClass());
			}
			RestrictedAgentCreator creator = new RestrictedAgentCreator(cfg);
			new Restricted(creator, "RestrictedAgentCreator", agentClass).exec();
			return creator.agent;
		}
		public void run() {
			agent = cfg.createAgent();
		}
		private Agent.Cfg cfg; // input
		private Agent agent; // output
	}

	// current config
	private Cfg cfg = new Cfg();

	// running agents
	private Collection<Agent> agents = new LinkedList<Agent>();

	// running agents, keyed by config class (useful for comparing configs on restart)
	private Map<Class, Collection<Agent>> agentsByCfgClass = new HashMap<Class, Collection<Agent>>();

	// private ctor - this is a singleton class
	private Scheduler() {
		// register timer
		setPeriodicTrigger(500);
		setLogLevel(1);
	}

	// the singleton instance
	private static Scheduler instance = new Scheduler();

	protected synchronized void senseSelectAct() {
		// check all running agents, whether they did not die
		for (Iterator<Agent> it = agents.iterator(); it.hasNext();) {
			Agent agent = it.next();
			if (!agent.isAlive()) {
				log(1, "agent " + agent + " died!");
				it.remove();
			}
		}
	}

	/**
	 *  Get the singleton instance.
	 */
	public static Scheduler getInstance() {
		return instance;
	}

	/**
	 *  Quit all running agents.
	 *  Set of exceptions (agents not to quit) can be passed.
	 */
	public void quitAll() {
		quitAll(null);
	}
	private void quitAll(Set<Agent> except) {
		// try to quit all agents peacefully
		for (Agent a: agents) {
			if (except == null || !except.contains(a)) {
				log(1, "stopping agent: " + a);
				a.exit();
			}
		}

		// give the agents some time to quit...
		sleep(1000);

		// kill all remaining agents (they are probably stuck if still alive)
		for (Agent a: agents) {
			if (except == null || !except.contains(a)) {
				if (a.isAlive()) {
					log(1, "forcefully killing agent: " + a);
					a.kill();
				}
			}
		}

		// update collection of currently running agents 
		if (except != null) {
			agents = except;
		} else {
			agents.clear();
		}
	}

	/**
	 *  Quit all current agents, and start agents specified in given config object.
	 *  This method may only be called from the main thread.
	 */
	public synchronized void restart(Cfg config) {
		// only callable from the main thread, and from client connection handler threads
		Thread thread = Thread.currentThread();
		if (thread != Main.getMainThread() && !thread.getClass().equals(Main.ClientHandler.class)) {
			log(0, "Scheduler.restart() denied for thread " + thread);
			throw new SecurityException("Scheduler.restart() called from bad thread!");
		}

		// we'll fill this with currently running agents, which do not have
		// to be killed, because there is same agent in the new configuration
		Set<Agent> agentsToReuse = new HashSet<Agent>();
		// and this with those agent configs from the new configuration, which
		// are disabled, or for which there is an already running same agent
		Set<Agent.Cfg> configsToSkip = new HashSet<Agent.Cfg>();

		// compare the configs, and try to find agents whose
		// config did not change, so they do not have to be restarted
		for (Cfg.AgentGroup group: config.groups) {
			for (Agent.Cfg agentCfg: group.agents) {
				// if this agent is disabled, skip it
				if (!agentCfg.enabled) {
					configsToSkip.add(agentCfg);
					continue;
				}

				// for each agent in the new config, we'll check all currently running
				// agents of the same class, whether there already is one with same config
				Collection<Agent> runningAgents = agentsByCfgClass.get(agentCfg.getClass());
				if (runningAgents != null) {
					for (Agent agent: runningAgents) {
						if (agent.getConfig().equals(agentCfg)) {
							log(2, "reusing agent: " + agent);
							agentsToReuse.add(agent);
							configsToSkip.add(agentCfg);
						}
					}
				}
			}
		}

		// quit current agents
		quitAll(agentsToReuse);

		// create agent objects for all the new agents (in temporary collection first)
		Collection<Agent> agents = new ArrayList<Agent>();
		for (Scheduler.Cfg.AgentGroup group: config.groups) {
			for (Agent.Cfg agentCfg: group.agents) {
				if (!configsToSkip.contains(agentCfg)) {
					// create the agent, but in restricted thread
					agents.add(RestrictedAgentCreator.exec(agentCfg));
				}
			}
		}

		// update current config
		this.cfg = config;

		// add all the new agents into the list of running agents, and launch them
		this.agents.addAll(agents);
		for (Agent agent: agents) {
			log(1, "starting agent: " + agent);
			agent.start();
		}

		// update running agents map
		agentsByCfgClass.clear();
		for (Agent agent: this.agents) {
			// get agent config class
			Agent.Cfg cfg = agent.getConfig();
			if (cfg == null)  {
				log(2, "custom agent without config, agent class " + agent);
				continue;
			}
			Class clazz = cfg.getClass();

			// add agent to the map
			Collection<Agent> agentsOfClass = agentsByCfgClass.get(clazz);
			if (agentsOfClass == null) {
				agentsByCfgClass.put(clazz, agentsOfClass = new ArrayList<Agent>());
			}
			agentsOfClass.add(agent);
		}
	}

	/**
	 *  Launches new agent.
	 *  The new agent won't get added into the current config object,
	 *  but only into the list of running agents. This means that this
	 *  agent won't be visible from outside (through getCurrentConfig).
	 */
	public synchronized void launch(Agent.Cfg cfg) {
		// create instance of the new agent
		Agent agent;
		if (SecurityManager.getAgentForThread(Thread.currentThread()) != null) {
			// this method is also callable by agents, in which case we'll create
			// the new agent directly, in the security context of the calling agent
			agent = cfg.createAgent();
		} else {
			// if not called from an agent, create the the new agent object in a restricted thread
			agent = RestrictedAgentCreator.exec(cfg);
		}

		// add it to list of running agents
		agents.add(agent);

		// start the new agent
		agent.start();
	}

	/**
	 *  @return Current configuration object
	 */
	public synchronized Cfg getCurrentConfig() {
		return cfg;
	}
}
