package sk.robotics.rlcs.agentspace;

import java.util.Collection;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;
import java.util.PriorityQueue;
import java.util.Comparator;

import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  The space.
 *
 *  Stores just references to objects, so it is a very good idea
 *  to try to store only immutable objects in the space. If mutable
 *  objects are stored, great care must be taken to avoid surprises.
 *
 *  Space has its own thread, which manages block expiration,
 *  and takes care of expiration triggers.
 */
public class Space {
	/** A block in space. */
	public static class Block implements CustomXmlMappable {
		// data
		private String name;
		private Object value;
		private long validFrom, validTo;
		private float priority;
		private Map<Agent, Integer> triggers; // Agent.trigger(Integer) will be called on change
		
		// ctors & mutators
		private Block(String name) {
			this.name = name;
			reset();
		}
		private Block(String name, Object value, long validFrom, long validTo, float priority) {
			this.name = name;
			set(value, validFrom, validTo, priority);
		}
		private void set(Object value, long validFrom, long validTo, float priority) {
			this.value     = value;
			this.validFrom = validFrom;
			this.validTo   = validTo;
			this.priority  = priority;
		}
		private void reset() {
			set(null, Long.MAX_VALUE, Long.MIN_VALUE, Float.MIN_VALUE);
		}
		
		// accessors
		public Object getValue() { return value; }
		public long getValidFrom() { return validFrom; }
		public long getValidTo() { return validTo; }
		public float getPriority() { return priority; }
		
		// fire all triggers (public space.trigger() calls this)
		private void trigger() {
			if (triggers != null)
				for (Map.Entry<Agent, Integer> t: triggers.entrySet())
					t.getKey().trigger(t.getValue());
		}
		
		// expiration comparator
		private static class ExpirationComparator implements Comparator<Block> {
			public int compare(Block b1, Block b2) {
				return b1.validTo < b2.validTo ? -1 : (b1.validTo == b2.validTo ? 0 : 1);
			}
		}

		// mapping to XML
		public class XmlMappable {
			public String name = Block.this.name;
			public Object value = Block.this.value;
			public long validFrom = Block.this.validFrom;
			public long validTo = Block.this.validTo;
			public float priority = Block.this.priority;
		}
		@Override public XmlMappable getXmlMappable() {
			XmlMappable m = new XmlMappable();
			if (value instanceof CustomXmlMappable) {
				m.value = ((CustomXmlMappable)value).getXmlMappable();
			}
			return m;
		}
	}
	
	// the singleton space object
	private static Space singleton = new Space();
	
	// the blocks in space
	private Map<String, Block> blocks = new HashMap<String, Block>();
	
	// queue of block expiration times
	private PriorityQueue<Block> expirationQueue = new PriorityQueue<Block>(16, new Block.ExpirationComparator());
	
	// private constructor - this is a singleton class
	private Space() {
		// start the expiration thread
		new Thread("SpaceExpirationThread") {
			public void run() {
				expirationThread();
			}
		}.start();
	}
	
	/**
	 *  Access to the space.
	 *
	 *  @return The singleton space object.
	 */
	public static Space getInstance() {
		return singleton;
	}
	
	/**
	 *  Read an object from the space.
	 *
	 *  @param blockName The block to read from space.
	 *  @param def Object to return, when there is no such block in space.
	 */
	public synchronized Object read(String blockName) {
		Block block = blocks.get(blockName);
		return block == null ? null : block.value;
	}
	public Object read(String blockName, Object def) {
		Object obj = read(blockName);
		return obj == null ? def : obj;
	}
	
	/**
	 *  Get a Block object (not only the stored value) from the space.
	 *  Useful to query block's meta data (validity, priority).
	 *
	 *  @param blockName The block to get.
	 *  @return A <b>copy</b> (except value, which is referenced) of the actual Block object.
	 *          For expired blocks, null is returned.
	 */
	public synchronized Block getBlock(String blockName) {
		Block b = blocks.get(blockName);
		return (b == null) ? null : new Block(blockName, b.value, b.validFrom, b.validTo, b.priority);
	}
	
	/**
	 *  Write an object into the space.
	 *
	 *  @param blockName The block in space to write to.
	 *  @param value The value to write to the block.
	 *  @param validFor Time period for which the data is valid (from now). Zero or less means infinite validity.
	 *  @param priority The write priority - existing values with greater priority won't get overwritten.
	 *  @return Object previously stored in given block.
	 */
	public synchronized Object write(String blockName, Object value, long validFrom, long validTo, float priority) {
		// get block and current value in it
		Block block = blocks.get(blockName);
		Object oldValue = (block == null) ? null : block.value;
		
		// check priority
		if (block != null && block.priority - priority > 0.00001) {
			return oldValue; // write denied, no change
		}
		
		// remove the block from the expiration queue (if it is there), before we change
		// its validity, because that would change ordering of items in the queue
		if (block != null) {
			expirationQueue.remove(block);
		}
		
		// update the block
		if (block == null) {
			blocks.put(blockName, block = new Block(blockName, value, validFrom, validTo, priority));
		} else {
			block.set(value, validFrom, validTo, priority);
		}
		
		// update the expiration queue
		if (validTo != Long.MAX_VALUE) {
			expirationQueue.add(block);
			notifyAll();
		}
		
		// fire triggers
		block.trigger();
		
		return oldValue;
	}
	public Object write(String blockName, Object value, long validFor, float priority) {
		long now = System.currentTimeMillis(),
		     validTo = validFor > 0 ? now + validFor : Long.MAX_VALUE;
		return write(blockName, value, now, validTo, priority);
	}
	public Object write(String blockName, Object value, long validFor) {
		return write(blockName, value, validFor, 0);
	}
	public Object write(String blockName, Object value) {
		return write(blockName, value, 0, 0);
	}
	
	/**
	 *  Delete a block from the space. This in fact just calls write()
	 *  to store null in the block, instead of completely removing the
	 *  block, to allow rejecting future lower-priority writes.
	 * <p>
	 *  A special case is delete() call with validFor set to zero.
	 *  In this case, the zero does not mean infinite validity, but instead
	 *  just deleting the block, and expring, so that any subsequent attempt
	 *  to write into that block will succeed.
	 *
	 *  @param blockName The block to delete.
	 *  @param validFor How long lower-priority writes will be ignored.
	 *  @param priority The write priority.
	 *  @return Object previously stored in given block.
	 */
	public synchronized Object delete(String blockName, long validFor, float priority) {
		Object oldValue = write(blockName, null, validFor, priority);
		
		// if validFor is zero, remove the block immediately from space
		if (validFor == 0) {
			removeBlock(blocks.get(blockName));
		}

		// return the value previously stored in the block
		return oldValue;
	}
	public Object delete(String blockName) {
		return delete(blockName, 0, 0);
	}
	
	/**
	 *  Add an agent trigger to a block.
	 *  When the block changes (on write(), delete(), or expiration), the agent gets trigerred.
	 *
	 *  @param blockName The block to attach the trigger to.
	 *  @param agent The agent to trigger, when the block changes.
	 */
	public synchronized void attachTrigger(String blockName, Agent agent, int triggerId) {
		Block block = blocks.get(blockName);
		if (block == null) {
			blocks.put(blockName, block = new Block(blockName));
		}
		if (block.triggers == null) {
			block.triggers = new HashMap<Agent, Integer>();
		}
		block.triggers.put(agent, new Integer(triggerId));
	}
	
	/**
	 *  Detach an agent trigger from all blocks.
	 *  Useful to release agent references on agent exit.
	 */
	public synchronized void detachTriggers(Agent agent) {
		for (Block block: blocks.values()) {
			// try to remove the trigger from this block
			if (block.triggers != null && block.triggers.remove(agent) != null) {
				// we have removed the trigger from this block,
				// if there are no more triggers attached,
				// and the block is expired, remove it from space
				if (block.triggers.isEmpty() && block.validTo < System.currentTimeMillis()) {
					removeBlock(block);
				}
			}
		}
	}
	
	/**
	 *  Simulate a block change, firing associated triggers.
	 *  Useful if application changes a mutable object in space.
	 *
	 *  @param blockName The block whose change to simulate.
	 */
	public synchronized void trigger(String blockName) {
		Block block = blocks.get(blockName);
		if (block != null) {
			block.trigger();
		}
	}
	
	// expiration thread
	private synchronized void expirationThread() {
		for (;;) {
			// get first-to-expire block
			Block block = expirationQueue.peek();
			if (block == null) {
				// no block in the queue, sleep
				try { wait(); }
				catch (InterruptedException e) {}
				continue;
			}
			
			// sleep until the expiration time
			long now = System.currentTimeMillis();
			if (block.validTo > now) {
				try { wait(block.validTo - now); }
				catch (InterruptedException e) {}
				continue;
			}
			
			// the block has expired, remove it from space
			removeBlock(block);
		}
	}
	
	// remove a block from space
	// (must by called from synchronized method)
	private void removeBlock(Block block) {
		// remove the block completely only if it has no
		// triggers attached, otherwise just reset it
		if (block.triggers == null || block.triggers.isEmpty()) {
			blocks.remove(block.name);
		} else {
			block.reset();
		}
		
		// and remove it also from the expiration queue (if it is there)
		expirationQueue.remove(block);
	}
}
