package sk.robotics.rlcs.data;

import java.util.Collection;
import java.util.ArrayList;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import sk.robotics.rlcs.constants.TurtleConstants;
import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  Collision object for given robot holds lists of nearby walls and robots,
 *  and has a method to check for collision in arbitrary directions.
 *  Collision objects are immutable.
 */
public class Collision implements CustomXmlMappable
{
	// actual position of robot, for which this object computes collisions
	private Position actualPos;

	// nearby robots
	private Collection<Position> robots = new ArrayList<Position>();

	// nearby walls
	// (we store the walls in a more complicated way than just two points,
	//  to make collision computation easier, and more effective)
	public static class Wall {
		public double direction, distance; // orthogonal distance & direction from the robot
		public double limitLeft, limitRight; // length of the wall (limitRight - limitLeft)

		public Wall() {}
		public Wall(double dir, double dist, double left, double right) {
			direction = dir;
			distance = dist;
			limitLeft = left;
			limitRight = right;
		}
	}
	private Collection<Wall> walls = new ArrayList<Wall>();

	/** Special "collision in all directions" object. */
	public static final Collision ALL = new Collision() {
		@Override public CheckResult checkMove(double direction, double border) { return checkResultForcedCollision; }
		@Override public CheckResult checkMoveRel(double direction, double border) { return checkResultForcedCollision; }
		@Override public double checkTurn(int direction) { return 0; }
	};

	// default ctor - private, used by ALL only
	private Collision() {}

	// normal ctor
	public Collision(Position actualPos, Collection<Line2D> walls, Collection<Position> robots) {
		this.actualPos = actualPos;
		if (robots != null) {
			this.robots.addAll(robots);
		}
		if (walls != null) {
			for (Line2D line: walls) {
				this.walls.add(line2wall(line));
			}
		}
	}

	/** @return Actual position associated with this Collision object */
	public Position getActualPos() {
		return actualPos;
	}

	/**
	 *  Object to be returned by checkMove().
	 */
	public static class CheckResult {
		private double collisionDistance; // maximal movement without collision
		private double obstacleDirection = 0; // direction to obstacle, which would cause the collision
		private double obstacleDistance = 0; // closest distance to the collision causing obstacle

		private CheckResult(double collisionDistance) {
			this.collisionDistance = collisionDistance;
		}

		public double getCollisionDistance() { return collisionDistance; }
		public double getObstacleDirection() { return obstacleDirection; }
		public double getObstacleDistance() { return obstacleDistance; }
	}
	private static CheckResult checkResultForcedCollision = new CheckResult(0);

	/**
	 *  Determine distance of closest obstacle in given direction.
	 *  @param direction Direction to check for collision in, in radians.
	 *  @return Distance of closest obstacle in that direction, in millimeters.
	 *          If there is no obstacle, Double.MAX_VALUE is returned.
	 */
	public CheckResult checkMove(double direction, double border) {
		// we'll cycle through all walls and other robots,
		// and find the closest in given direction
		CheckResult result = new CheckResult(Double.MAX_VALUE);

		// check walls
		for (Wall w: walls) {
			// compute angle between the wall normal, and our direction
			double angle = Math.abs(Position.headingDiff(w.direction, direction));
			if (angle >= Math.PI/2) continue; // no collision with this wall

			// compute real distance to the wall
			double dst = (w.distance - border) / Math.cos(angle);

			// compute crash point in the wall (its distance from closest point in the wall)
			double pt = dst * Math.sin(angle);
			if (pt < w.limitLeft || pt > w.limitRight) continue; // no collision with this wall

			if (dst < result.collisionDistance) {
				result.collisionDistance = dst;
				result.obstacleDirection = w.direction;
				result.obstacleDistance = w.distance - border;
			}
		}

		// check other robots
		for (Position p: robots) {
			// compute distance and direction to the robot
			double dst = actualPos.distance(p);
			double dir = actualPos.directionTo(p);

			// compute angle between direction to the robot, and the direction to check collision for
			double angle = Math.abs(Position.headingDiff(dir, direction));

			// if the robot is not in the right half of the plane, ignore it
			if (angle >= Math.PI/2) {
				continue;
			}

			// minimal distance of turtle centres
			double minDist = 2 * TurtleConstants.RADIUS_MM + border;

			// collision distance (amount of movement in given direction to cause collision)
			double collDist;

			if (angle == 0) {
				// if the angle is zero, we are heading straight towards the robot
				collDist = dst - minDist;
			} else {
				// compute maximal colliding angle
				double angleLimit = Math.abs(Math.asin(minDist / dst));
				if (angle > angleLimit) continue; // no collision with this robot

				// compute the collision distance (from sine theorem)
				double q = minDist / Math.sin(angle);
				collDist = q * Math.sin(Math.asin(dst / q) - angle);
			}

			if (collDist < result.collisionDistance) {
				result.collisionDistance = collDist;
				result.obstacleDirection = dir;
				result.obstacleDistance = dst - minDist;
			}
		}

		return result;
	}

	/**
	 *  Similar like checkMove(), but takes relative direction, instead of absolute.
	 */
	public CheckResult checkMoveRel(double direction, double border) {
		return checkMove(direction + actualPos.getHeading(), border);
	}

	/**
	 *  Determine allowed angle to turn without collision.
	 *  @param direction Turning direction, positive for CW, negative for CCW.
	 *  @return Allowed turning angle, or Double.MAX_VALUE if there is no obstacle.
	 */
	public double checkTurn(int direction) {
		return Double.MAX_VALUE;
	}

	/**
	 *  Determine whether a movement would cause a collision.
	 *  @param direction Movement direction: FD/BK/LT/RT.
	 *  @param border Turtle border, included in collision computation.
	 */
	public CheckResult check(Movement.Direction direction, double border) {
		if (direction.isFdOrBk()) {
			return checkMoveRel(direction.toRadians(), border);
		} else {
			return new CheckResult(checkTurn(direction.isLt() ? -1 : 1));
		}
	}

	/**
	 *  Determine whether a movement would cause a collision.
	 *  @param move Movement command.
	 *  @param border Turtle border, included in collision computation.
	 */
	public CheckResult check(Movement move, double border) {
		// TODO: computation form movements with different speeds for left and right wheels!!!
		//       (such movements are currently not used at all in RLCS, so its not that important)
		return check(move.getDirection(), border);
	}

	// conversion of wall representation from absolute Line2D to relative Wall
	public Wall line2wall(Line2D line) {
		Wall wall = new Wall();

		// directions from actual position to the wall limit points
		double dir1 = actualPos.directionTo(line.getP1()),
		       dir2 = actualPos.directionTo(line.getP2());
		if (dir1 == dir2) {
			// the wall is a single point from our POV
			wall.distance = Math.min(actualPos.distance(line.getP1()), actualPos.distance(line.getP2()));
			wall.direction = dir1;
			return wall;
		}

		// the wall vector
		double dx = line.getX2() - line.getX1(),
		       dy = line.getY2() - line.getY1();
		if (dx == 0 && dy == 0) {
			throw new IllegalArgumentException("Wall is single point!");
		}

		// normal vector of the wall
		double nx = -dy, ny = dx;

		// constant member of the the wall equation (x*nx + y*ny + c = 0)
		double c = - (line.getX1() * nx + line.getY1() * ny);

		// compute orthogonal direction & distance of actual position from the wall
		double dist = (actualPos.getX() * nx + actualPos.getY() * ny + c) / Math.sqrt(nx * nx + ny * ny);
		double dir;
		if (dist > 0)
			dir = Math.atan2(-nx, -ny);
		else {
			dir = Math.atan2(nx, ny);
			dist = -dist;
		}

		// compute wall limits
		double limit1 = dist * Math.tan(Position.headingDiff(dir1, dir)),
		       limit2 = dist * Math.tan(Position.headingDiff(dir2, dir));

		// init the wall object data
		wall.distance = dist - TurtleConstants.RADIUS_MM;
		wall.direction = dir;
		wall.limitLeft = Math.min(limit1, limit2);
		wall.limitRight = Math.max(limit1, limit2);
		return wall;
	}

	// mapping to Xml
	public class XmlMappable {
		public Position.XmlMappable actualPos = (Collision.this.actualPos == null) ? null : Collision.this.actualPos.getXmlMappable();
		public Collection<Position.XmlMappable> robots = new ArrayList<Position.XmlMappable>();
		public Collection<Wall> walls = new ArrayList<Wall>();
	}
	@Override public XmlMappable getXmlMappable() {
		XmlMappable m = new XmlMappable();
		for (Position r: robots) {
			m.robots.add(r.getXmlMappable());
		}
		for (Wall w: walls) {
			m.walls.add(new Wall(w.direction, w.distance, w.limitLeft, w.limitRight));
		}
		return m;
	}
}
