package sk.robotics.rlcs.data;

import java.util.*;
import java.awt.geom.AffineTransform;
import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  Structure for data received from detection.
 *
 *  Objects of this class are mutable, detected robot positions
 *  get associated with robots upon identification, but no synchronization
 *  is needed to access objects of this class.
 */
public class DetectionData implements CustomXmlMappable
{
	// structure representing a detected robot,
	// with mutable robot number and actual position block name,
	// both filled in by AgentIdentify
	public static class Robot implements CustomXmlMappable
	{
		public final Position position;
		public Integer robotNum;
		public String blockActualPos;
		
		private Robot(Position pos) { position = pos; }
		
		public void identify(int robotNum, String blockActualPos) {
			// these assignments have to be in this order, so that when
			// robotNum is set, blockActualPos is already set as well
			// (this allows reading DetectionData without locking)
			this.blockActualPos = blockActualPos;
			this.robotNum = new Integer(robotNum);
		}

		public class XmlMappable {
			public Position.XmlMappable position = Robot.this.position.getXmlMappable();
			public Integer robotNum = Robot.this.robotNum;
			public String blockActualPos = Robot.this.blockActualPos;
		}
		@Override public XmlMappable getXmlMappable() { return new XmlMappable(); }
	}
	
	// data
	private long timestamp;
	private Collection<Robot> robots;
	private AffineTransform transform; // transformation from calibrated image to our coordinates
	
	// ctor
	public DetectionData(long timestamp, Collection<Position> robotPositions, AffineTransform transform) {
		this.timestamp = timestamp;
		
		robots = new ArrayList<Robot>(robotPositions.size());
		for (Position p: robotPositions)
			robots.add(new Robot(p));
		robots = Collections.unmodifiableCollection(robots);

		this.transform = (transform == null) ? new AffineTransform() : transform;
	}
	
	// accessors
	public long getTimeStamp() {
		return timestamp;
	}
	public Collection<Robot> getRobots() {
		return robots;
	}
	public AffineTransform getTransform() {
		return new AffineTransform(transform); // AffineTransform is mutable, so return a copy
	}

	// mapping to XML
	public class XmlMappable {
		public long timestamp = DetectionData.this.timestamp;
		public Collection<Robot.XmlMappable> robots = new ArrayList<Robot.XmlMappable>();
		public String transform;
	}
	@Override public Object getXmlMappable() {
		XmlMappable m = new XmlMappable();
		for (Robot r: robots) {
			m.robots.add(r.getXmlMappable());
		}
		m.transform = String.format("[[%.3f %.3f %.3f] [%.3f %.3f %.3f]]",
			transform.getScaleX(), transform.getShearX(), transform.getTranslateX(),
			transform.getShearY(), transform.getScaleY(), transform.getTranslateY());
		return m;
	}
}
