package sk.robotics.rlcs.data;

import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  Robot movement - direction, speed of left wheel, speed of right wheel.
 *  Objects of this class are immutable.
 */
public class Movement
	implements Comparable<Movement>, CustomXmlMappable
{
	// directions
	public enum Direction {
		FD, BK, LT, RT;

		public boolean isFd() { return this == FD; }
		public boolean isBk() { return this == BK; }
		public boolean isLt() { return this == LT; }
		public boolean isRt() { return this == RT; }
		public boolean isFdOrBk() { return this == FD || this == BK; }
		public boolean isLtOrRt() { return this == LT || this == RT; }

		public double toRadians() {
			switch (this) {
				case FD: return 0;
				case BK: return Math.PI;
				case LT: return -Math.PI/2;
				case RT: return Math.PI/2;
				default: throw new IllegalStateException();
			}
		}
	}

	// null move
	public static final Movement NO_MOVE = new Movement(Direction.FD, 0, 0);

	// full speed moves
	public static final Movement FD_FULLSPEED = new Movement(Direction.FD, 99, 99);
	public static final Movement BK_FULLSPEED = new Movement(Direction.BK, 99, 99);
	public static final Movement LT_FULLSPEED = new Movement(Direction.LT, 99, 99);
	public static final Movement RT_FULLSPEED = new Movement(Direction.RT, 99, 99);

	// data
	private Direction dir;
	private int lspeed, rspeed;

	// ctor
	public Movement(Direction dir, int lspeed, int rspeed) {
		this.dir = dir;
		this.lspeed = lspeed;
		this.rspeed = rspeed;
	}

	// factory, reusing common movement objects
	public static Movement valueOf(Direction dir, int lspeed, int rspeed) {
		if (lspeed == 99 && rspeed == 99) {
			switch (dir) {
				case FD: return FD_FULLSPEED;
				case BK: return BK_FULLSPEED;
				case LT: return LT_FULLSPEED;
				case RT: return RT_FULLSPEED;
				default: throw new IllegalArgumentException();
			}
		}
		if (lspeed == 0 && rspeed == 0) {
			return NO_MOVE;
		}
		return new Movement(dir, lspeed, rspeed);
	}

	// accessors
	public Direction getDirection() { return dir; }
	public int getLeftSpeed() { return lspeed; }
	public int getRightSpeed() { return rspeed; }
	public int getMaxSpeed() { return Math.max(lspeed, rspeed); }

	public boolean isFd() { return dir.isFd(); }
	public boolean isBk() { return dir.isBk(); }
	public boolean isLt() { return dir.isLt(); }
	public boolean isRt() { return dir.isRt(); }
	public boolean isFdOrBk() { return dir.isFdOrBk(); }
	public boolean isLtOrRt() { return dir.isLtOrRt(); }

	// comparing & hashing
	public boolean equals(Object o) {
		if (!(o instanceof Movement)) return false;
		Movement m = (Movement)o;
		return dir == m.dir && lspeed == m.lspeed && rspeed == m.rspeed;
	}
	public int compareTo(Movement m) {
		int cmp;
		if ((cmp = dir.compareTo(m.dir)) != 0) return cmp;
		if ((cmp = lspeed - m.lspeed) != 0) return cmp;
		if ((cmp = rspeed - m.rspeed) != 0) return cmp;
		return 0;
	}
	public int hashCode() {
		int result = 71;
		result = 37 * result + dir.hashCode();
		result = 37 * result + lspeed;
		result = 37 * result + rspeed;
		return result;
	}

	// string representation
	public String toString() {
		StringBuilder str = new StringBuilder(16);
		str.append(dir).append(' ').append(lspeed).append(' ').append(rspeed);
		return str.toString();
	}

	// mapping to XML
	public class XmlMappable {
		public String dir = Movement.this.dir.toString();
		public int lspeed = Movement.this.lspeed;
		public int rspeed = Movement.this.rspeed;
	}
	@Override public XmlMappable getXmlMappable() { return new XmlMappable(); }
}
