package sk.robotics.rlcs.data;

import sk.robotics.rlcs.constants.TurtleConstants;
import java.awt.geom.Point2D;
import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  Robot position - [x,y] position, and heading (0..2pi).
 *  Objects of this class are immutable.
 */
public class Position
	implements Comparable<Position>, CustomXmlMappable
{
	// data
	private double x, y, h;

	// ctor
	public Position(double x, double y, double h) {
		this.x = x;
		this.y = y;
		this.h = normalizeHeading(h);
	}

	/** Contructs position from string returned by Robot.pos(). */
	public Position(String pos) {
		String[] parts = pos.trim().split(" ");
		if (parts.length != 3)
			throw new IllegalArgumentException("wrong format of odometry string!");

		this.x = java.lang.Double.parseDouble(parts[0]) * TurtleConstants.STEPSIZE_MM;
		this.y = java.lang.Double.parseDouble(parts[1]) * TurtleConstants.STEPSIZE_MM;
		this.h = normalizeHeading(java.lang.Double.parseDouble(parts[2]));
	}

	// heading normalization - modulo 2*PI
	private static double normalizeHeading(double h) {
		while (h < 0) h += 2 * Math.PI;
		while (h >= 2 * Math.PI) h -= 2 * Math.PI;
		return h;
	}

	// accessors
	public double getPosX() { return x; }
	public double getPosY() { return y; }
	public double getHeading() { return h; }

	// Point2D compatible accessors
	public double getX() { return getPosX(); }
	public double getY() { return getPosY(); }

	// Point2D adapter
	private class Point2DAdapter extends Point2D {
		public double getX() { return x; }
		public double getY() { return y; }

		public void setLocation(double x, double y) {
			throw new UnsupportedOperationException();
		}
	}
	private Point2DAdapter point2DAdapter = new Point2DAdapter();

	/** Get the position as Point2D (without heading). */
	public Point2D asPoint2D() { return point2DAdapter; }

	// string representation
	public String toString() {
		StringBuilder str = new StringBuilder(64);
		str.append('[').append(x).append(' ').append(y).append(' ').append(h).append(']');
		return str.toString();
	}

	// comparing & hashing
	public boolean equals(Object o) {
		if (!(o instanceof Position)) return false;
		Position p = (Position)o;
		return x == p.x && y == p.y && h == p.h;
	}
	public int compareTo(Position p) {
		if (x < p.x) return -1;
		if (x > p.x) return  1;
		if (y < p.y) return -1;
		if (y > p.y) return  1;
		if (h < p.h) return -1;
		if (h > p.h) return  1;
		return 0;
	}
	public int hashCode() {
		int result = 73;
		result = 37 * result + hashDouble(x);
		result = 37 * result + hashDouble(y);
		result = 37 * result + hashDouble(h);
		return result;
	}
	private static int hashDouble(double val) {
		long v = Double.doubleToLongBits(val);
		return (int)(v ^ (v >>> 32));
	}

	/** Computes distance from another point. */
	public double distance(Point2D p) {
		return asPoint2D().distance(p);
	}
	public double distance(Position p) {
		return asPoint2D().distance(p.asPoint2D());
	}

	/** Computes direction to another point. */
	public double directionTo(Position p) {
		return directionTo(asPoint2D(), p.asPoint2D());
	}
	public double directionTo(Point2D p) {
		return directionTo(asPoint2D(), p);
	}
	public static double directionTo(Point2D p1, Point2D p2) {
		return Math.atan2(p2.getX() - p1.getX(), p2.getY() - p1.getY());
	}

	/** Computes heading difference between 2 positions.
	 *  @return heading direction from interval [-pi,pi] */
	public double headingDiff(Position p) {
		return headingDiff(getHeading(), p.getHeading());
	}
	public double headingDiff(double h) {
		return headingDiff(getHeading(), h);
	}
	public static double headingDiff(double h1, double h2) {
		double diff = h1 - h2;
		while (diff >  Math.PI) diff -= 2 * Math.PI;
		while (diff < -Math.PI) diff += 2 * Math.PI;
		return diff;
	}

	// mapping to XML
	public class XmlMappable {
		public double x = Position.this.x;
		public double y = Position.this.y;
		public double h = Position.this.h;
	}
	@Override public XmlMappable getXmlMappable() { return new XmlMappable(); }
}
