package sk.robotics.rlcs.data;

import java.util.Collection;
import java.util.AbstractCollection;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;

import com.microstepmis.xplatform.CustomXmlMappable;

/**
 *  Collection class holding queue of positions.
 *
 *  Allows constant-time push and pop, as well as constant-time random access.
 *  Positions in the queue are always sorted by time (but this is application's
 *  responsibility), so the queue also provides logarithmic-time searching
 *  by position age.
 * <p>
 *  This class also allows a shift operation, which shifts all current stored
 *  positions by given delta x, y, and h. But this shift operation does not
 *  actually change the stored positions, it only remembers that all positions
 *  older than the shift are to be treated as shifted. Thus, it is a constant
 *  time operation.
 * <p>
 *  This class' implementation is strongly inspired by java.util.ArrayDeque.
 */
public class PositionQueue
	extends AbstractCollection<TimedPosition>
	implements CustomXmlMappable
{
	// A nested queue of position shifts.
	private PositionQueue shifts;
	
	// The array in which the positions are stored. It is always a power of 2.
	private TimedPosition[] elements;
	
	// Head & Tail of the queue.
	private int head, tail;
	
	// Double the capacity of this deque.
	private void doubleCapacity() {
		assert head == tail; // queue must be full
		int p = head;
		int n = elements.length;
		int r = n - p; // number of elements to the right of p
		int newCapacity = n << 1;
		if (newCapacity < 0)
			throw new IllegalStateException("Sorry, deque too big");
		TimedPosition[] a = new TimedPosition[newCapacity];
		System.arraycopy(elements, p, a, 0, r);
		System.arraycopy(elements, 0, a, r, p);
		elements = a;
		head = 0;
		tail = n;
	}
	
	/** Constructs an empty queue with an initial capacity for 16 elements. */
	public PositionQueue() {
		elements = new TimedPosition[16];
	}
	
	/** Constructs an empty queue with specified initial capacity. */
	public PositionQueue(int numElements) {
		int initialCapacity = 16;
		if (numElements >= initialCapacity) {
			initialCapacity = numElements;
			initialCapacity |= (initialCapacity >>>  1);
			initialCapacity |= (initialCapacity >>>  2);
			initialCapacity |= (initialCapacity >>>  4);
			initialCapacity |= (initialCapacity >>>  8);
			initialCapacity |= (initialCapacity >>> 16);
			if (++initialCapacity < 0) initialCapacity >>>= 1;
		}
		elements = new TimedPosition[initialCapacity];
	}
	
	// *** Queue Methods ***
	
	/** Adds the specified element to (the tail of) this queue. */
	public boolean add(TimedPosition e) {
		if (e == null) // we can't store nulls!
			throw new NullPointerException();
		elements[tail] = e;
		if ( (tail = (tail + 1) & (elements.length - 1)) == head)
			doubleCapacity();
		return true;
	}
	
	/** Removes the head element of this queue.
	 *  @return True if an item was removed, false if queue was empty */
	public boolean removeOldest() {
		TimedPosition oldest = elements[head]; // null if queue is empty
		if (oldest == null) return false;
		
		// remove the element
		elements[head] = null; // null out the slot
		head = (head + 1) & (elements.length - 1);
		
		// remove all redundant shifts
		while (shifts != null && !shifts.isEmpty() && shifts.getOldest().getTimeStamp() <= oldest.getTimeStamp())
			shifts.removeOldest();
		
		return true;
	}
	
	/** Retrieves the head element of this queue, without removing it.
	 *  @return the oldest item, or null if queue is empty */
	public TimedPosition getOldest() {
		return applyShifts(getOldestNoShift());
	}
	private TimedPosition getOldestNoShift() {
		return elements[head]; // null if queue is empty
	}
	
	/** Retrieves the tail element of this queue, without removing it.
	 *  @return the newest item, or null if queue is empty */
	public TimedPosition getNewest() {
		return applyShifts(getNewestNoShift());
	}
	private TimedPosition getNewestNoShift() {
		return elements[(tail - 1) & (elements.length - 1)]; // null if queue is empty
	}
	
	// *** Collection Methods ***
	
	/** Returns the number of elements in this queue. */
	public int size() {
		return (tail - head) & (elements.length - 1); // elements.length is always power of 2
	}
	
	/** Checks for queue emptyness. */
	public boolean isEmpty() {
		return head == tail;
	}
	
	/** Returns an iterator over the elements in this queue. The elements
	 *  will be ordered from head to tail (from oldest to newest). */
	public Iterator<TimedPosition> iterator() {
		return new Iterator<TimedPosition>()
		{
			// Index of element to be returned by subsequent call to next.
			private int cursor = head;
			
			// Tail recorded at construction (also in remove), to stop
			// iterator and also to check for comodification.
			private int fence = tail;
			
			public boolean hasNext() {
				return cursor != fence;
			}
			
			public TimedPosition next() {
				if (cursor == fence)
					throw new NoSuchElementException();
				TimedPosition result = elements[cursor];
				if (tail != fence || result == null) // catches comodifications which corrupt traversal
					throw new ConcurrentModificationException();
				cursor = (cursor + 1) & (elements.length - 1);
				return applyShifts(result);
			}
			
			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}
	
	/** Removes all of the elements from this queue. */
	public void clear() {
		for (int i = head; i != tail; i = (i + 1) & (elements.length - 1))
			elements[i] = null;
		head = tail = 0;
		
		if (shifts != null)
			shifts.clear();
	}
	
	// *** Random access methods ***
	
	/** Retrieve element by index. The head element has index 0.
	 *  Does not apply any shifts. */
	private TimedPosition get(int index) {
		if (index < 0 || index >= size())
			throw new IndexOutOfBoundsException();
		return elements[(index + head) & (elements.length - 1)];
	}
	
	/**
	 *  Find index by time.
	 *  If the queue is empty, zero is returned.
	 *  If given time is older than oldest position in the queue, -1 is returned.
	 *  If given time is newer than newest position in the queue, size() is returned.
	 *  If there is a position from exactly the time given, its index is returned.
	 *  Otherwise, a fractional index corresponding to that exact time is returned.
	 */
	private double findIndex(long timestamp) {
		// special cases
		if (isEmpty()) return 0; // empty queue
		long oldestTime = getOldestNoShift().getTimeStamp(),
		     newestTime = getNewestNoShift().getTimeStamp();
		if (timestamp <= oldestTime) return (timestamp == oldestTime) ? 0 : -1;
		if (timestamp >= newestTime) return (timestamp == newestTime) ? size() - 1 : size();
		// if we got here, it is sure that the queue contains at least 2 elements,
		// and its head is strictly older, and tail strictly newer than the given timestamp
		
		// find closest older and newer elements
		int minIndex = 0, maxIndex = size() - 1;
		while (minIndex + 1 < maxIndex) {
			int midIndex = (minIndex + maxIndex) / 2;
			long midTime = get(midIndex).getTimeStamp();
			
			if (midTime < timestamp) minIndex = midIndex;
			else if (midTime > timestamp) maxIndex = midIndex;
			else return midIndex; // this position's timestamp matches exactly
		}
		// if we got here, then maxIndex == minIndex + 1, and minTime < timestamp < maxTime
		
		// interpolate the result
		TimedPosition minPos = get(minIndex),
		              maxPos = get(maxIndex);
		double q = (timestamp - minPos.getTimeStamp()) / (double)(maxPos.getTimeStamp() - minPos.getTimeStamp());
		return minIndex + q;
	}
	
	/**
	 *  Get position estimation for given time.
	 *  If the queue is empty, null is returned.
	 *  For times older than oldest position in the queue, the oldest position is returned.
	 *  For times newer than newest position in the queue, the newest position is returned.
	 *  If there is a position from exactly the time given, it is returned.
	 *  Otherwise, a position is interpolated from the closest older and newer positions.
	 */
	public Position getPositionByTime(long timestamp) {
		// check for queue emptyness
		if (isEmpty()) return null;
		
		// find index of matching element
		double index = findIndex(timestamp);
		
		// is result the oldest or newest item?
		if (index <= 0) return getOldest();
		if (index >= size() - 1) return getNewest();
		
		// is result exactly one of stored positions?
		if (index == Math.round(index))
			return applyShifts(get((int)index));
		
		// otherwise, interpolate the result
		TimedPosition minPos = get((int)Math.floor(index)),
		              maxPos = get((int)Math.ceil(index));
		double q = index - Math.floor(index);
		double headingDiff = maxPos.getHeading() - minPos.getHeading();
		if (headingDiff != 0 && Math.abs(headingDiff) > Math.PI)
			headingDiff = (2*Math.PI - Math.abs(headingDiff)) * -Math.signum(headingDiff);
		return new Position(
			minPos.getPosX() + q * (maxPos.getPosX() - minPos.getPosX()),
			minPos.getPosY() + q * (maxPos.getPosY() - minPos.getPosY()),
			minPos.getHeading() + q * headingDiff);
	}
	
	/** Shift all positions in the queue. */
	public void shiftBy(double dx, double dy, double dh, long timestamp) {
		if (isEmpty()) return; // no positions to shift
		if (dx == 0 && dy == 0 && dh == 0) return; // zero shift, ignore
		if (shifts == null) shifts = new PositionQueue(); // init the shifts queue
		shifts.add(new TimedPosition(new Position(dx, dy, dh), timestamp));
	}
	public void shiftBy(double dx, double dy, double dh) {
		shiftBy(dx, dy, dh, System.currentTimeMillis());
	}
	
	/** Apply all applicable shifts to a position. */
	public TimedPosition applyShifts(TimedPosition pos) {
		// special cases
		if (pos == null) return null;
		if (shifts == null || shifts.isEmpty()) return pos;
		
		// find index in the shifts queue, corresponding to timestamp of the position
		double index = shifts.findIndex(pos.getTimeStamp());
		if (index >= shifts.size() - 1) return pos; // no applicable shifts
		
		// cummulate all shifts
		double dx = 0, dy = 0, dh = 0;
		for (int i = (int)Math.floor(index) + 1; i < shifts.size(); ++i) {
			Position p = shifts.get(i);
			dx += p.getPosX();
			dy += p.getPosY();
			dh += p.getHeading();
		}
		
		// if the cummulated shift is zero, return the original position
		if (dx == 0 && dy == 0 && dh == 0) return pos;
		
		// otherwise, shift the original position
		Position shiftedPos = new Position(
			pos.getPosX() + dx,
			pos.getPosY() + dy,
			pos.getHeading() + dh);
		return new TimedPosition(shiftedPos, pos.getTimeStamp());
	}

	// mapping to XML
	@Override public synchronized Collection<TimedPosition.XmlMappable> getXmlMappable() {
		Collection<TimedPosition.XmlMappable> positions = new ArrayList<TimedPosition.XmlMappable>(size());
		for (TimedPosition p: this) {
			positions.add(p.getXmlMappable());
		}
		return positions;
	}
}
