package sk.robotics.rlcs.security;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Scheduler;
import sk.robotics.rlcs.Main;

import java.security.*;
import java.util.*;
import java.io.File;
import java.lang.reflect.Constructor;

public class SecurityManager extends java.lang.SecurityManager {

	// set of allowed subclasses of Trusted thread class
	private static Collection<Class> allowedTrusted = new HashSet<Class>();
	static {
		allowedTrusted.add(Agent.Cfg.Validator.class);
		allowedTrusted.add(Agent.ThreadCreator.class);
	}

	// sets of allowed permissions, per agent class
	private Map<Class, Permissions> permissions = new HashMap<Class, Permissions>();

	/**
	 *  Constructor, initializes the allowed permissions.
	 */
	public SecurityManager(PermissionDesc[] configPerms) {
		try {
			// FIXME: when the Restricted class is loaded only on first call to checkRestricted(),
			//  LinkageError si throw. This workaround forces the JVM to load the Restricted class now.
			//  This may be a bug in Java 1.6.0.
			//Restricted.class.equals(null);

			// fill the map
			for (PermissionDesc permDesc: configPerms) {
				// agent class
				Class<?> agentClass = Class.forName(permDesc.agent);

				// current permissions for this agent
				Permissions agentPerms = permissions.get(agentClass);
				if (agentPerms == null) {
					permissions.put(agentClass, agentPerms = new Permissions());
				}

				// class and ctor for the new permission
				Class<?> permClass = Class.forName(permDesc.perm);
				Constructor<?> permCtor = permClass.getConstructor(String.class, String.class);

				// add the new permission
				Permission newPerm = (Permission) permCtor.newInstance(permDesc.name, permDesc.action);
				agentPerms.add(newPerm);
			}

			// deny further changing of permissions
			for (Permissions perms: permissions.values()) {
				perms.setReadOnly();
			}
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	// this helper, if called from an agent or an agent-launched worker thread,
	// returns the agent, and returns null otherwise (if called from another thread)
	public static Agent getAgentForThread(Thread thread) {
		// check for agents
		if (thread instanceof Agent.AgentThread) {
			return ((Agent.AgentThread)thread).getAgent();
		}

		// check for agent-launched worker threads
		ThreadGroup group = thread.getThreadGroup();
		if (group instanceof Agent.WorkerThreadGroup) {
			return ((Agent.WorkerThreadGroup)group).getAgent();
		}

		// for other threads, return null
		return null;
	}

	// check whether a thread is privileged (privileged threads have all permissions granted)
	private static boolean isPrivileged(Thread thread) {
		// all system threads are privileged
		if (thread.getThreadGroup() == Main.getSystemThreadGroup()) return true;

		// the main thread is privileged
		if (thread == Main.getMainThread()) return true;

		// shutdown hooks are privileged (only main thread can create shutdown hooks)
		if (thread instanceof Main.ShutdownHook) return true;

		// other threads are not privileged
		return false;
	}

	// check whether a thread is restricted, and throw SecurityException if so
	// (the Permission parameter is just for debugging)
	private static void checkRestricted(Thread thread, Object perm) {
		if (thread instanceof Restricted) {
			logDenied(((Restricted)thread).getAgentClass(), thread, perm);
			throw new SecurityException("Restricted thread");
		}
	}
	private static void checkRestricted(Thread thread) {
		checkRestricted(thread, null);
	}

	// log a permission denied message
	private static void logDenied(Object agent, Thread thread, Object perm) {
		synchronized (System.err) {
			System.err.print("Permission denied (");
			if (agent != null) {
				System.err.printf("agent: %s, ", agent);
			}
			if (!(thread instanceof Agent.AgentThread) || ((Agent.AgentThread)thread).getAgent() != agent) {
				System.err.printf("thread: %s, ", thread);
			}
			System.err.printf("permission: %s)\n", perm);
		}
	}

	/**
	 *  Select thread group for new threads.
	 *  If an agent is creating a new thread, place it inside its worker threads group.
	 */
	public ThreadGroup getThreadGroup() {
		Agent agent = getAgentForThread(Thread.currentThread());
		/*if (agent != null) {
			synchronized (System.err) {
				System.err.printf("SecurityManager.getThreadGroup, thread %s, agent %s\n", Thread.currentThread(), agent);
			}
		}*/
		return (agent == null) ? super.getThreadGroup() : agent.getWorkerThreadGroup();
	}

	/**
	 *  Check for permission.
	 */
	public void checkPermission(Permission perm) {
		// get calling thread
		Thread thread = Thread.currentThread();

		// grant all permissions to privileged threads
		if (isPrivileged(thread)) return;

		// grant needed permissions for client handler threads
		if (thread instanceof Main.ClientHandler) {
			/*if (Main.ClientHandler.permissions.implies(perm)) return;
			logDenied(null, thread, perm);
			throw new SecurityException("Permission denied for client handler.");*/

			// FIXME: for now, we grant all permissions to client handlers
			return;
		}

		// grant needed permissions to allowed trusted threads
		if (thread instanceof Trusted) {
			if (!allowedTrusted.contains(thread.getClass())) {
				logDenied(null, thread, perm);
				throw new SecurityException("Permission denied for fake trusted thread.");
			}
			if (!((Trusted)thread).getPermissions().implies(perm)) {
				logDenied(null, thread, perm);
				throw new SecurityException("Permission denied, not one of allowed permissions for trusted thread.");
			}
			return; // OK
		}

		// check whether this is an agent thread, or a thread launched from an agent,
		// or a restricted thread with agent permissions
		Class agentClass = null;
		Agent agent = getAgentForThread(thread);
		if (agent != null) {
			agentClass = agent.getClass();
		} else if (thread instanceof Restricted) {
			agentClass = ((Restricted)thread).getAgentClass();
		}
		if (agentClass != null) {
			// agent thread, allow only explicitely allowed permissions
			Permissions allowed = permissions.get(agentClass);
			if (allowed != null && allowed.implies(perm)) return;

			logDenied(agentClass, thread, perm);
			throw new SecurityException("Permission denied for agent.");
		}

		// do nothing for dead threads (app shutdown)
		// (this really happens under linux, it is actually 100% reproducible, and
		//  throwing exception in this case stops the JVM immediately, Java bug maybe?)
		if (thread.getThreadGroup() == null) return;

		// NB: we don't use java.util.logging at all, but it still registers a shutdown hook,
		// which requires LoggingPermission, so allow it here as a special case, just to
		// avoid the security exception in logs on every app shutdown
		if (perm instanceof java.util.logging.LoggingPermission &&
		    thread.getClass().getName().equals("java.util.logging.LogManager$Cleaner")) {
			return;
		}

		// unknown calling thread, reject
		logDenied(null, thread, perm);
		throw new SecurityException("Unknown calling thread: " + thread);
	}

	/**
	 *  Check thread access.
	 */
	public void checkAccess(Thread thread) {
		Thread callingThread = Thread.currentThread();

		// every thread has access to itself
		if (thread == callingThread) return;

		// deny all permissions for restricted threads,
		// and grant all permissions to privileged threads
		checkRestricted(callingThread, thread);
		if (isPrivileged(callingThread)) return;

		// allow client handler threads to...
		if (callingThread instanceof Main.ClientHandler) {
			// start restricted threads
			if (thread instanceof Restricted) return;

			// start agent threads
			if (thread instanceof Agent.AgentThread) return;
		}

		// grant agent thread creator access to agent threads
		if (callingThread instanceof Agent.ThreadCreator && thread instanceof Agent.AgentThread) {
			return;
		}

		// grant agents access to their worker threads
		Agent agent = getAgentForThread(callingThread);
		if (agent != null && thread.getThreadGroup() == agent.getWorkerThreadGroup()) {
			return;
		}

		// do nothing for dead threads (app shutdown)
		if (callingThread.getThreadGroup() == null) return;

		// access denied
		synchronized (System.err) {
			System.err.printf("Access to thread %s denied for thread %s.\n", thread, callingThread);
		}
		throw new SecurityException("Thread access denied");
	}

	/**
	 *  Check thread group access.
	 */
	public void checkAccess(ThreadGroup group) {
		Thread callingThread = Thread.currentThread();

		// deny all permissions for restricted threads,
		// and grant all permissions to privileged threads
		checkRestricted(callingThread, group);
		if (isPrivileged(callingThread)) return;

		// any thread can access its own thread group
		if (callingThread.getThreadGroup() == group) return;

		// allow agents with granted RLCSPermission("subThread") to access
		// their worker thread groups (this allows also spawning worker threads)
		Agent agent = getAgentForThread(callingThread);
		if (agent != null && group == agent.getWorkerThreadGroup()) {
			checkPermission(new RLCSPermission("subThread"));
			return; // OK, access granted
		}

		// do nothing for dead threads (app shutdown)
		if (callingThread.getThreadGroup() == null) return;

		synchronized (System.err) {
			System.err.printf("Access to thread group %s denied for thread %s.\n", group, callingThread);
		}
		throw new SecurityException("ThreadGroup access denied");
	}
}
