package sk.robotics.rlcs.security;

import java.security.Permissions;
import java.util.*;

/**
 *  This class provides a way to execude some code in a thread with more privileges
 *  than the calling thread.
 * <p>
 *  Of course, threads of this class have no permissions,
 *  only subclasses explicitely allowed in SecurityManager are granted permissions
 *  specified through methods of this class.
 * <p>
 *  Threads of this class, as opposed to Restricted threads, are not spawned
 *  on demand, because that would require special permissions from the calling thread.
 *  Instead, multiple request are queued (serialized) and executed in single
 *  instance of a thread of this class.
 */
public class Trusted extends Thread {

	// set of granted permissions for this thread
	private Permissions permissions;

	// queue of pending execution requests
	private Queue<Runnable> queue = new LinkedList<Runnable>();

	// map of finished runnables, with thrown exceptions
	private Map<Runnable, RuntimeException> finished = new HashMap<Runnable, RuntimeException>();

	/**
	 *  Constructor creates and starts the thread, initializing set of permissions.
	 */
	protected Trusted(String name, Permissions perms) {
		super(name);
		this.permissions = perms;

		start();
	}

	/**
	 *  Get set of permissions associated with this thread.
	 *  These permissions are granted to this thread, but only
	 *  if it is explicitely allowed in SecurityManager.
	 */
	public Permissions getPermissions() {
		return permissions;
	}

	/**
	 *  The thread function.
	 *  Waits for requests, and executes them. Stores finished targets,
	 *  along with exceptions they throw, in the finished map.
	 */
	public void run() {
		for (;;) {
			// wait for a target
			Runnable target = null;
			synchronized (this) {
				for (;;) {
					target = queue.poll();
					if (target != null) break;

					try { wait(); }
					catch (InterruptedException e) {}
				}
			}

			// execute the target, storing exception it throws (if any)
			RuntimeException exception = null;
			try { target.run(); }
			catch (RuntimeException e) { exception = e; }

			// store the target, along with the exception, in the finished map,
			// and notify exec method waiting for it
			synchronized (this) {
				finished.put(target, exception);
				notifyAll();
			}
		}
	}

	/**
	 *  Execute a Runnable target in this thread.
	 *  Waits for the target to get executed, and rethrows any
	 *  RuntimeException it might throw.
	 */
	public synchronized void exec(Runnable target) {
		// add the target to queue, and notify the thread
		queue.add(target);
		notifyAll();

		// wait for the target to get executed
		do {
			try { wait(); }
			catch (InterruptedException e) {}
		}
		while (!finished.containsKey(target));

		// if the target threw an exception, rethrow it
		RuntimeException ex = finished.remove(target);
		if (ex != null) throw new RuntimeException(ex);
	}

}
