package sk.robotics.rlcs.test;

import sk.robotics.rlcs.data.*;
import sk.robotics.rlcs.constants.*;
import sk.robotics.rlcs.agentspace.*;

import org.junit.Test;
import static org.junit.Assert.*;

import java.util.*;
import java.awt.geom.Line2D;

public class Tests
{
	@Test public void testSpace()
	{
		Space space = Space.getInstance();

		space.delete("test", 10000, 1);
		space.write("test", "hello");
		assertNull(space.read("test"));

		space.delete("test", 10000, Float.MAX_VALUE);
		space.write("test", "world", 0, 100.0f);
		assertNull(space.read("test"));

		space.delete("test", 0, Float.POSITIVE_INFINITY);
		space.write("test", "hu!");
		assertEquals("hu!", space.read("test"));
	}

	@Test public void testPosition()
	{
		assertEquals(0.0, new Position(0, 0, 4 * Math.PI).getHeading(), 0.0);
		assertEquals(Math.PI, new Position(0, 0, 4 * Math.PI + Math.PI).getHeading(), 0.0);
		
		Position p = new Position("1.0 2.0 3.0");
		assertEquals(1.0 * TurtleConstants.STEPSIZE_MM, p.getPosX(), 0.0);
		assertEquals(2.0 * TurtleConstants.STEPSIZE_MM, p.getPosY(), 0.0);
		assertEquals(3.0, p.getHeading(), 0.0);
		
		assertEquals(0.0, new Position(0, 0, 0).directionTo(new Position(0, 1, 0)), 0.0);
		assertEquals(Math.PI, new Position(0, 0, 0).directionTo(new Position(0, -1, 0)), 0.0);
		assertEquals(Math.PI/2, new Position(0, 0, 0).directionTo(new Position(1, 0, 0)), 0.0);
		assertEquals(-Math.PI/2, new Position(0, 0, 0).directionTo(new Position(-1, 0, 0)), 0.0);
		
		assertEquals(-Math.PI, Position.headingDiff(0, Math.PI));
		assertEquals(Math.PI, Position.headingDiff(Math.PI, 0));
		assertEquals(-Math.PI/2, Position.headingDiff(-Math.PI/4, Math.PI/4));
		assertEquals(Math.PI/2, Position.headingDiff(5*Math.PI/4, 3*Math.PI/4));
	}
	
	@Test public void testCollision()
	{
		double R = TurtleConstants.RADIUS_MM;

		double X = 500;
		double Y = 300;
		double H = 0;
		Position pos = new Position(X, Y, H);

		Line2D w1 = new Line2D.Double(X - 2000, Y + R + 1, X + 2000, Y + R + 1),
		       w2 = new Line2D.Double(X - 2000, Y - R - 1, X + 2000, Y - R - 1);
		
		Collection<Line2D> walls = new ArrayList<Line2D>();
		Collection<Position> robots = new ArrayList<Position>();

		walls.add(w1);
		walls.add(w2);

		Collision coll = new Collision(pos, walls, robots);
		
		Collision.Wall w1x = coll.line2wall(w1),
		               w2x = coll.line2wall(w2);
		assertEquals(0.0, w1x.direction, 0.000001);
		assertEquals(1.0, w1x.distance, 0.000001);
		assertEquals(-2000.0, w1x.limitLeft, 0.000001);
		assertEquals(2000.0, w1x.limitRight, 0.000001);
		assertEquals(Math.PI, w2x.direction, 0.000001);
		assertEquals(1.0, w2x.distance, 0.000001);
		assertEquals(-2000.0, w2x.limitLeft, 0.000001);
		assertEquals(2000.0, w2x.limitRight, 0.000001);

		assertEquals(1.0, coll.checkMove(0, 0).getCollisionDistance(), 0.0);
		assertEquals(Math.sqrt(2), coll.checkMove(Math.PI/4, 0).getCollisionDistance(), 0.00001);
		assertEquals(Math.sqrt(2), coll.checkMove(-Math.PI/4, 0).getCollisionDistance(), 0.00001);
		
		assertEquals(1.0, coll.checkMove(Math.PI, 0).getCollisionDistance(), 0.0);
		assertEquals(Math.sqrt(2), coll.checkMove(3*Math.PI/4, 0).getCollisionDistance(), 0.00001);
		assertEquals(Math.sqrt(2), coll.checkMove(5*Math.PI/4, 0).getCollisionDistance(), 0.00001);


		walls.clear();
		robots.add(new Position(X + TurtleConstants.DIAMETER_MM + 1, Y, 0));
		coll = new Collision(pos, walls, robots);

		assertEquals(1.0, coll.checkMove(Math.PI/2, 0).getCollisionDistance(), 0.0);
		assertEquals(1.005, coll.checkMove(Math.PI/2 + 0.1, 0).getCollisionDistance(), 0.0001);
		assertEquals(1.005, coll.checkMove(Math.PI/2 - 0.1, 0).getCollisionDistance(), 0.0001);


		robots.clear();
		robots.add(new Position(485.795, 505.072, 4.43968));
		coll = new Collision(new Position(592.2435413124851, 526.1196376916109, 4.443298820832862), walls, robots);
		assertTrue(coll.check(Movement.Direction.FD, 0).getCollisionDistance() <= 0);
	}
	
	@Test public void testPositionQueue()
	{
	}
}
