package sk.robotics.rlcs.web.display;

import java.net.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.applet.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.font.*;
import javax.swing.*;
import javax.imageio.*;
import javax.imageio.stream.*;
import java.util.concurrent.atomic.*;

/**
 *  Applet for displaying images streamed by ImageStreamAgent.
 *
 *  For now, this class assumes ImageStreamAgent streams a MJPEG
 *  stream. In particular, it reads the input line by line,
 *  until it finds a "ContentLength: xxx" line, followed by an empty
 *  line. Then it reads the image, with the specified xxx length.
 */
public class Display extends Applet
{
	// default port where ImageServerAgent is listening
	// (default host is the one serving the page, and both
	//  are overridable by parameters in HTML <applet> tag)
	private final static int DEFAULT_REMOTE_PORT = 35755;

	// real host and port
	private String remoteHost;
	private int remotePort;

	// other parameters, overridable from HTML (by parameters of same names like these)
	private boolean scaleImage = false; // scale image to applet size (keeping aspect ratio)
	private boolean centerImage = false; // center image on applet

	// last received image
	private AtomicReference<BufferedImage> lastImage = new AtomicReference<BufferedImage>();

	// error message to display
	private AtomicReference<String> lastMessage = new AtomicReference<String>();

	// background color
	private Color bgColor;

	// a back-buffer (reduces flicker)
	private Image bufferImage = null;
	private Graphics bufferGraphics = null;

	/**
	 *  Initialization.
	 */
	@Override public void init()
	{
		// get bg color from html attribute of the applet
		try {
			String bgstr = getParameter("bgcolor");
			if (!bgstr.startsWith("#"))
				bgstr = "#" + bgstr;
			bgColor = Color.decode(bgstr);
		}
		catch (Exception e) {
			bgColor = Color.WHITE;
		}

		// get image properties from html params
		scaleImage = Boolean.valueOf(getParameter("scaleImage"));
		centerImage = Boolean.valueOf(getParameter("centerImage"));

		// determine the remote host and port
		remoteHost = getParameter("host");
		if (remoteHost == null || remoteHost.length() == 0) {
			remoteHost = getDocumentBase().getHost();
		}
		try { remotePort = Integer.parseInt(getParameter("port")); }
		catch (Exception e) { remotePort = DEFAULT_REMOTE_PORT; }
	}

	/**
	 *  Startup.
	 */
	@Override public void start() {
		connectionThread = new ConnectionThread();
		connectionThread.start();
	}

	/**
	 *  Stop.
	 */
	@Override public void stop() {
		connectionThread.quit();
	}

	// ImageInputStream implementation backed by byte array
	private static class ImgInput extends ImageInputStreamImpl {
		private byte[] data;
		private int off, len;
		private ImgInput(byte[] data, int off, int len) {
			this.data = data;
			this.off = off;
			this.len = len;
		}

		public int read() throws IOException {
			if (streamPos >= len) throw new EOFException();
			int val = data[off + (int)streamPos++];
			return val < 0 ? val + 256 : val;
		}
		public int read(byte[] buff, int off, int len) throws IOException {
			if (streamPos >= this.len) throw new EOFException();
			if (streamPos + len > this.len) len = (int)(this.len - streamPos);
			System.arraycopy(data, this.off + (int)streamPos, buff, off, len);
			streamPos += len;
			return len;
		}
	}

	/**
	 *  The connection thread.
	 */
	private class ConnectionThread extends Thread {
		private SocketChannel channel = null;
		private volatile boolean quitFlag = false;

		private ConnectionThread() {
			super("RLCS/Display/ConnectionThread");
		}

		public void quit() {
			quitFlag = true;
			interrupt();
		}

		@Override public void run() {
			SocketChannel sock = null; // socket, connected to AgentImageStream
			DataInput input = null; // input stream
			byte[] buffer = null; // image buffer
			boolean connectFailed = false;

			for (;;) {
				try {
					// check quit flag
					if (quitFlag) {
						try { if (channel != null) channel.close(); }
						catch (Exception e) {}
						return;
					}

					// reconnect if needed
					if (channel == null) {
						channel = SocketChannel.open();
						channel.connect(new InetSocketAddress(remoteHost, remotePort));
						input = new DataInputStream(new BufferedInputStream(channel.socket().getInputStream()));
						updateMessage("connected");
						connectFailed = false;
					}

					// read lines, until we get one with ContentLength
					int contentLength = 0;
					for (;;) {
						String line = input.readLine();
						if (line == null) { // EOF
							throw new EOFException();
						}
						if (line.startsWith("Content-Length: ")) {
							contentLength = Integer.parseInt(line.substring(16));

							// after ContentLength, there should be an empty line
							line = input.readLine();
							if (line == null) {
								throw new EOFException();
							}
							if (line.equals("")) {
								break; // ok, let's read the image
							}
						}
					}

					// realloc the buffer if needed
					if (buffer == null || buffer.length < contentLength) {
						buffer = new byte[contentLength];
					}

					// read the image data
					input.readFully(buffer, 0, contentLength);

					// load the image
					BufferedImage image = ImageIO.read(new ImgInput(buffer, 0, contentLength));
					if (image == null) {
						throw new Exception("Failed to load the image, compatible reader not found");
					}

					// display the image
					lastImage.set(image);
					repaint();

					// short sleep
					Thread.sleep(100);
				}
				catch (InterruptedException e) {}
				catch (ClosedByInterruptException e) {}
				catch (EOFException e) {
					updateMessage("disconnected");
					try { channel.close(); }
					catch (IOException ee) {}
					channel = null;
				}
				catch (IOException e) {
					if (channel == null) {
						// connection failed, display the error only once
						if (!connectFailed) {
							e.printStackTrace(System.err);
							connectFailed = true;
						}
					} else {
						e.printStackTrace(System.err);
					}
					updateMessage("Socket I/O error: " + e.getMessage());
					try { if (channel != null) channel.close(); }
					catch (IOException ee) {}
					channel = null;

					try { Thread.sleep(1000); }
					catch (InterruptedException ee) {}
				}
				catch (Exception e) {
					e.printStackTrace(System.err);
					updateMessage("Error: " + e);

					try { Thread.sleep(1000); }
					catch (InterruptedException ee) {}
				}
			}
		}
	}
	private ConnectionThread connectionThread;

	/**
	 *  Update the displayed message.
	 */
	private void updateMessage(String message) {
		lastMessage.set(message);
		repaint();
	}

	/**
	 *  Paint the applet.
	 */
	@Override public void paint(Graphics appletGraphics) {
		// [re]allocate the back-buffer, if needed (if not yet allocated, or is too small)
		if (bufferImage == null || bufferImage.getWidth(null) < getWidth() || bufferImage.getHeight(null) < getHeight()) {
			bufferImage = createImage(getWidth(), getHeight());
			bufferGraphics = bufferImage.getGraphics();
		}

		Graphics2D g = (Graphics2D) bufferGraphics;

		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
		                   scaleImage ? RenderingHints.VALUE_ANTIALIAS_ON
		                              : RenderingHints.VALUE_ANTIALIAS_OFF);
		g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
		                   scaleImage ? RenderingHints.VALUE_INTERPOLATION_BICUBIC
		                              : RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);

		BufferedImage image = lastImage.get();
		String message = lastMessage.get();

		double w = getWidth();
		double h = getHeight();

		// draw background
		g.setPaint(bgColor);
		g.fill(new Rectangle2D.Double(0, 0, w, h));
		
		// draw the image
		if (image != null) {
			double iw = image.getWidth();
			double ih = image.getHeight();

			AffineTransform t = new AffineTransform();
			double scale = scaleImage ? Math.min(w / iw, h / ih) : 1;
			if (scaleImage) {
				t.scale(scale, scale);
			}
			if (centerImage) {
				t.translate((w - iw * scale) / 2, (h - ih * scale) / 2);
			}

			g.drawImage(image, t, null);
		}

		// draw message
		if (message != null && message.length() > 0) {
			TextLayout text = new TextLayout(message, g.getFont().deriveFont(Font.BOLD), g.getFontRenderContext());
			Rectangle2D rc = text.getBounds();

			double x = w - rc.getX() - rc.getWidth();
			double y = h - rc.getY() - rc.getHeight();

			g.setPaint(Color.red);
			text.draw(g, (float)x, (float)y);
		}

		// blit the back-buffer to screen
		appletGraphics.drawImage(bufferImage, 0, 0, this);
	}

	/**
	 *  Overriden update() to avoid flicker.
	 */
	@Override public void update(Graphics g) {
		paint(g);
	}
}
