package sk.robotics.rlcs.web.editor;

import sk.robotics.rlcs.agentspace.Agent;

import java.util.*;
import java.lang.reflect.*;
import static java.lang.reflect.Modifier.*;

/**
 *  Agent description, as stored in XML, including field descriptions.
 */
public class AgentDesc implements Editor.ItemInfo {
	/** Information about a field in agent config. */
	public static class FieldDesc {
		/** Field name. */
		public String name;

		/** Field description. HTML formatted. */
		public String description;

		/** Block data type. Used only for block name fields. */
		public String blockType;

		/** Block access. Used only for block name fields. */
		public String blockAccess;
	}

	/** Agent config class fully qualified name. */
	public String configClassName;

	/** Display name. If null, is generated from the config class name,
	 *  by stripping standard prefix (sk.robotics.rlcs.agents.) and suffix ($Cfg). */
	public String displayName;
	private static final String STD_AGENT_PREFIX = "sk.robotics.rlcs.agents.";
	private static final String STD_AGENT_SUFFIX = "$Cfg";

	/** Agent description. HTML formatted. */
	public String description = "";

	/** List of fields. */
	public FieldDesc[] fields = new FieldDesc[0];

	// class object for configClassName, not storen in XML
	private transient Class configClass;

	// map for fkwyws by fiwld name, for faster access to fields, not stored in XML
	private transient Map<String, FieldDesc> fieldsMap = new HashMap<String, FieldDesc>();

	// reference to map of descriptions of all other agents, keyed by config class name (not stored in XML of course)
	private transient Map<String, AgentDesc> descMap;

	/** Initialization, to be called after loading from XML. */
	void init(Map<String, AgentDesc> descMap) {
		// reference to the descriptions map
		this.descMap = descMap;

		// agent display name
		if (displayName == null) {
			int trimbeg = configClassName.startsWith(STD_AGENT_PREFIX) ? STD_AGENT_PREFIX.length() : 0;
			int trimend = configClassName.endsWith(STD_AGENT_SUFFIX) ? STD_AGENT_SUFFIX.length() : 0;
			displayName = configClassName.substring(trimbeg, configClassName.length() - trimend);
		}

		// config class object
		try {
			configClass = Class.forName(configClassName);
		} catch (Exception e) {
			System.err.println("class '" + configClassName + "' not found");
		}

		// fields map
		for (FieldDesc field: fields) {
			fieldsMap.put(field.name, field);
		}
	}

	/**
	 *  Get field description for given field.
	 *  This method also searches for the field in parent classes, if needed.
	 */
	public FieldDesc getField(String fieldName) {
		// if we have description for the field, return it
		FieldDesc field = fieldsMap.get(fieldName);
		if (field != null) return field;

		// otherwise try to find class which declares this field, and get description from it
		if (configClass != null) {
			try {
				Class cl = configClass.getField(fieldName).getDeclaringClass();
				if (!cl.equals(configClass)) {
					AgentDesc agentDesc = descMap.get(cl.getName());
					return agentDesc == null ? null : agentDesc.getField(fieldName);
				}
			}
			catch (Exception e) {
				e.printStackTrace();
			}
			System.err.printf("field '%s' not found for class '%s'\n", fieldName, configClassName);
		}
		return null;
	}

	/**
	 *  Create new instance of agent config object. All fields will be cleared
	 *  to null, except those annotated with <code>Agent$Cfg$NoDefault</code>.
	 */
	public Agent.Cfg createEmptyCfg() {
		return createCfg(false);
	}

	/**
	 *  Create config instance with default field values.
	 *  Contains field filled with default values as initialized by constructor,
	 *  except fields annotated with <code>Agent$Cfg$NoDefault</code>, whose
	 *  values are cleared to null.
	 */
	public Agent.Cfg getDefaultCfg() {
		return createCfg(true);
	}

	// implementation for the 2 methods above
	private Agent.Cfg createCfg(boolean defaults) {
		try {
			// create instance
			Agent.Cfg cfg = (Agent.Cfg) configClass.newInstance();

			// clear either those fields annotated with NoDefault, or those others
			clearDefaults(cfg, defaults);

			return cfg;
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 *  Set default values for unset fields of agent config.
	 */
	public void setDefaults(Agent.Cfg cfg) {
		// check class off passed config
		if (!configClass.isInstance(cfg)) {
			throw new IllegalArgumentException("Passed config is not of class "+configClass.getName());
		}

		// create a default instance
		Object def = getDefaultCfg();

		// iterate through all fields
		for (Field f: configClass.getFields()) {
			// skip static and final fields
			int m = f.getModifiers();
			if (isStatic(m) || isFinal(m)) continue;

			// if the field is not set in the passed in cfg, fill with default value
			try {
				if (f.get(cfg) == null) {
					f.set(cfg, f.get(def));
				}
			} catch (IllegalAccessException e) {
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 *  Replace default field values with nulls in agent config.
	 */
	public void clearDefaults(Agent.Cfg cfg) {
		clearDefaults(cfg, false);
	}

	// implementation of the above, and also helper for createCfg
	private void clearDefaults(Agent.Cfg cfg, boolean clearNoDefaults) {
		// check class off passed config
		if (!configClass.isInstance(cfg)) {
			throw new IllegalArgumentException("Passed config is not of class "+configClass.getName());
		}

		try {
			// create a default instance to compare with
			// (we can't use getDefaultCfg here, that would cause infinite recursion)
			Object def = configClass.newInstance();

			// iterate through all fields
			for (Field f: configClass.getFields()) {
				// skip primitive fields
				if (f.getType().isPrimitive()) continue;

				// skip static and final fields
				int m = f.getModifiers();
				if (isStatic(m) || isFinal(m)) continue;

				// skip fields annotated (or not) with Agent.Cfg.NoDefault
				if (f.isAnnotationPresent(Agent.Cfg.NoDefault.class) != clearNoDefaults) continue;

				// if the field equals the default value, clear it
				Object val = f.get(cfg);
				if (val != null && val.equals(f.get(def))) {
					f.set(cfg, null);
				}
			}
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	// Editor.ItemInfo implementaition
	public String getDisplayName() { return displayName; }
	public String getDescription() { return description; }
	public boolean allowAdd() {
		// deny adding agents with config annotated as Hidden
		return configClass != null && !configClass.isAnnotationPresent(Agent.Hidden.class);
	}
}
