package sk.robotics.rlcs.web.editor;

import sk.robotics.rlcs.agentspace.Agent;

import java.util.*;
import java.util.regex.*;
import java.text.*;

import com.microstepmis.xplatform.X2O;
import org.jdom.Element;

/**
 *  Information about a behaviour (group of agents).
 *  Objects of this class are loaded from XML config,
 *  they are never created in any other way.
 */
public class BehaviourDesc implements Editor.ItemInfo {
	/** Behaviour name. Used for identification, as well as for informative
	 *  purposes, so it must be unique, but should be human readable. */
	public String name;

	/** HTML formatted description of the behaviour. */
	public String description;

	/** Array of agent configs, in XML representation.
	 *  The configs are parametrized, with parameters denoted as %x%.
	 *  Use generateAgentConfigs method, passing values of the parameters,
	 *  to get array of agent config objects. */
	public Element agents;

	/** Parameters descriptions, types,  and default values. */
	public ParamDesc[] params = new ParamDesc[0];
	public static class ParamDesc {
		/** Parameter name. */
		public String name;
		
		/** Parameter type (fully qualified class name). */;
		public String type;

		/** HTML formatted description of the parameter. */
		public String description;

		/** Default value for the parameter. */
		public Object def;

		/** Required flag. If set, and default value is not set,
		 *  user has to specify a value for the parameter.
		 *  Default value for this flag is true. */
		public boolean required = true;
	}

	// xml string representation of agents
	private transient String xmlAgentsStr;

	// map of param descriptions and default values, keyed by param name
	private transient Map<String, ParamDesc> paramDescMap = new HashMap<String, ParamDesc>();

	// initialization, to be called after loading from xml
	public void init() {
		// add xclass attribute to the xml element
		// (allows converting it to java object later)
		agents.setAttribute("xclass", "sk.robotics.rlcs.agentspace.Agent$Cfg[]");

		// init xml string representation of agents
		try { xmlAgentsStr = X2O.mapToXML(agents); }
		catch (Exception e) {
			throw new RuntimeException(e);
		}

		// init map of param descriptions
		for (ParamDesc param: params) {
			paramDescMap.put(param.name, param);
		}
	}

	/** Get ParamDesc object for a parameter. */
	public ParamDesc getParamDesc(String paramName) {
		ParamDesc paramDesc = paramDescMap.get(paramName);
		if (paramDesc == null) {
			throw new RuntimeException("Unknown parameter: '" + paramName + "'");
		}
		return paramDesc;
	}

	/** Get list of parameters, found in agents. */
	public Collection<String> getParamNames() {
		Collection<String> params = new HashSet<String>();
		for (ParamSubStr p: new ParamIterator(xmlAgentsStr, false)) {
			params.add(p.name);
		}
		return params;
	}

	/** Genarate array of agent config object, with filled parameter values. */
	public Agent.Cfg[] generateAgentConfigs(Map<String, Object> paramValues) {
		// generate xml string with parameter values filled
		StringBuilder builder = new StringBuilder();
		int pos = 0;
		for (ParamSubStr param: new ParamIterator(xmlAgentsStr, true)) {
			builder.append(xmlAgentsStr, pos, param.start);

			if (param.name.isEmpty()) {
				// no parameter, just a doubled % char
				builder.append('%');
			} else {
				// try to get parameter value from the map passed as parameter
				Object val = paramValues.get(param.name);

				// if the param is not in the passed map, use its default value, if it has one
				if (val == null) {
					ParamDesc paramDesc = paramDescMap.get(param.name);
					if (paramDesc == null) {
						throw new IllegalStateException("Undefined parameter '"+param.name+"'");
					}
					val = paramDesc.def;
				}

				// append the value to the output (if it is set)
				if (val != null) {
					builder.append(val);
				}
			}

			pos = param.end + 1;
		}
		if (pos < xmlAgentsStr.length()) {
			builder.append(xmlAgentsStr, pos, xmlAgentsStr.length());
		}

		// map from XML
		try { return (Agent.Cfg[]) X2O.mapFromXML(builder.toString()); }
		catch (Exception e) {
			System.err.println(builder);
			throw new RuntimeException(e);
		}
	}

	// classes used to iterate through parameters in the xml
	private static class ParamSubStr {
		private String name;
		private int start, end;
	}
	private static class ParamIterator implements Iterator<ParamSubStr>, Iterable<ParamSubStr> {
		public ParamIterator(String source, boolean includeEmpty) {
			this.source = source;
			this.includeEmpty = includeEmpty;
			this.param = findNext(0);
		}
		public Iterator<ParamSubStr> iterator() {
			// the Iterable and Iterator are probably really not meant
			// to be used like this, but its simple and works for us
			return this;
		}
		public ParamSubStr next() {
			ParamSubStr result = param;
			param = findNext(param.end + 1);
			return result;
		}
		public boolean hasNext() {
			return param != null;
		}
		public void remove() {
			throw new UnsupportedOperationException();
		}
		private ParamSubStr findNext(int from) {
			ParamSubStr param = new ParamSubStr();

			// find next parameter bounds
			for (;;from = param.end + 1) {
				// find parameter start
				param.start = source.indexOf('%', from);
				if (param.start == -1) return null;

				// find parameter end
				param.end = source.indexOf('%', param.start + 1);
				if (param.end == -1) {
					throw new IllegalArgumentException("Unmatched % character at position " + param.start);
				}

				// if the parameter name is empty, it is not a parameter, but a doubled % character,
				// so continue searching, unless we are told to include also empty-named params
				if (param.end > param.start + 1 || includeEmpty) break;
			};

			// check parameter name validity
			for (int i = param.start + 1; i < param.end; ++i) {
				if (!isCharOk(source.charAt(i))) {
					throw new IllegalArgumentException("Invalid character in parameter name at position " + i);
				}
			}

			// ok, we have next parameter
			param.name = source.substring(param.start + 1, param.end);
			return param;
		}
		private static boolean isCharOk(char c) {
			if (Character.isLetter(c)) return true;
			if (Character.isDigit(c)) return true;
			return c == '_';
		}
		private String source;
		private boolean includeEmpty;
		private ParamSubStr param;
	}

	// Editor.ItemInfo implementaition
	public String getDisplayName() { return name; }
	public String getDescription() { return description; }
	public boolean allowAdd() { return true; }
}
