package sk.robotics.rlcs.web.editor;

import sk.robotics.rlcs.agentspace.Scheduler;
import sk.robotics.rlcs.agentspace.Agent;

import java.util.*;

/**
 *  Config of running agents as used by Web Editor.
 *  It is an extension of Scheduler.Cfg, which supports behaviours in addition to agents.
 */
public class Config {

	// the config contains set of named groups
	public Group[] groups = new Group[0];

	/**
	 *  Class for storing one named logical group of agents and behaviours.
	 *  Agents and behaviours are displayed in the property sheet divided into these groups.
	 */
	public static class Group {
		/** Group name, as displayed in the property sheet. */
		public String name;

		/** Set of behaviours. */
		public Behaviour[] behaviours = new Behaviour[0];

		/** Set of agents. */
		public Agent.Cfg[] agents = new Agent.Cfg[0];
	}
		
	/**
	 *  Class for storing config for one behaviour.
	 *  Contains just the behaviour name, group name (as seen in editor),
	 *  and map of behaviour parameters.
	 */
	public static class Behaviour {
		/** Behaviour name, as defined in the behaviours XML file. */
		public String name;

		/** Map of behaviour parameter values. */
		public HashMap<String, Object> params = new HashMap<String, Object>();
	}

	/** Default ctor, needed for XML mapping. */
	public Config() {}

	/** Constructs Config from plain Scheduler.Cfg. */
	public Config(Scheduler.Cfg cfg) {
		groups = new Group[cfg.groups.length];
		for (int i = 0, n = cfg.groups.length; i < n; ++i) {
			groups[i] = new Group();
			groups[i].name = cfg.groups[i].name;
			groups[i].agents = cfg.groups[i].agents;
		}
	}

	/** Conversion to plain Scheduler.Cfg. Flattens behaviours to agent configs. */
	public Scheduler.Cfg asSchedulerCfg(Map<String, BehaviourDesc> behaviourDescriptions) {
		Scheduler.Cfg cfg = new Scheduler.Cfg();
		cfg.groups = new Scheduler.Cfg.AgentGroup[groups.length];

		for (int i = 0, n = groups.length; i < n; ++i) {
			Collection<Agent.Cfg> agents = new ArrayList<Agent.Cfg>(16);

			// flatten behaviours
			for (Behaviour beh: groups[i].behaviours) {
				BehaviourDesc behDesc = behaviourDescriptions.get(beh.name);
				agents.addAll(Arrays.asList(behDesc.generateAgentConfigs(beh.params)));
			}

			// copy raw agents
			agents.addAll(Arrays.asList(groups[i].agents));

			// construct the agent group for scheduler config
			cfg.groups[i] = new Scheduler.Cfg.AgentGroup();
			cfg.groups[i].name = groups[i].name;
			cfg.groups[i].agents = agents.toArray(new Agent.Cfg[0]);
		}

		return cfg;
	}
}
