package sk.robotics.rlcs.web.editor;

import sk.robotics.rlcs.agentspace.Agent;
import sk.robotics.rlcs.agentspace.Scheduler;

import com.microstepmis.xplatform.X2O;

import com.l2fprod.common.propertysheet.*;
import com.l2fprod.common.beans.editor.ComboBoxPropertyEditor;

import org.xngr.browser.editor.*;
import static org.xngr.browser.editor.Constants.*;

import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.*;

import java.net.*;
import java.applet.*;

/**
 *  Applet used to remotely control the Virtual Robotics Laboratory Control System.
 *
 *  Allows editing the XML configuration, either manually in syntax-highlighted editor,
 *  or visually using a property editor component. Also allows to query current
 *  configuration from running system, and also to update it (restart) with new config.
 */
public class Editor extends JApplet
{
	// default TCP port, where RLCS is listening
	// (default host is the one serving the page,
	//  and both are overridable by applet tag params)
	private static final int DEFAULT_PORT = 35759;

	// app entry point - testing only, this is an applet
	public static void main(String[] args) {
		// create a frame
		JFrame frame = new JFrame();

		// create the applet
		final Editor editor = new Editor();

		// when the frame gets closed, stop the applet, and exit the app
		frame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				editor.stop();
				System.exit(0);
			}
		});

		// applet stub, making the applet connect to localhost
		URL xurl = null;
		try { xurl = new URL("http://localhost/"); }
		catch (Exception e) {}
		final URL url = xurl;
		editor.setStub(new java.applet.AppletStub() {
			@Override public void appletResize(int w, int h) {}
			@Override public AppletContext getAppletContext() { return null; }
			@Override public URL getCodeBase() { return url; }
			@Override public URL getDocumentBase() { return url; }
			@Override public String getParameter(String name) { return null; }
			@Override public boolean isActive() { return true; }
		});

		// init the applet, and place it inside the frame
		frame.add(editor);
		editor.init();

		// show the frame
		frame.pack();
		frame.setLocationRelativeTo(null);
		frame.setVisible(true);

		// start the applet
		editor.start();
	}

	// ------------------------------------
	//  descriptions & behaviours
	// ------------------------------------

	/** Interface implemented by AgentDesc and BehaviourDesc, used by the Add dialog. */
	static interface ItemInfo {
		// name displayed in the choice component on top of add dialog
		public String getDisplayName();

		// descriptions displayed in the add dialog
		public String getDescription();

		// allow adding this item through the add dialog?
		public boolean allowAdd();
	}

	// agent descriptions, loaded from XML at runtime
	private AgentDesc[] agentDescriptions;

	// behaviour descriptions, loaded from XML at runtime
	private BehaviourDesc[] behaviourDescriptions;

	// maps of agent and behaviour descriptions keyed by config class name for agents
	// and by behaviour name for behaviours, for faster access to descriptions
	private Map<String, AgentDesc> agentDescMap = new HashMap<String, AgentDesc>();
	private Map<String, BehaviourDesc> behaviourDescMap = new HashMap<String, BehaviourDesc>();

	// ------------------------------------
	//  applet init
	// ------------------------------------

	// gui components
	private JTabbedPane tabs;
	private JTextComponent xmlSourceEditor;
	private PropertySheetPanel propertySheet;

	// RLCS host & port
	private String rlcsHost;
	private int rlcsPort;

	/**
	 *  Applet initialization.
	 */
	public void init() {
		// determine RLCS host and port
		rlcsHost = getParameter("host");
		if (rlcsHost == null || rlcsHost.length() == 0) {
			rlcsHost = getDocumentBase().getHost();
		}
		try { rlcsPort = Integer.parseInt(getParameter("port")); }
		catch (Exception e) { rlcsPort = DEFAULT_PORT; }

		// load agent descriptions and behaviours from XML
		try {
			agentDescriptions = (AgentDesc[]) X2O.mapFromXML(Editor.class.getResourceAsStream("descriptions.xml"));
			behaviourDescriptions = (BehaviourDesc[]) X2O.mapFromXML(Editor.class.getResourceAsStream("behaviours.xml"));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		// build maps of descriptions, and init all AgentDesc and BehaviourDesc objects
		for (AgentDesc agentDesc: agentDescriptions) {
			agentDescMap.put(agentDesc.configClassName, agentDesc);
			agentDesc.init(agentDescMap);
		}
		for (BehaviourDesc behDesc: behaviourDescriptions) {
			behaviourDescMap.put(behDesc.name, behDesc);
			behDesc.init();
		}

		// set layout
		setLayout(new BorderLayout());

		// create top panel with apply & load buttons
		JPanel toolbar = new JPanel();
		add(toolbar, BorderLayout.NORTH);
		toolbar.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.gridy = 0;
		c.weightx = 1;
		c.fill = GridBagConstraints.BOTH;
		JButton btn;
		toolbar.add(btn = new JButton("Load current config"), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) { actionLoadCurrent(); } });
		toolbar.add(btn = new JButton("Apply config"), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) { actionApply(); } });
		toolbar.add(btn = new JButton("Clear"), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) { setCurrentConfig(null); } });

		// create the tabbed pane
		tabs = new JTabbedPane();
		tabs.setModel(new DefaultSingleSelectionModel() {
			public void setSelectedIndex(int index) {
				if (tabSelectionChange(index))
					super.setSelectedIndex(index);
			}
		});
		add(tabs, BorderLayout.CENTER);

		// create the property editor pane
		tabs.add("Agents", propertySheet = new PropertySheetPanel());
		propertySheet.setDescriptionVisible(true);
		propertySheet.setToolBarVisible(false);
		propertySheet.setMode(PropertySheet.VIEW_AS_CATEGORIES);

		// create toolbar above the property editor
		toolbar = new JPanel();
		propertySheet.add(toolbar, BorderLayout.NORTH);
		toolbar.setLayout(new GridBagLayout());
		toolbar.add(btn = new JButton("Add agent..."), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) { actionAddAgent(); } });
		toolbar.add(btn = new JButton("Add behaviour..."), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) { actionAddBehaviour(); } });
		toolbar.add(btn = new JButton("Remove"), c);
		btn.addActionListener(new ActionListener() { public void actionPerformed(ActionEvent e) {
			int sel = propertySheet.getTable().getSelectedRow();
			Object item = propertySheet.getTable().getSheetModel().getObject(sel);
			Property prop = ((PropertySheetTableModel.Item)item).getProperty();
			propertySheet.removeProperty(prop);
		} });

		// enable the 'remove' button only if an agent or behaviour is selected
		final JButton btnRem = btn;
		btnRem.setEnabled(false);
		propertySheet.getTable().getSelectionModel().addListSelectionListener(new ListSelectionListener() {
			public void valueChanged(ListSelectionEvent evt) {
				int sel = propertySheet.getTable().getSelectedRow();
				if (sel == -1) return;
				Object item = propertySheet.getTable().getSheetModel().getObject(sel);
				Property prop = ((PropertySheetTableModel.Item)item).getProperty();
				btnRem.setEnabled(prop != null && prop.getParentProperty() == null);
			}
		});

		// create the XML source editor pane
		JScrollPane xmlScrollPane = new JScrollPane(xmlSourceEditor = new XmlEditorPane());
		xmlScrollPane.setPreferredSize(new Dimension(600, 500));
		tabs.add("XML Source", xmlScrollPane);
		xmlSourceEditor.getDocument().putProperty(PlainDocument.tabSizeAttribute, 2);
		xmlSourceEditor.setFont(new Font(Font.MONOSPACED, 0, 12));
		{
			XmlEditorKit kit = (XmlEditorKit) ((XmlEditorPane)xmlSourceEditor).getEditorKit();
			XmlContext ctx = kit.getContext();
			ctx.setAttributes(ELEMENT_NAME, new Color(0, 128, 128), Font.BOLD);
			//ctx.setAttributes(ELEMENT_PREFIX, Color.BLACK, Font.BOLD);
			ctx.setAttributes(ELEMENT_VALUE, Color.BLACK, Font.BOLD);
			ctx.setAttributes(ATTRIBUTE_NAME, new Color(0, 0, 128), Font.BOLD);
			//ctx.setAttributes(ATTRIBUTE_PREFIX, Color.BLACK, Font.BOLD);
			ctx.setAttributes(ATTRIBUTE_VALUE, new Color(255, 0, 255), Font.BOLD);
			ctx.setAttributes(NAMESPACE_NAME, Color.BLACK, Font.BOLD);
			//ctx.setAttributes(NAMESPACE_PREFIX, Color.BLACK, Font.BOLD);
			ctx.setAttributes(NAMESPACE_VALUE, Color.BLACK, Font.BOLD);
			ctx.setAttributes(ENTITY, Color.BLACK, Font.BOLD);
			ctx.setAttributes(COMMENT, Color.BLACK, Font.BOLD);
			ctx.setAttributes(DECLARATION, Color.BLACK, Font.BOLD);
			ctx.setAttributes(SPECIAL, Color.BLACK, Font.BOLD);
			ctx.setAttributes(STRING, Color.BLACK, Font.BOLD);
		}

		// setup undo/redo for the XML source editor
		final UndoManager undo = new UndoManager();
		xmlSourceEditor.getDocument().addUndoableEditListener(new UndoableEditListener() {
			public void undoableEditHappened(UndoableEditEvent evt) {
				undo.addEdit(evt.getEdit());
			}
		});
		xmlSourceEditor.getActionMap().put("Undo", new AbstractAction("Undo") {
			public void actionPerformed(ActionEvent evt) {
				try { if (undo.canUndo()) undo.undo(); }
				catch (CannotUndoException e) {}
			}
		});
		xmlSourceEditor.getActionMap().put("Redo", new AbstractAction("Redo") {
			public void actionPerformed(ActionEvent evt) {
				try { if (undo.canRedo()) undo.redo(); }
				catch (CannotRedoException e) {}
			}
		});
		xmlSourceEditor.getInputMap().put(KeyStroke.getKeyStroke("control Z"), "Undo");
		xmlSourceEditor.getInputMap().put(KeyStroke.getKeyStroke("control Y"), "Redo");
	}

	public void start() {
		// hack the propertyeditor to make the desription pane bigger initially
		// NB: just setting preferred size of the description pane works when running
		//     the editor as standalone app, but it does not work in applet in browser!
		BorderLayout layout = (BorderLayout)propertySheet.getLayout();
		JSplitPane splitter = (JSplitPane) layout.getLayoutComponent("Center");
		splitter.setDividerLocation(splitter.getHeight() - splitter.getDividerSize() - 100);

		// NB: don't load current config on start, because if RLCS
		//     does not run, the error messages are very disturbing
		//actionLoadCurrent();
	}

	// ----------------------------------------------
	//  transfering config to/from editor components
	// ----------------------------------------------

	/**
	 *  Tab selection change handler.
	 *  Returning false vetoes the change
	 */
	private boolean tabSelectionChange(int newIndex) {
		// get new tab index
		int oldIndex = tabs.getSelectedIndex();

		if (oldIndex == 0 && newIndex == 1) {
			// switch from property editor to xml editor
			Config cfg = getFromPropertyEditor();
			updateSourceEditor(cfg);
		}
		else if (oldIndex == 1 && newIndex == 0) {
			// switch from xml editor to
			Config cfg = getFromSourceEditor();
			if (cfg == null) return false;
			updatePropertyEditor(cfg);
		}

		return true;
	}

	/**
	 *  Get current config.
	 *  Returns null if there is an error, and also pops up an error message in this case.
	 */
	private Config getCurrentConfig() {
		switch (tabs.getSelectedIndex()) {
			case 0: return getFromPropertyEditor();
			case 1: return getFromSourceEditor();
			default:
				JOptionPane.showMessageDialog(this, "Bad tab selection!", "Error", JOptionPane.ERROR_MESSAGE);
				return null;
		}
	}

	/**
	 *  Set current config.
	 *  Updates currently displayed editor.
	 */
	private void setCurrentConfig(Config cfg) {
		if (cfg == null) {
			cfg = new Config();
		}
		switch (tabs.getSelectedIndex()) {
			case 0: updatePropertyEditor(cfg); break;
			case 1: updateSourceEditor(cfg); break;
			default:
				JOptionPane.showMessageDialog(this, "Bad tab selection!", "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 *  Validates a configuration. Also fills unset fields with default values.
	 *  Displays an error message, and returns false, if the validation fails.
	 *  Otherwise (if config is OK), it returns true.
	 */
	private boolean validateConfig(Config cfg) {
		for (Config.Group group: cfg.groups) {
			// validate agents
			for (Agent.Cfg agent: group.agents) {
				if (!validateAgentCfg(agent, null)) return false;
			}

			// validate behaviours
			for (Config.Behaviour beh: group.behaviours) {
				BehaviourDesc behDesc = behaviourDescMap.get(beh.name);
				if (behDesc == null) {
					JOptionPane.showMessageDialog(this, "Unknown behaviour: " + beh.name,
						"Error in config", JOptionPane.ERROR_MESSAGE);
					return false;
				}
				for (String paramName: behDesc.getParamNames()) {
					// check params, whethey they are filled if required, and of corrent type
					BehaviourDesc.ParamDesc paramDesc = behDesc.getParamDesc(paramName);
					Object paramValue = beh.params.get(paramName);
					if (paramValue == null) {
						paramValue = paramDesc.def;
					}
					if (paramValue == null && paramDesc.required) {
						JOptionPane.showMessageDialog(this,
							"Parameter '"+paramName+"' in behaviour '"+beh.name+"' must be filled.",
							"Error in config", JOptionPane.ERROR_MESSAGE);
						return false;
					}
					if (paramValue != null && !paramValue.getClass().getName().equals(paramDesc.type)) {
						JOptionPane.showMessageDialog(this,
							"Parameter '"+paramName+"' in behaviour '"+beh.name+"' must be of type "+paramDesc.type,
							"Error in config", JOptionPane.ERROR_MESSAGE);
						return false;
					}
				}
				for (Agent.Cfg agent: behDesc.generateAgentConfigs(beh.params)) {
					if (!validateAgentCfg(agent, beh)) return false;
				}
			}
		}
		return true;
	}
	private boolean validateAgentCfg(Agent.Cfg agent, Config.Behaviour beh) {
		// fill unset fields with default values
		AgentDesc desc = agentDescMap.get(agent.getClass().getName());
		if (desc == null) {
			JOptionPane.showMessageDialog(this,
				"Unknown agent config class: " + agent.getClass().getName(),
				"Error in config", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		desc.setDefaults(agent);

		if (!agent.enabled) return true;
		try { agent.validate(); }
		catch (Exception e) {
			e.printStackTrace();
			String agentName = agent.getClass().getName();
			if (agent.id != null) {
				agentName = agentName + "/" + agent.id;
			}
			if (beh != null) {
				agentName = agentName + " in behaviour " + beh.name;
			}
			JOptionPane.showMessageDialog(this,
				"Error in configuration of agent " + agentName + ":\n" + e,
				"Error in config", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		return true;
	}

	/**
	 *  Update the property editor pane with a new config.
	 */
	private void updatePropertyEditor(Config cfg) {
		// clear the property sheet
		for (Property property: propertySheet.getProperties()) {
			propertySheet.removeProperty(property);
		}

		// fill it
		for (Config.Group group: cfg.groups) {
			for (Agent.Cfg agentCfg: group.agents) {
				AgentDesc agentDesc = agentDescMap.get(agentCfg.getClass().getName());
				if (agentDesc == null) agentDesc = new AgentDesc();
				addAgentToPropertyEditor(agentDesc, agentCfg, group.name);
			}
			for (Config.Behaviour beh: group.behaviours) {
				addBehaviourToPropertyEditor(beh, group.name);
			}
		}
	}

	/**
	 *  Add one agent to the property editor.
	 */
	private void addAgentToPropertyEditor(AgentDesc agentDesc, Agent.Cfg config, String group) {
		// the config class
		Class cfgClass = config.getClass();

		// default config
		Object defaultConfig = agentDesc.getDefaultCfg();

		// the agent property (parent for real properties - the agent's config fields)
		DefaultProperty parentProperty = new DefaultProperty();
		parentProperty.setCategory(group);
		parentProperty.setType(cfgClass); // we just store it there, it doesn't do anything
		parentProperty.setName(cfgClass.getName());
		parentProperty.setDisplayName(agentDesc.displayName);
		parentProperty.setShortDescription(agentDesc.description);

		// hidden fields
		Set<String> hiddenFields = config.getHiddenFields();

		// config field properties
		for (Field field: cfgClass.getFields()) {
			// skip final, static, and hidden fields
			if (Modifier.isFinal(field.getModifiers())) continue;
			if (Modifier.isStatic(field.getModifiers())) continue;
			if (hiddenFields.contains(field.getName())) continue;

			// determine default value for the field
			Object defaultValue;
			try { defaultValue = field.get(defaultConfig); }
			catch (Exception e) {
				throw new RuntimeException(e);
			}

			// construct the property object
			DefaultProperty prop = new DefaultProperty();
			prop.setName(field.getName());
			prop.setDisplayName(field.getName());
			prop.setType(field.getType());
			prop.setEditable(true);
			registerPropertyEditor(prop.getType());
			try {
				Object value = field.get(config);

				// escape strings
				if (value != null && field.getType().equals(String.class)) {
					value = encodeString((String)value);
				}

				// booleans are special, they are edited with a checkbox,
				// so they cannot be unset, so replace null booleans with default value
				if (value == null && field.getType().equals(Boolean.class)) {
					value = defaultValue;
				}

				prop.setValue(value);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}

			// construct description
			StringBuilder txt = new StringBuilder(512);

			// basic description
			AgentDesc.FieldDesc fieldDesc = agentDesc.getField(field.getName());
			if (fieldDesc != null && fieldDesc.description != null) {
				txt.append("<p>").append(fieldDesc.description).append("<p>");
			}

			// field type
			txt.append("<i>Type:</i> ").append(field.getType().getName()).append("<br>");

			// block data type & access (present for block name fields only)
			if (fieldDesc != null && fieldDesc.blockType != null) {
				txt.append("<i>Block data type:</i> ").append(fieldDesc.blockType).append("<br>");
			}
			if (fieldDesc != null && fieldDesc.blockAccess != null) {
				txt.append("<i>Block access:</i> ").append(fieldDesc.blockAccess).append("<br>");
			}

			// allowed range
			if (field.isAnnotationPresent(Agent.Cfg.Limited.class)) {
				Agent.Cfg.Limited limits = field.getAnnotation(Agent.Cfg.Limited.class);
				txt.append("<i>Range:</i> ").append(limits.min()).append(" - ").append(limits.max()).append("<br>");
			}

			// required/optional field, or field with default value?
			// (the required flag does not make sense for fields with default value)
			if (defaultValue == null) {
				boolean req = field.isAnnotationPresent(Agent.Cfg.Required.class);
				txt.append("<i>Required:</i> ").append(req ? "yes" : "no").append("<br>");
			} else {
				txt.append("<i>Default value:</i> ").append(defaultValue).append("<br>");
			}

			// set property description
			if (txt.length() > 0) {
				prop.setShortDescription(txt.toString());
			}

			// add the field sub-property to the agent property
			parentProperty.addSubProperty(prop);
		}

		// add the agent property to the sheet
		propertySheet.addProperty(parentProperty);
	}

	/**
	 *  Add one behaviour to the property editor.
	 */
	private void addBehaviourToPropertyEditor(Config.Behaviour beh, String group) {
		BehaviourDesc behDesc = behaviourDescMap.get(beh.name);

		// the behaviour property (parent for real properties - the behaviour's parameters)
		DefaultProperty parentProperty = new DefaultProperty();
		parentProperty.setCategory(group);
		parentProperty.setName(beh.name);
		parentProperty.setType(null); // means 'behaviour, not agent'
		parentProperty.setDisplayName(beh.name);
		parentProperty.setShortDescription(behDesc.description);

		// behaviour params
		for (String paramName: behDesc.getParamNames()) {
			BehaviourDesc.ParamDesc paramDesc = behDesc.getParamDesc(paramName);

			// construct the property object
			DefaultProperty prop = new DefaultProperty();
			prop.setName(paramName);
			prop.setDisplayName(paramName);
			try { prop.setType(Class.forName(paramDesc.type)); }
			catch (ClassNotFoundException e) {
				throw new RuntimeException(e);
			}
			Object value = beh.params.get(paramName);
			if (prop.getType().equals(String.class) && value != null) {
				value = encodeString((String)value);
			}
			prop.setValue(value);
			prop.setEditable(true);
			registerPropertyEditor(prop.getType());

			StringBuilder desc = new StringBuilder();
			desc.append("<p>").append(paramDesc.description).append("<p>");
			desc.append("<i>Type:</i> ").append(prop.getType().getName()).append("<br>");
			if (paramDesc.def != null) {
				desc.append("<i>Default value:</i> ").append(paramDesc.def);
			} else {
				desc.append("<i>Required:</i> ").append(paramDesc.required ? "yes" : "no");
			}
			prop.setShortDescription(desc.toString());

			// add the param sub-property to the behaviour property
			parentProperty.addSubProperty(prop);
		}

		// add the behaviour property to the sheet
		propertySheet.addProperty(parentProperty);
	}

	/**
	 *  Register property editor for given property type, if not yet registered.
	 */
	private void registerPropertyEditor(Class type) {
		PropertyEditorRegistry editorRegistry = (PropertyEditorRegistry) propertySheet.getEditorFactory();

		// if this is an enum field, and a corresponding property editor
		// is not yet registered, create and register it
		if (type.isEnum() && editorRegistry.getEditor(type) == null) {
			Collection<ComboBoxPropertyEditor.Value> values = new ArrayList<ComboBoxPropertyEditor.Value>();
			values.add(new ComboBoxPropertyEditor.Value(null, "(not set)"));
			for (Object value: type.getEnumConstants()) {
				values.add(new ComboBoxPropertyEditor.Value(value, value));
			}

			ComboBoxPropertyEditor editor = new ComboBoxPropertyEditor();
			editor.setAvailableValues(values.toArray());
			editorRegistry.registerEditor(type, editor);
		}
	}

	/**
	 *  Get config from the property editor.
	 */
	private Config getFromPropertyEditor() {
		// commit current editing, if in progress
		propertySheet.getTable().commitEditing();

		// fill a map of groups
		class Group {
			Collection<Config.Behaviour> behaviours = new ArrayList<Config.Behaviour>();
			Collection<Agent.Cfg> agents = new ArrayList<Agent.Cfg>();
		}
		Map<String, Group> groups = new LinkedHashMap<String, Group>();

		for (Iterator it = propertySheet.propertyIterator(); it.hasNext();) {
			// the agent/behaviour property
			Property parentProp = (Property) it.next();

			// create group if not yet done
			Group grp = groups.get(parentProp.getCategory());
			if (grp == null) {
				groups.put(parentProp.getCategory(), grp = new Group());
			}

			// is it an agent or a behaviour?
			Class cfgClass = parentProp.getType();
			if (cfgClass == null) {
				// it is a behaviour
				Config.Behaviour beh = new Config.Behaviour();
				beh.name = parentProp.getName();
				beh.params = new HashMap<String, Object>();
				for (Property prop: parentProp.getSubProperties()) {
					if (prop.getValue() != null) {
						beh.params.put(prop.getName(), prop.getValue());
					}
				}
				grp.behaviours.add(beh);
			} else {
				// the property is an agent config, create empty agent config
				AgentDesc desc = agentDescMap.get(cfgClass.getName());
				Agent.Cfg cfg = desc.createEmptyCfg();

				// fill in fields
				for (Property prop: parentProp.getSubProperties()) {
					try {
						Field field = cfgClass.getField(prop.getName());
						Object value = prop.getValue();
						Class type = field.getType();
						if (type.equals(String.class) && value != null) {
							if (value.equals("")) {
								// treat empty strings as nulls
								value = null;
							} else {
								// decode escaped chars
								value = decodeString((String)value);
							}
						}
						field.set(cfg, value);
					} catch (Exception e) {
						throw new RuntimeException(e);
					}
				}

				// clear default values
				desc.clearDefaults(cfg);

				// insert into group (creating it if neccessary)
				grp.agents.add(cfg);
			}
		}

		// generate the configuration object
		Config cfg = new Config();
		cfg.groups = new Config.Group[groups.size()];
		int groupIndex = 0;
		for (Map.Entry<String, Group> entry: groups.entrySet()) {
			Config.Group grp = new Config.Group();
			grp.name = entry.getKey();
			grp.behaviours = entry.getValue().behaviours.toArray(new Config.Behaviour[0]);
			grp.agents = entry.getValue().agents.toArray(new Agent.Cfg[0]);
			cfg.groups[groupIndex++] = grp;
		}
		return cfg;
	}

	/**
	 *  Update the source editor with new config.
	 */
	private void updateSourceEditor(Config cfg) {
		if (cfg == null) cfg = new Config();
		try {
			xmlSourceEditor.setText(X2O.mapToXML(cfg));
			xmlSourceEditor.setCaretPosition(0);
		}
		catch (Exception e) {
			throw new RuntimeException(e); // this really shouldn't happen
		}
	}
	/**
	 *  Get config from the source editor.
	 */
	private Config getFromSourceEditor() {
		// check for empty editor, use default (empty) config in that case
		String xml = xmlSourceEditor.getText();
		if (xml.trim().length() == 0) {
			return new Config();
		}

		// map from xml
		Config result;
		try { result = (Config) X2O.mapFromXML(xml); }
		catch (Exception e) {
			e.printStackTrace();
			StringBuilder msg = new StringBuilder("The XML is malformed:");
			for (String line: e.toString().split(":")) {
				msg.append('\n').append(line);
			}
			JOptionPane.showMessageDialog(this, msg, "Oops!", JOptionPane.ERROR_MESSAGE);
			return null;
		}

		// clear default values
		for (Config.Group grp: result.groups) {
			for (Agent.Cfg cfg: grp.agents) {
				AgentDesc desc = agentDescMap.get(cfg.getClass().getName());
				if (desc == null) {
					JOptionPane.showMessageDialog(this, "Unknown agent config class "+cfg.getClass().getName(),
					                              "Oops!", JOptionPane.ERROR_MESSAGE);
					return null;
				}
				desc.clearDefaults(cfg);
			}
		}

		return result;
	}

	// encoding/decoding strings for displaying in the property editor
	// (show linefeeds as \n, carriage returns as \r, tabs as \t,
	//  all other characters below 0x20 as \xNN, and backslash as \\)
	public static String encodeString(String data) {
		StringBuilder builder = new StringBuilder((int)(data.length() * 1.1));
		for (int i = 0; i < data.length(); ++i) {
			char c = data.charAt(i);
			if      (c == '\\') builder.append("\\\\");
			else if (c >= 0x20) builder.append(c);
			else if (c == '\r') builder.append("\\r");
			else if (c == '\n') builder.append("\\n");
			else if (c == '\t') builder.append("\\t");
			else {
				builder.append("\\x");
				builder.append(c >= 0x10 ? '1' : '0');
				int cx = c & 0x0F;
				builder.append(cx >= 0x0A ? (char)('a' + cx - 0x0A) : (char)('0' + cx));
			}
		}
		return builder.toString();
	}
	public static String decodeString(String data) {
		StringBuilder builder = new StringBuilder(data.length());
		for (int i = 0; i < data.length(); ++i) {
			char c = data.charAt(i);
			if (c != '\\') {
				builder.append(c);
				continue;
			}
			if (++i < data.length()) {
				c = data.charAt(i);
				if      (c == 'r') builder.append('\r');
				else if (c == 'n') builder.append('\n');
				else if (c == 't') builder.append('\t');
				else if (c != 'x') builder.append(c);
				else if (i < data.length() - 2) {
					int n1 = Character.getNumericValue(data.charAt(i + 1)),
					    n2 = Character.getNumericValue(data.charAt(i + 2));
					if (n1 >= 0x0 && n1 <= 0xF && n2 >= 0x0 && n2 <= 0xF) {
						builder.append((char)((n1 << 1) | n2));
						i += 2;
					}
				}
			}
		}
		return builder.toString();
	}

	// ----------------------------------------------------------------
	//  actions in property editor (adding agents and behaviours)
	// ----------------------------------------------------------------

	private void actionAddAgent() {
		// show a dialog, select an agent
		AddDialog dialog = new AddDialog("Add agent", "Agent:", Arrays.asList(agentDescriptions));
		AgentDesc agentDesc = (AgentDesc) dialog.getResult();
		if (agentDesc == null) return;

		// add the agent
		try { addAgentToPropertyEditor(agentDesc, agentDesc.createEmptyCfg(), dialog.getGroup()); }
		catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this, e, "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	private void actionAddBehaviour() {
		// show a dialog, select a behaviour
		AddDialog dialog = new AddDialog("Add group of agents", "Behaviour:", Arrays.asList(behaviourDescriptions));
		BehaviourDesc behDesc = (BehaviourDesc) dialog.getResult();
		if (behDesc == null) return;

		// construct default behaviour object
		Config.Behaviour beh = new Config.Behaviour();
		beh.name = behDesc.name;

		// add the behaviour
		try { addBehaviourToPropertyEditor(beh, dialog.getGroup()); }
		catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this, e, "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 *  Add dialog.
	 *  Shows one text field (group name), a combobox choice of ItemInfo objects,
	 *  and a description pane, which shows description of selected item.
	 */
	private class AddDialog extends JDialog {
		// class for storing AgentDesc in JComboBox
		private class ComboItem {
			private ItemInfo itemInfo;
			public ComboItem(ItemInfo itemInfo) { this.itemInfo = itemInfo; }
			public ItemInfo getItemInfo() { return itemInfo; }
			public String toString() { return itemInfo.getDisplayName(); }
		}

		// components
		private JTextField textGroup;
		private JComboBox comboItems;
		private JEditorPane textDesc;

		// result
		private ItemInfo result;
		public ItemInfo getResult() { return result; }
		public String getGroup() { return textGroup.getText(); }

		public AddDialog(final String title, String itemsLabel, Collection<? extends ItemInfo> items) {
			// construct vector of items, directly displayable in a combo-box
			Vector<ComboItem> comboVector = new Vector<ComboItem>(items.size());
			for (ItemInfo item: items) {
				if (item.allowAdd()) {
					comboVector.add(new ComboItem(item));
				}
			}

			// setup layout
			setTitle(title);
			((JComponent)getContentPane()).setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
			setLayout(new GridBagLayout());
			GridBagConstraints c = new GridBagConstraints();
			c.anchor = GridBagConstraints.WEST;

			// group name
			c.gridy = 0;
			c.gridx = 0;
			c.insets.right = 4;
			add(new JLabel("Group name:"), c);
			c.insets.right = 0;
			c.gridx++;
			c.fill = GridBagConstraints.HORIZONTAL;
			c.weightx = 1;
			add(textGroup = new JTextField(), c);

			// choice box
			c.gridy++;
			c.gridx = 0;
			c.fill = GridBagConstraints.NONE;
			c.weightx = 0;
			c.insets.top = c.insets.bottom = 4;
			c.insets.right = 4;
			add(new JLabel(itemsLabel), c);
			c.insets.right = 0;
			c.gridx++;
			c.fill = GridBagConstraints.HORIZONTAL;
			c.weightx = 1;
			add(comboItems = new JComboBox(comboVector), c);
			c.insets.top = c.insets.bottom = 0;

			// description pane
			c.gridy++;
			c.gridx = 0;
			c.fill = GridBagConstraints.NONE;
			c.weightx = 0;
			add(new JLabel("Description:"), c);
			c.gridy++;
			c.gridwidth = 2;
			c.fill = GridBagConstraints.BOTH;
			c.weightx = 1;
			c.weighty = 1;
			add(new JScrollPane(textDesc = new JEditorPane()), c);
			textDesc.setEditable(false);
			textDesc.setContentType("text/html");

			// ok/cancel buttons
			Box buttonBox = Box.createHorizontalBox();
			c.gridy++;
			c.fill = GridBagConstraints.NONE;
			c.weightx = c.weighty = 0;
			c.anchor = GridBagConstraints.CENTER;
			c.insets.top = 4;
			add(buttonBox, c);
			JButton btnOk = new JButton("OK"),
			        btnCancel = new JButton("Cancel");
			buttonBox.add(btnOk);
			buttonBox.add(Box.createHorizontalStrut(4));
			buttonBox.add(btnCancel);

			// update the description on item selection
			comboItems.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (e.getStateChange() == ItemEvent.SELECTED) {
						ComboItem item = (ComboItem) e.getItem();
						textDesc.setText(item.getItemInfo().getDescription());
						textDesc.setCaretPosition(0); // scroll to top
					}
				}
			});
			ComboItem selected = (ComboItem) comboItems.getSelectedItem();
			if (selected != null) {
				textDesc.setText(selected.getItemInfo().getDescription());
				textDesc.setCaretPosition(0);
			}

			// ok/cancel clicks
			btnOk.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					ComboItem selected = (ComboItem) comboItems.getSelectedItem();
					if (selected == null) {
						JOptionPane.showMessageDialog(AddDialog.this, "Nothing selected.", title, JOptionPane.ERROR_MESSAGE);
					} else {
						result = selected.getItemInfo();
						setVisible(false);
					}
				}
			});
			ActionListener closeAction = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					setVisible(false);
				}
			};
			btnCancel.addActionListener(closeAction);
			getRootPane().registerKeyboardAction(closeAction,
				KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
				JComponent.WHEN_IN_FOCUSED_WINDOW);

			// size and show the dialog
			setPreferredSize(new Dimension(400, 400));
			pack();
			setLocationRelativeTo(Editor.this);
			setModal(true);
			setVisible(true);
		}
	}

	// -------------------------------------------------------------------
	//  global config actions (load current, apply, load file, save file)
	// -------------------------------------------------------------------

	private void actionLoadCurrent() {
		Socket sock = null;
		try {
			// connect to RLCS
			System.err.printf("Connecting to %s:%d...", rlcsHost, rlcsPort);
			sock = new Socket(rlcsHost, rlcsPort);

			// request configuration
			Writer writer = new OutputStreamWriter(sock.getOutputStream());
			writer.write("GETCFG\n");
			writer.flush();

			// read reply
			Reader reader = new BufferedReader(new InputStreamReader(sock.getInputStream()));
			Scheduler.Cfg cfg = (Scheduler.Cfg) X2O.mapFromXML(reader);

			// load the reply
			setCurrentConfig(new Config(cfg));

			// if the config is empty, display an info box
			if (cfg.groups.length == 0) {
				JOptionPane.showMessageDialog(this, "Current config is empty (no agents).");
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this, e, "Error", JOptionPane.ERROR_MESSAGE);
		}

		// close the socket
		try { if (sock != null) sock.close(); }
		catch (IOException e) {}
	}

	private void actionApply() {
		// get current config
		Config config = getCurrentConfig();
		if (config == null) return;

		// validate the config
		if (!validateConfig(config)) return;

		// determine RLCS host and port
		String host = getDocumentBase().getHost();
		int port = DEFAULT_PORT;
		try { port = Integer.parseInt(getParameter("port")); }
		catch (NumberFormatException e) {}
		catch (NullPointerException e) {}

		// this exception is thrown when we get unexpected or no reply from server
		class ServerReplyException extends Exception {
			public ServerReplyException(String msg) { super(msg); }
		}

		Socket sock = null;
		try {
			// connect to RLCS
			sock = new Socket(host, port);
			Writer writer = new BufferedWriter(new OutputStreamWriter(sock.getOutputStream()));

			// send 'put' command
			writer.write("PUTCFG\n");

			// send the config
			writer.write(X2O.mapToXML(config.asSchedulerCfg(behaviourDescMap)));

			// flush the writer, and shutdown output
			writer.flush();
			sock.shutdownOutput();

			System.err.println(X2O.mapToXML(config.asSchedulerCfg(behaviourDescMap)));

			// wait for reply
			BufferedReader reader = new BufferedReader(new InputStreamReader(sock.getInputStream()));
			sock.setSoTimeout(10000);
			String reply = reader.readLine();
			if (reply == null) {
				throw new ServerReplyException("No reply from server.");
			}
			if (reply.equals("ERR")) {
				throw new ServerReplyException("Error on server.");
			}
			else if (!reply.equals("OK")) {
				throw new ServerReplyException("Unexpected reply from server.");
			}

			// OK
			JOptionPane.showMessageDialog(this, "OK");
		}
		catch (Exception e) {
			if (!(e instanceof ServerReplyException)) {
				e.printStackTrace();
			}
			JOptionPane.showMessageDialog(this, e, "Error", JOptionPane.ERROR_MESSAGE);
		}

		// close the socket
		try { if (sock != null) sock.close(); }
		catch (IOException e) {}
	}
}
