// Contains a single class for calling functions of the robotjava dll
package sk.robotics.robot;

/** A wrapper class around java dll version of the library for controling robot (robotjava.dll); Use this to control local or remote robot directly from Java application.
 *  Most of the methods in this class call directly methods in the dll using JNI.
 *  You can use it to control a robot connected either to some local communication port
 *  (for example Bluetooth serial port - such as COM4), or a robot that is located
 *  at a remote computer and reachable in the local TCP/IP network or on the Internet.
 */
public class Robot {

    /** do not use (internal id) */
    private int id = 0; //internal id which identifies this object for dll

    /** Constructs a new robot control object.
     *  It will be destroyed either by explicitly calling destructor(), which is a recommended method that
     *  attempts to close the communication port, or when the object is collected by the garbage collector.
     *  The conctructor throws Exception in case the pool of robot objects is full.
     *  @param showMsgBoxes  use 1, if you do not mind messageboxes telling you about problems/errors, otherwise use 0
     *  @see #destructor
     */
    public Robot(int showMsgBoxes) throws Exception
    {
        id = constructor(showMsgBoxes);
        if (id < 0) throw new Exception("Robot.java: Robot(): Too many objects in robot object pool");
    }

    /** Standard finalize method, deletes the robot control object, not for explicit use. */
    protected void finalize() throws Throwable
    {
        destructor();
        super.finalize();
    }

    /** Internal native constructor, not for explicit use. */
    public native short constructor(int showMsgBoxes);

    /** Explicit destructor, you can call it when you stop using the object.
      * However, it is your responsibility to forget all references to this object. */
    public native void destructor();


    ///////////////////////////////////////////////////////////////////////////////////////
    /// HERE START THE MESSAGE HANDLERS
    ///////////////////////////////////////////////////////////////////////////////////////

    /** Move some steps forward.
     *  @param a  number of small steps to go forward
     *  @see #bk
     *  @see #lt
     *  @see #rt
     *  @see #isdone
     */
    public native void fd(int a);
    /** Move some steps backwards.
     *  @param a  number of small steps to go back
     *  @see #fd
     *  @see #lt
     *  @see #rt
     *  @see #isdone
     */
    public native void bk(short a);
    /** Turn some steps left. 720 steps should correspond 360 degrees.
     *  @param a  number of small steps to turn left
     *  @see #fd
     *  @see #bk
     *  @see #rt
     *  @see #isdone
     */
    public native void lt(short a);
    /** Turn some steps right. 720 steps should correspond 360 degrees.
     *  @param a  number of small steps to turn right
     *  @see #fd
     *  @see #bk
     *  @see #lt
     *  @see #isdone
     */
    public native void rt(short a);
    /** Start moving both wheels forward. Use l = 0, r = 0 to stop moving.
     *  @param l  new speed for the left wheel
     *  @param r  new speed for the right wheel
     *  @see #bkspeed
     *  @see #ltspeed
     *  @see #rtspeed
     *  @see #getactivity
     */
    public native void fdspeed(short l, short r);
    /** Start moving both wheels backwards. Use l = 0, r = 0 to stop moving.
     *  @param l  new speed for the left wheel
     *  @param r  new speed for the right wheel
     *  @see #fdspeed
     *  @see #ltspeed
     *  @see #rtspeed
     *  @see #getactivity
     */
    public native void bkspeed(short l, short r);
    /** Start moving right (that is left wheel will move backwards, while right wheel will move forward). Use l = 0, r = 0 to stop moving.
     *  @param l  new speed for the left wheel
     *  @param r  new speed for the right wheel
     *  @see #fdspeed
     *  @see #bkspeed
     *  @see #ltspeed
     *  @see #getactivity
     */
    public native void rtspeed(short l, short r);
    /** Start moving left (that is right wheel will move backwards, while left wheel will move forward). Use l = 0, r = 0 to stop moving.
     *  @param l  new speed for the left wheel
     *  @param r  new speed for the right wheel
     *  @see #fdspeed
     *  @see #bkspeed
     *  @see #rtspeed
     *  @see #getactivity
     */
    public native void ltspeed(short l, short r);
    /** Pen up. This command will instruct the robot to raise the pen and stop drawing.
     *  @see #pd
     */
    public native void pu();
    /** Pen down. This command will instruct the robot to put its pen down and stop drawing.
     *  @see #pu
     */
    public native void pd();
    /** Set the current condition.
     *  The condition will be satisfied, when the current sensory readings S will satisfy the condition:
     *  ((S XOR xorbits) OR orbits) > 0) || ((S XOR xorbits) AND andbits) == andbits).
     *  This commands only sets the condition for future use until reset by another call to condition().
     *  Use xorbits = 0, orbits = 0, andbits = 0 to reset.
     *  @param xorbits  the 6-bit XOR mask
     *  @param orbits   the 6-bit OR mask
     *  @param andbits  the 6-bit AND mask
     *  @see #satisfied
     *  @see #getsensors
     */
    public native void condition(short xorbits, short orbits, short andbits);
    /** Pings the robot.
     *  @see #connect
     *  @return 0, if the robot cannot be reached or was not connected properly. Otherwise returns 1.
     */
    public native short isalive();
    /** Determines whether the last command preceded by a conditional prefix was interrupted by a condition that was satisfied, or if the command finished properly.
     *  @see #condition
     *  @return 1, if the command was interrupted, otherwise returns 0.
     */
    public native short satisfied();
    /** Determines a 6-bit binary vector of sensory readings. Each bit corresponds to a single bit.
     *  @return current state of sensors
     */
    public native short getsensors();
    /** Determined the current voltage of the battery in deciVolts. When the voltage falls below 5.3, the robot
     *  will stop functioning.
     *  @return the current battery voltage level
     */
    public native short getvoltage();
    /** Determine what is the robot doing: executing position command (fd, bk, rt, lt), speed command (fdspeed, bkspeed, rtspeed, ltspeed), sitting idle, or being in error.
     *  @return 1, if robot is executing speed command;<br> 2, if the robot is executing position command;<br> 3, if the robot is sitting idle;<br> -1, if error occured and the robot cannot be reached
     */
    public native short getactivity();
    /** Determine if the position command has finished.
     *  @return 1, if the command has finished, otherwise returns 0.
     */
    public native short isdone();
    /** Connect to the robot at given port or IP address. Note that the user() must be called first.
     *  @param port communication port, such as COM4, or IP-address, such as 158.195.89.120
     *  @see #user
     *  @return 1, if connected successfully, otherwise 0 */
    public native short connect(String port);
    /** Set communication parameters. Should be called before connect().
     *  The meaning of these communication parameters is explained by Microsoft(TM) at:
     *  <a href="http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/dcb_str.asp">http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/dcb_str.asp</a>.
     *  Use -1 for the arguments where you wish to use the default values.
     */
    public native void setcomparam(long baud, short fparity, short ctsflow, short dsrflow, short dtrcontrol, short dsrsensitivity, short continueonxoff, short foutx, short finx, short rtscontrol, short bytesize, short parity, short stopbits);
    /** Returns version string of the robot firmware.
     *  @return typically 4-character version string
     */
    public native String getversion();
    /** Make the robot available on the network and start accepting network connections, or stop listning.
     *  Must be called only after successful connect().
     *  @param state  a string equal to "on" (start listening) or "off" (stop listening)
     *  @see #connect
     *  @return 1, if robot network server could be started, otherwise returns 0
     */
    public native short server(String state);
    /** Determine the robot number.
     *  It can be used to determine the robot internal number.
     *  @see #connect
     *  @return robot number
     */
    public native short getnum();
    /** Determine the robot local communication port name or IP address.
     * @see #connect
     * @return name of the local communication port where the robot is connected, or ip-address if we are connected to a remote robot.
     */
    public native String getport();
    /** Send message to another network user. Messages are not shown, if the object was constructed with parameter = 0, or if messages "off" was called.
     *  @param user  name of the remote user
     *  @param msg   text message to be shown in the message box on the screen of the remote user
     *  @see #messages
     *  @see #Robot
     *  @return 1, if the message was successfully sent (which does not mean that it was successfully received).
     */
    public native short message(String user, String msg);
    /** Toggle displaying of the message boxes containing informative or error messages, or messages from remote users.
     *  @param state  must be either "on" or "off
     */
    public native void messages(String state);
    /** Set the admin password for this robot network server. Unless this method has been called at the server, the password argument in the kickout() call can be ignored.
     *  @param password  a new network password
     */
    public native void password(String password);
    /** Lock or unlock the robot for exclusive access. After calling exclusive("on"), other users will not be able to
     *  controll the robot, until exclusive("off") will be called. The exclusive access can be terminated by administrator
     *  using kickout() call. By default, all users can access the robot simultaneously. The user is responsible to call
     *  exclusive("off") as soon as it stopped using the robot. Users can negotiate the robot use by sending messages each to other.
     *  @param state  a string containing "on" or "off"
     *  @see #logoff
     *  @see #message
     *  @return string "ok", if exclusive lock was perfomed, "err" if there is some error, or the name of the user who has already locked the robot, if the robot is locked.
     */
    public native String exclusive(String state);
    /** This call is the last possibility for the administrator to unlock a robot that has been locked by some user who forgot to unlock it.
     *  User performing kickout() will acquire exclusive lock for the robot. It must release it by exclusive("off") if it does not plan to use it.
     *  @param password the administrator password. It is verified only if it was configured on the server using password() command.
     *  @see #exclusive
     */
    public native void logoff(String password);
    /** Setup the user name. This call is obligatory before connecting to the robot. The password should be "turtle" unless the user file is in use.
     *  User performing kickout() will acquire exclusive lock for the robot. It must release it by exclusive("off") if it does not plan to use it.
     *  @param password the user password
     *  @see #userfile
     *  @see #connect
     */
    public native void user(String name, String password);
    /** Configure the file where the user names and scrambled passwords are stored and read from. Applies only to the
     *  local robot or the robot server from its console.
     *  @param filename fully qualified file path to the data file with the user names and passwords. The program must have read and write access to this file and the folder it is contained in.
     *  @see #connect
     *  @see #user
     *  @see #adduser
     *  @see #deluser
     *  @see #listusers
     */
    public native void userfile(String filename);
    /** Add user to the local user file.
     * @param name name of the user
     * @param passw for this user
     * @see #user
     * @see #deluser
     * @see #listusers
     * @see #userfile
     * @return 1, if the operation was successful, otherwise returns 0.
     */
    public native short adduser(String name, String passw);
    /** Delete user from the local user file.
     * @param name name of the user
     * @see #user
     * @see #adduser
     * @see #listusers
     * @see #userfile
     * @return 1, if the operation was successful, otherwise returns 0.
     */
    public native short deluser(String name);
    /** Determine the list of the local users.
     * @see #user
     * @see #adduser
     * @see #deluser
     * @see #userfile
     * @return a string containing a space-separated list of local users listed in the userfile
     */
    public native String listusers();
    /** Setup network port number. Use this call before connect() to remote robot server, if it is
     *  using a non-standard port. Use it on the server before calling server("on").
     * @param num network port number, use -1 to use the default port number
     * @see #connect
     * @see #server
     */
    public native void portnumber(long num);
    /** Wait until position command completes or timeout occures. This method will
     * block your application until the position command completes. However, if
     * the given number of milliseconds pass, the call will return anyway.
     * @param time number of milliseconds to wait
     * @see #isdone
     * @see #fd
      * @see #bk
      * @see #lt
      * @see #rt
     * @return 1, if the position command completed, and 0, if timeout occured.
     */
    public native short wait(short time);
    /** Determine the current position of the robot based on odometry locally.
     * When the robot is connected, it is assumed that it is located at point [0 0]
     * and heading towards [0 1].
     * @param reset allows to reset the robot position to the origin [0 0 0]
     * @return string of three space-separated double-precision floating-point numbers "x y heading"
     */
    public native String pos(short reset);
    /** Determine the current position of the robot based on odometry on server (applies only when connected to server).
     * The global position is updated at the server. Consequently, all clients connecting to the same server share
     * the global odometry. It is possible to reset the global odometry with the reset argument, but it by convention
     * it is not allowed, only in special applications (i.e. when all [your] clients can be prepared for this happening).
     * @param reset allows to reset the robot position to the origin [0 0 0]
     * @return string of three space-separated double-precision floating-point numbers "x y heading"
     */
    public native String globalpos(short reset);
    /** Move some steps forward with specific left and right wheel speeds (draw circle arc)
     *  @param l speed of the left wheel (0 - 99)
     *  @param r speed of the right wheel (0 - 99)
     *  @param steps number of steps of the faster wheel
     *  @see #bkx
     *  @see #ltx
     *  @see #rtx
     */
    public native void fdx(short l, short r, short steps);
    /** Move some steps backwards with specific wheel speeds.
     *  @param l  speed of the left wheel (0 - 99)
     *  @param r  speed of the right wheel (0 - 99)
     *  @param steps number of steps (of the faster wheel)
     *  @see #fdx
     *  @see #ltx
     *  @see #rtx
     */
    public native void bkx(short l, short r, short steps);
    /** Move some steps left with specific wheel speeds.
     *  @param l  speed of the left wheel (0-99)
     *  @param r  speed of the right wheel (0-99)
     *  @param steps number of steps (of the faster wheel)
     *  @see #fdx
     *  @see #bkx
     *  @see #rtx
     */
    public native void ltx(short l, short r, short steps);
    /** Move some steps right with specific wheel speeds.
     *  @param l  speed of the left wheel (0-99)
     *  @param r  speed of the right wheel (0-99)
     *  @param steps number of steps (of the faster wheel)
     *  @see #fdx
     *  @see #bkx
     *  @see #ltx
     */
    public native void rtx(short l, short r, short steps);
    /** Calibrate the robot.
     *  @param d1 calibration parameter 1 (see ReadMe.txt for details)
     *  @param d2 calibration parameter 2 (see ReadMe.txt for details)
     *  @param d3 calibration parameter 3 (see ReadMe.txt for details)
     *  @param d4 calibration parameter 4 (see ReadMe.txt for details)
     *  @param d5 calibration parameter 5 (see ReadMe.txt for details)
     *  @param d6 calibration parameter 6 (see ReadMe.txt for details)
     *  @param d7 calibration parameter 7 (see ReadMe.txt for details)
     *  @see #drawcalib
     */
    public native void calibrate(String d1, String d2, String d3, String d4, String d5, String d6, String d7);
    /** Draw the calibration image.
     *  Note that the robot has to be placed so that it has more than a meter of empty space to the right and to the front
     *  and 20cm to the left and behind. See ReadMe.txt for more details.
     *  @see #calibrate
     */
    public native void drawcalib();

    /** Switch synchroneous/asynchroneous position commands mode.
     *  In the default asynchroneous mode, position commands (fd,bk,lt...) return
     *  immediatelly. In the synchroneous mode, they return only after the
     *  robot has completed the movement (i.e. the same as if wait(0) was
     *  called after them).
     *  @param state  "on" turns the synchroneous mode on, "off" switches back to asynchroneous mode
     *  @see #wait
     */
    public native void alwayswait(String state);


    /** Start or stop recording commands to a file or memory.
     *  The only parameter should contain the full path to the logfile,
     *  or one of the following keywords:
     *    "memory" - to start recording to memory,
     *    "off" - to stop recording,
     *  @param name  name of the logfile, or one of the keywords described above.
     *  @see #replay
     */
    public native void logfile(String name);


    /** Start replay of previously recorded sequence.
     *  The only parameter should contain the full path to the logfile,
     *  or one of the following keywords:
     *    "memory" - to start recording to memory,
     *    "stop" - to stop recording,
     *    "status" - to return the number of commands remaining,
     *    "timeleft" - to return the number of milliseconds remaining
     *    "skip" - to skip over the next command
     *    "<any number>" - to skip the given number of milliseconds
     *  @param filename  name of the logfile, or one of the keywords described above.
     *  @return 1 - if operation successful, 0 - if operation unsuccessful
     *  @see #logfile
     */
    public native long replay(String filename);

    /** Control robot gripper, if installed.
     *  The following commands are recognized:
     *    "up" - move the gripper to its highest position,
     *    "down" - move the gripper to its bottom position,
     *    "open" - open the gripper completely,
     *    "close" - close the gripper to its most closed position
     *    "upwards" - move the gripper a bit upwards
     *    "downwards" - move the gripper a bit downwards
     *    "tighten" - close the gripper a little bit
     *    "loose" - open the gripper a little bit
     *  @param cmd  one of the above strings
     */
    public native void gripper(String cmd);

    /** Move the robot to given coordinates.
     *  @param x x-coordinate of destination robot location
     *  @param y y-coordiante of destination robot location
     *  @see #pos
     */
    public native void moveto(short x, short y);

    /** Draw text from current turtle location (into right and forward quadrant) using specified font.
     *  The font must be in the Borland CHR stroked fonts format. (for example C:\TP55\LITT.CHR).
     *  These fonts can be scaled by any size. Size 10 is  normal size.
     *  @param caption string containing the text to be drawn
     *  @param size multiplicative scale factor
     *  @param fontfile full path to the file containing the font
     *  @see #textsize
     *  @return 1 - if operation successful, 0 - if font not found or some problem occurs
     */
    public native short drawtext(String caption, short size, String fontfile);

    /** Returns the dimensions of a rectangle, which will be occupied by a drawn text, if drawtext() is called with the same parameters.
     *  @see #drawtext
     *  @return string containing the dimensions in format "width height"
     */
    public native String textsize(String caption, short size, String fontfile);

    /** Plays tone with given frequency [Hz] and duration [ms]. Also turns off beeping of the turtle. Use negative frequency to turn beeping on again.
     *  @return 1 if ok, 1 if parameters out of range
     */
    public native short tone(float freq, float duration);


    /** Determines whether the connection has been lost.
     *  @return true, if connection has not been lost, otherwise returns false.
     */
    public native boolean connected();

    /** Set position and heading of the robot (applies only to simulated robots).
     *  @param x the x-coordinate in pixels (0-9999)
     *  @param y the y-coordinate in pixels (0-9999)
     *  @param heading heading is in tenths of degree, 0 means facing upwards, incresing in clockwise-direction
     */
    public native void setsimpos(short x, short y, short heading);

    static
    {
      System.loadLibrary("robotjava");
    }

}
