package sk.robotics.simulator;

import java.lang.*;
import java.util.*;
import java.util.zip.*;
import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.color.*;
import java.awt.font.*;

/**
 *  Animator for Simulator of Virtual Robotics Laboratory.
 *
 *  This class generates an animation from data sent by the simulator.
 * <p>
 *  Ctor takes a Runnable object as parameter, whose run() method is called
 *  whenever a change ocurs. Animator provides a paint() method, which can
 *  be used then by the Runnable object to generate an image of the arena.
 * <p>
 *  This class is used by:<br>
 *   - Viewer - applet which displays the animation on-screen<br>
 *   - ImgGen - generates an image repeatedly or on request<br>
 */

public class Animator
{
	// timeout settings
	final static int  RECV_TIMEOUT = 1000; // timeout for receive() blocking
	final static long PING_PERIOD  = 1000; // ping frequency
	final static long ECHO_TIMEOUT = 5000; // timeout, after which without an echo we delete all turtles

	// number of remembered ping timestamps (see pingTimeStamps)
	final static int PING_STAMPS_LIMIT = 10;

	// turtle geometry
	final static int TURTLE_DIAMETER_MM = 210; // turtle diameter in millimeters
	final static int TURTLE_WHEELDST_MM = 180; // wheel distance in millimeters
	final static int TURTLE_TURN360_STP = 2880; // number of steps for 360-rotation
	final static int WHEEL_DIAMETER_MM  = 50; // this is only a guess!!
	final static double TURTLE_STEP_MM  = TURTLE_WHEELDST_MM * Math.PI / TURTLE_TURN360_STP;
	final static double TURTLE_DIAMETER = TURTLE_DIAMETER_MM / TURTLE_STEP_MM; // turtle diameter in steps
	final static double TURTLE_RADIUS   = TURTLE_DIAMETER / 2; // turtle radius in steps
	final static double WHEEL_DISTANCE  = TURTLE_WHEELDST_MM / TURTLE_STEP_MM; // wheel distance in steps
	final static double WHEEL_DIAMETER  = WHEEL_DIAMETER_MM / TURTLE_STEP_MM; // diameter of a wheel

	// laboratory size
	final static int LAB_WIDTH_MM  = 1500; // laboratory width in millimeters
	final static int LAB_HEIGHT_MM = 1200; // laboratory height in millimeters
	final static double LAB_WIDTH  = LAB_WIDTH_MM  / TURTLE_STEP_MM; // in steps
	final static double LAB_HEIGHT = LAB_HEIGHT_MM / TURTLE_STEP_MM; // in steps

	// desk image resolution
	final static int DESK_IMG_WIDTH  = LAB_WIDTH_MM;
	final static int DESK_IMG_HEIGHT = LAB_HEIGHT_MM;

	// robot sensors
	final static Point2D.Double[] SENSOR_POS = {
		new Point2D.Double( 90 / TURTLE_STEP_MM, -35 / TURTLE_STEP_MM),
		new Point2D.Double(  0,                  -97 / TURTLE_STEP_MM),
		new Point2D.Double(-90 / TURTLE_STEP_MM, -35 / TURTLE_STEP_MM),
		new Point2D.Double(-90 / TURTLE_STEP_MM,  35 / TURTLE_STEP_MM),
		new Point2D.Double(  0,                   97 / TURTLE_STEP_MM),
		new Point2D.Double( 90 / TURTLE_STEP_MM,  35 / TURTLE_STEP_MM) };

	// colours
	final static Color DESK_BACKGROUND = Color.lightGray;
	final static Color LINE_COLOR = Color.darkGray; // color for 1 bits in the desk bitmap
	final Color OUTER_BACKGROUND; // set by ctor

	// the listener (the object which actually draws the generated animation)
	Runnable listener;

	// real pixel-size of the animation (set in ctor)
	// (fullWidth & fullHeight is size of the applet/image/whatever,
	//  while deskWidth & deskWidth is size of the lab, scaled to fit
	//  the full size, but retaining aspect ratio of the real lab)
	int fullWidth, fullHeight, deskWidth, deskHeight;

	// simulator host and port (also set in ctor)
	InetAddress simulatorAddr;
	int simulatorPort = 35753;

	// sockets used to communicate with the simulator
	// (only dedicated threads may receive data through these sockets,
	//  but any thread may send, if it locks corresponding monitor)
	Socket tcpSocket;
	DatagramSocket udpSocket;
	Object tcpSendMonitor, udpSendMonitor;

	// received, but not yet processed movement datagrams
	SortedMap movements = new TreeMap(); // map<timestamp, collection<movement-data>>

	// delay (in milliseconds), used to make movement smoother
	int delay = 1; // small delay saves CPU time

	// simulator's timestamp of last already processed movement
	// (access to this variable is synchronized with monitor of the map above)
	long lastMovement = Long.MIN_VALUE;

	// difference between simd timestamps, und local system timestamps
	// (will be computed from echo packets)
	long timeDiff = Long.MAX_VALUE; // MAX_VALUE means uninitialized
	Object timeDiffMonitor = new Object();

	// timestamps of last few sent pings, we compute the difference above
	// from these upon receipt of corresponding echos
	SortedMap pingTimeStamps = new TreeMap(); // map<ping-id, timestamp>

	// the actual positions
	Map turtles = new TreeMap(); // map<turtle-id, turtle-data>

	// coordinates transformation
	AffineTransform coordsTransformCentered, // used to paint into the applet
	                coordsTransformTopLeft;  // used to paint into the merged image
	double scale;

	// desk background image (original, and with merged finished shapes)
	BufferedImage deskImage, mergedImage;

	// drawn shapes (lines & arcs)
	Map drawingShapes = new TreeMap(); // map<id, drawing-shape> - currently drawn shapes
	Collection finishedShapes = new Vector(); // collection<drawing-shape> - shapes to merge into background
	Collection mergedShapes = new Vector(); // collection<drawing-shape> - shapes merged into background

	// termination flag
	boolean termFlag = false;
	Object termMonitor = new Object();
	boolean terminating()
	{
		synchronized (termMonitor)
		{
			return termFlag;
		}
	}

	// structures
	class TurtleData
	{
		int id, type, shape;
		Color color;
		double xpos, ypos, heading;
		Shape identify;
	}
	class MovementData
	{
		long timestamp;
		TurtleData turtle;
		DrawingShape shape;
	}
	class DrawingShape
	{
		int id;
		Shape shape;
		Color color;
		Stroke stroke;
	}

	// overridable logger
	public static class Logger {
		public void log(Object message) {
			System.err.println(message);
		}
		public void log(Exception e) {
			e.printStackTrace(System.err);
		}
		public void log(Exception e, Object message) {
			log(message);
			log(e);
		}
	}
	private Logger logger;

	// our threads
	private Collection threads = new ArrayList(4);

	//------------------------------------------------------------------
	//    ctor & dtor
	//------------------------------------------------------------------

	public Animator(String host, Runnable listener, int width,
	                int height, Color bg, Logger logger) // bg - outer background (outside of desk)
	{
		// init logger
		this.logger = (logger == null) ? new Logger() : logger;

		// init simulator host and port
		String[] hostParts = host.split(":");
		if (hostParts.length > 2) {
			throw new IllegalArgumentException("Invalid host");
		}
		try { simulatorAddr = InetAddress.getByName(hostParts[0]); }
		catch (UnknownHostException e) {
			throw new IllegalArgumentException("Unknown host");
		}
		if (hostParts.length == 2) {
			simulatorPort = Integer.parseInt(hostParts[1]);
		}

		// init image params
		fullWidth     = width;
		fullHeight    = height;
		OUTER_BACKGROUND = bg;

		// the listener (viewer or img-generator)
		this.listener = listener;

		// setup coordinates transformation (scale to fit, but lock aspect ratio)
		scale = Math.min(fullWidth / (double)LAB_WIDTH, fullHeight / (double)LAB_HEIGHT);
		deskWidth  = (int)Math.round(scale * LAB_WIDTH);
		deskHeight = (int)Math.round(scale * LAB_HEIGHT);
		coordsTransformTopLeft = new AffineTransform();
		coordsTransformTopLeft.scale(scale, -scale); // minus -> swap the y-axis
		coordsTransformTopLeft.translate(0, -LAB_HEIGHT); // normalize after the scale
		coordsTransformCentered = new AffineTransform();
		coordsTransformCentered.translate((fullWidth - deskWidth) / 2.0, (fullHeight - deskHeight) / 2.0); // centre
		coordsTransformCentered.concatenate(coordsTransformTopLeft);

		// init desk image (it is scaled to real pixel-size of the animation)
		deskImage = new BufferedImage(deskWidth, deskHeight, BufferedImage.TYPE_INT_RGB);
		Graphics2D g = deskImage.createGraphics();
		g.setPaint(DESK_BACKGROUND);
		g.fill(new Rectangle(0, 0, deskWidth, deskHeight));
		resetMergedImage();

		// init the sockets
		tcpSocket = null;
		tcpSendMonitor = new Object();
		try {
			udpSocket = new DatagramSocket();
			udpSocket.connect(simulatorAddr, simulatorPort);
			udpSocket.setSoTimeout(RECV_TIMEOUT);
		} catch (SocketException e) {
			throw new RuntimeException("Failed to create the UDP socket!");
		}
		udpSendMonitor = new Object();

		threads.add(new TCPReceiverThread());
		threads.add(new UDPReceiverThread());
		threads.add(new PingerThread());
		threads.add(new DelayThread());
	}

	public void start() {
		for (Iterator it = threads.iterator(); it.hasNext();) {
			Thread thread = (Thread) it.next();
			thread.start();
		}
	}

	public void destroy()
	{
		// set termination flag, and wake all threads
		synchronized (termMonitor)
		{
			termFlag = true;
		}
		synchronized (tcpSendMonitor)
		{
			if (tcpSocket != null)
				try { tcpSocket.close(); }
				catch (Exception e) {}
		}
		synchronized (udpSendMonitor)
		{
			if (udpSocket != null)
				try { udpSocket.close(); }
				catch (Exception e) {}
		}
		synchronized (movements)
		{
			movements.notify();
		}

		// interrupt all threads
		for (Iterator it = threads.iterator(); it.hasNext();) {
			Thread thread = (Thread) it.next();
			thread.interrupt();
		}
	}

	private void resetMergedImage()
	{
		AffineTransformOp op = new AffineTransformOp(new AffineTransform(), AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		mergedImage = op.filter(deskImage, mergedImage);
	}
	
	//------------------------------------------------------------------
	//    runtime configuration
	//------------------------------------------------------------------

	public void setDelay(int millis)
	{
		synchronized (movements)
		{
			delay = Math.max(millis, 1); // minimal delay 1 - saves CPU time
			movements.notify();
		}
	}

	public int getDelay()
	{
		synchronized (movements)
		{
			return delay;
		}
	}

	public Rectangle2D getDeskBounds()
	{
		return new Rectangle2D.Double((fullWidth - deskWidth) / 2.0,
			(fullHeight - deskHeight) / 2.0, deskWidth, deskHeight);
	}

	/** Get IDs of turtles on desk. */
	public int[] getTurtleIds()
	{
		synchronized (turtles)
		{
			int[] result = new int[turtles.size()];
			int idx = 0;
			for (Iterator it = turtles.values().iterator(); it.hasNext();
			     result[idx++] = ((TurtleData)it.next()).id);
			return result;
		}
	}

	/** Get positions of turtles on desk. Returne array of 3-tuple (x,y,h) arrays.
	 *  The returned position coordinates are in millimetres. */
	public double[][] getTurtlePositions() {
		synchronized (turtles) {
			double[][] result = new double[turtles.size()][3];
			int idx = 0;
			for (Iterator it = turtles.values().iterator(); it.hasNext(); ++idx) {
				TurtleData t = (TurtleData) it.next();
				result[idx][0] = t.xpos * TURTLE_STEP_MM;
				result[idx][1] = t.ypos * TURTLE_STEP_MM;
				result[idx][2] = t.heading;
			}
			return result;
		}
	}

 	//------------------------------------------------------------------
	//    the pinger thread
 	//------------------------------------------------------------------
 
	class PingerThread extends Thread
 	{
		public PingerThread()
		{
			super("Animator/PingerThread");
			setDaemon(true);
		}
 
		public void run()
		{
			for (long pingId = 1;; ++pingId)
			{
				try
				{
					// check for termination
					if (terminating()) break;

					// get current timestamp
					long timestamp = System.currentTimeMillis();
 
					// construct the datagram
					String data = "ping\n" + pingId + "\n";
					DatagramPacket datagram = new DatagramPacket(data.getBytes(), data.length());

					// add a record into last pings map
					// (removing oldest record if the map is full)
					synchronized (pingTimeStamps)
					{
						pingTimeStamps.put(new Long(pingId), new Long(timestamp));
						if (pingTimeStamps.size() > PING_STAMPS_LIMIT)
							pingTimeStamps.remove(pingTimeStamps.firstKey());
					}
 
					// send the ping datagram to simulator
					udpSocket.send(datagram);
 
					// sleep one second
					Thread.sleep(PING_PERIOD);
				}
				catch (InterruptedException e)
				{
					// no problem
					continue;
				}
				catch (Exception e)
				{
					logger.log("Error in Pinger thread: " + e);
				}
			}
 		}
 	}
 
	//------------------------------------------------------------------
	//    update on turtle move
	//------------------------------------------------------------------

	void moveTurtle(MovementData data)
	{
		// update the turtles & shapes maps
		Integer turtleId = new Integer(data.turtle.id);
		Integer finishedShape = null;
		synchronized (turtles)
		{
			TurtleData t = (TurtleData)turtles.get(turtleId);
			if (t == null)
				turtles.put(turtleId, data.turtle);
			else
			{
				t.xpos    = data.turtle.xpos;
				t.ypos    = data.turtle.ypos;
				t.heading = data.turtle.heading;
				if (t.shape != 0 && t.shape != data.turtle.shape)
					finishedShape = new Integer(t.shape);
				t.shape   = data.turtle.shape;
			}
		}
		if (data.shape != null || finishedShape != null)
		{
			synchronized (drawingShapes)
			{
				if (data.shape != null)
					drawingShapes.put(new Integer(data.shape.id), data.shape);
				if (finishedShape != null) {
					Object shape = drawingShapes.remove(finishedShape);

					// a clear-desk command could have cleared the shapes map
					if (shape != null) {
						finishedShapes.add(shape);
					}
				}
			}
		}

		// repaint
		listener.run();
	}

	void addTurtle(TurtleData data)
	{
		// update the turtles map
		Integer id = new Integer(data.id);
		synchronized (turtles)
		{
			TurtleData t = (TurtleData)turtles.get(id);
			if (t == null)
				turtles.put(id, data);
			else
			{
				// we already received newer position of this robot through UDP
				t.type  = data.type;
				t.color = data.color;
			}
		}

		// repaint
		listener.run();
	}

	//------------------------------------------------------------------
	//    parsing of received data
	//------------------------------------------------------------------

	TurtleData ParseTurtleData(DataInput data)
		throws IOException
	{
		TurtleData t = new TurtleData();
		t.id      = Integer.parseInt(data.readLine());
		t.type    = Integer.parseInt(data.readLine());
		t.color   = Color.decode("#" + data.readLine());
		t.xpos    = Double.parseDouble(data.readLine());
		t.ypos    = Double.parseDouble(data.readLine());
		t.heading = Double.parseDouble(data.readLine());
		t.shape   = Integer.parseInt(data.readLine());
		return t;
	}

	MovementData ParseMovementData(DataInput data)
		throws IOException
	{
		// read turtle number
		MovementData m = new MovementData();
		m.turtle = new TurtleData();
		m.turtle.id = Integer.parseInt(data.readLine());

		// read timestamp
		m.timestamp = Long.parseLong(data.readLine()) * 1000; // reads seconds part
		m.timestamp += Long.parseLong(data.readLine()); // reads milliseconds part

		// read turtle position & heading
		m.turtle.xpos    = Double.parseDouble(data.readLine());
		m.turtle.ypos    = Double.parseDouble(data.readLine());
		m.turtle.heading = Double.parseDouble(data.readLine());

		// read drawing shape
		String shapetype = data.readLine();
		if (!shapetype.equals("noshape"))
			m.shape = ParseDrawingShape(shapetype, data);
		m.turtle.shape = m.shape == null ? 0 : m.shape.id;
		return m;
	}

	DrawingShape ParseDrawingShape(String type, DataInput data)
		throws IOException
	{
		if (!type.equals("line") && !type.equals("arc"))
			throw new IOException("invalid shape: " + type);

		DrawingShape s = new DrawingShape();

		// read basic shape values: id, color, pen-width
		s.id     = Integer.parseInt(data.readLine());
		s.color  = Color.decode(data.readLine());
		s.stroke = new BasicStroke((float)(Double.parseDouble(data.readLine()) / scale));

		if (type.equals("line"))
		{
			// read line values: 2 end points
			double[] d = new double[4];
			for (int i = 0; i < 4; d[i++] = Double.parseDouble(data.readLine()));
			s.shape = new Line2D.Double(d[0], d[1], d[2], d[3]);
		}
		else if (type.equals("arc"))
		{
			// read arc values: center, radius, angle, extent
			double[] d = new double[5];
			for (int i = 0; i < 5; d[i++] = Double.parseDouble(data.readLine()));
			s.shape = new Arc2D.Double(d[0] - d[2], d[1] - d[2], 2 * d[2],
				2 * d[2], d[3] * 180 / Math.PI - 90, d[4] * 180 / Math.PI, Arc2D.OPEN);
		}
		return s;
	}

	//------------------------------------------------------------------
	//    the TCP receiver thread
	//------------------------------------------------------------------

	class TCPReceiverThread extends Thread
	{
		public TCPReceiverThread()
		{
			super("Animator/TCPReceiverThread");
			setDaemon(true);
		}

		public void run()
		{
			boolean unreachable = false;
			try
			{
				for (;;)
				{
					// if we are not connected, connect now
					synchronized (tcpSendMonitor) // exclude writing threads
					{
						// check for termination
						if (terminating()) break;

						if (tcpSocket == null)
						{
							try
							{
								// unfortunately Socket.connect() is Java 1.4+ only
								tcpSocket = new Socket(simulatorAddr, simulatorPort);
								unreachable = false;
							}
							catch (Exception e)
							{
								if (!unreachable)
									logger.log("simulator unreachable");
								unreachable = true;
								try { Thread.sleep(1000); }
								catch (InterruptedException ex) {}
								continue;
							}
							try
							{
								// send magic packet & local UDP port
								OutputStream out = tcpSocket.getOutputStream();
								out.write(("view\n" + udpSocket.getLocalPort() + "\n").getBytes());

								// receive actual robot positions
								DataInput in = new DataInputStream(tcpSocket.getInputStream());
								Map unfinishedShapes = new HashMap(); // map<shape-id, turtle>
								for (;;)
								{
									String magic = in.readLine();
									if (magic.equals("end")) break;
									if (!magic.equals("turtle"))
										throw new IOException("invalid turtle data: " + magic);

									TurtleData t = ParseTurtleData(in);
									unfinishedShapes.put(new Integer(t.shape), t);
									addTurtle(t);
								}

								// receive desk image
								DeskUpdate(in);

								// receive drawn shapes
								for (;;)
								{
									String magic = in.readLine();
									if (magic.equals("end")) break;

									DrawingShape shape = ParseDrawingShape(magic, in);
									synchronized (drawingShapes)
									{
										Integer id = new Integer(shape.id);
										if (drawingShapes.containsKey(id))
											// we already received newer version of this shape through UDP
											continue;
										TurtleData turtle = (TurtleData)unfinishedShapes.get(id);
										if (turtle != null && turtle.shape == id.intValue())
											// this is a currently-drawn shape
											drawingShapes.put(id, shape);
										else // this is an already finished shape
											finishedShapes.add(shape);
									}
									listener.run();
								}
							}
							catch (Exception e)
							{
								logger.log(e, "simulator communication error");

								try { tcpSocket.close(); }
								catch (Exception ex) {}
								tcpSocket = null;
								try { Thread.sleep(1000); }
								catch (Exception ex) {}
								continue;
							}
							logger.log("connected to simulator");
						}
					}

					// wait for data
					try
					{
						// read a magic packet
						DataInput in = new DataInputStream(tcpSocket.getInputStream());
						String magic = in.readLine();

						// identify the magic packet
						if (magic == null)
						{
							// EOF, probably disconnected
							throw new EOFException("EOF");
						}

						// process incoming desk images
						if (magic.equals("desk"))
						{
							DeskUpdate(in);
							continue;
						}

						// process clear-shapes commands
						if (magic.equals("clear"))
						{
							synchronized (deskImage)
							{
								resetMergedImage();
								
								synchronized (drawingShapes)
								{
									drawingShapes.clear();
									finishedShapes.clear();
									mergedShapes.clear();
								}
							}
							listener.run();
							continue;
						}

						// process new-turtle notifications
						if (magic.equals("turtle"))
						{
							addTurtle(ParseTurtleData(in));
							continue;
						}

						// process turtle-removed notifications
						if (magic.equals("remove"))
						{
							Integer turtleId = new Integer(in.readLine());
							synchronized (turtles)
							{
								turtles.remove(turtleId);
							}
							listener.run();
							continue;
						}

						// unknown magic packet
						logger.log("invalid command: " + magic);
					}
					catch (Exception e)
					{
						logger.log("simulator connection closed");
						logger.log("TCPReceiverThread error: " + e);
						//logger.log(e);

						// simulator disconnected, remove all turtles & shapes
						synchronized (tcpSendMonitor)
						{
							try { tcpSocket.close(); }
							catch (Exception ex) {}
							tcpSocket = null;
						}
						synchronized (turtles)
						{
							turtles.clear();
						}
						synchronized (drawingShapes)
						{
							drawingShapes.clear();
							finishedShapes.clear();
						}
						listener.run();
					}
				}
			}
			finally
			{
				synchronized (tcpSendMonitor)
				{
					if (tcpSocket != null)
						try { tcpSocket.close(); }
						catch (Exception e) {}
					tcpSocket = null;
				}
			}
		}
	}

	//------------------------------------------------------------------
	//    the UDP receiver thread
	//------------------------------------------------------------------

	class UDPReceiverThread extends Thread
	{
		public UDPReceiverThread()
		{
			super("Animator/UDPReceiverThread");
			setDaemon(true);
		}

		public void run()
		{
			DatagramPacket datagram = new DatagramPacket(new byte[512], 512);
			long lastEcho = System.currentTimeMillis(); // last received echo
			long fastestEcho = Long.MAX_VALUE; // fastest echo so far (used to compute timeDiff)
			boolean udpError = false; // UDP connectivity problems
			for (;;)
			{
				// read new turtle position & heading
				try
				{
					// check for termination
					if (terminating()) break;

					// receive a datagram
					for (;;)
					{
						try { udpSocket.receive(datagram); } // blocks for max 1 second
						catch (InterruptedIOException e)
							// java 1.4+ throws SocketTimeoutException on timeout, but it is derived
							// from InterruptedIOException (throwed by lower java versions),
							// so this should work for all versions
						{
							// check for simulator time-out
							if (lastEcho + ECHO_TIMEOUT < System.currentTimeMillis() && !udpError)
							{
								udpError = true;
								logger.log("UDP connectivity problem");
							}
							continue; // continue listening
						}
						catch (SocketException e)
						{
							// we wanted to catch PortUnreachableException (derived from SocketException),
							// but it is java 1.4+ only, so hack it like this
							if (e.getClass().getName().equals("java.net.PortUnreachableException"))
								continue; // no problem, try again
							// if this is not a PortUnreachableException, pretend we didn't catch it
							else throw e;
						}
						break; // we have a datagram
					}

					// identify the datagram
					DataInput data = new DataInputStream(new ByteArrayInputStream(datagram.getData(), 0, datagram.getLength()));
					String magic = data.readLine();
					if (magic == null)
					{
						logger.log("received empty datagram");
						continue;
					}

					// process echo datagram
					if (magic.equals("echo"))
					{
						// get current time
						long currTime = System.currentTimeMillis();

						// read ping id from the echo
						long pingId = Long.parseLong(data.readLine());

						// read simulator's timestamp from the echo
						long timestamp = Long.parseLong(data.readLine()) * 1000; // reads seconds part
						timestamp += Long.parseLong(data.readLine()); // reads milliseconds part

						// if this is the fastest echo so far, update timestamp diff
						Long pingTime;
						synchronized (pingTimeStamps)
						{
							pingTime = (Long)pingTimeStamps.get(new Long(pingId));
						}
						if (pingTime != null && currTime - pingTime.longValue() < fastestEcho)
						{
							fastestEcho = currTime - pingTime.longValue();
							//logger.log("fastest echo: " + fastestEcho);

							// approximate time of the echo-send time as right between
							// ping-send time and echo-receive time (now)
							long echoTime = (pingTime.longValue() + currTime) / 2;
							synchronized (timeDiffMonitor)
							{
								timeDiff = echoTime - timestamp;
								//logger.log("time diff: " + timeDiff);
							}
							// time-diff changed, wake the delay thread
							synchronized (movements)
							{
								movements.notify();
							}
						}

						// update timestamp of last echo
						lastEcho = currTime;

						// imform about recovered UDP connectivity
						if (udpError)
						{
							udpError = false;
							logger.log("UDP connectivity recovered");
						}

						continue; // continue listening
					}

					// process move datagram
					if (magic.equals("move"))
					{
						MovementData move = ParseMovementData(data);

						// add this movement into list of movements
						// (the delay thread will manage it further)
						synchronized (movements)
						{
							// only add this movement, if we didn't process a newer one yet
							if (move.timestamp >= lastMovement)
							{
								Long tm = new Long(move.timestamp);
								Vector v = (Vector)movements.get(tm);
								if (v == null)
								{
									v = new Vector();
									v.add(move);
									movements.put(tm, v);
								}
								else v.add(move);
								movements.notify();
							}
							//else logger.log("dropping move datagram");
						}
						continue; // continue listening
					}

					// unknown datagram
					logger.log("received unknown datagram: " + magic);
				}
				catch (Exception e)
				{
					logger.log("UDPReceiverThread error: " + e);
					//logger.log(e);
				}
			}
		}
	}

	//------------------------------------------------------------------
	//    delay thread
	//------------------------------------------------------------------

	class DelayThread extends Thread
	{
		public DelayThread()
		{
			super("Animator/DelayThread");
			setDaemon(true);
		}

		public void run()
		{
			Vector moves;
			for (;;)
			{
				try
				{
					synchronized (movements)
					{
						for (;;)
						{
							// check for termination
							if (terminating()) return;

							// if there is movement to process, wait
							while (movements.isEmpty())
							{
								movements.wait();
								if (terminating()) return; // we have to check again, because there is another wait() below
							}

							// get simulator timestamp of first movement in the set
							Long simulatorTime = (Long)movements.firstKey();

							// convert it to local system timestamp
							long localTime;
							synchronized (timeDiffMonitor)
							{
								if (timeDiff == Long.MAX_VALUE) localTime = Long.MAX_VALUE;
								else localTime = simulatorTime.longValue() + timeDiff + delay;
							}

							// if the time is in future, wait
							long currentTime = System.currentTimeMillis();
							if (localTime > currentTime)
							{
								movements.wait(localTime - currentTime);
								continue; // we might have received an older movement in the meantime
							}

							// get the movement data
							moves = (Vector)movements.remove(simulatorTime);

							// remember simulator timestamp of last processed movement
							lastMovement = simulatorTime.longValue();
							break;
						}
					}

					// move the turtles
					for (Iterator it = moves.iterator(); it.hasNext(); moveTurtle((MovementData)it.next()));
					//logger.log("draw movements: " + moves.size() + " (queued: " + remaining + ")");
				}
				catch (InterruptedException e)
				{
					// no problem, just continue...
				}
				catch (Exception e)
				{
					logger.log("Error in DelayThread: " + e.toString());
					//logger.log(e);
				}
			}
		}
	}

	//------------------------------------------------------------------
	//    desk update
	//------------------------------------------------------------------

	void DeskUpdate(DataInput in)
		throws IOException
	{
		// receive checksum & compressed length
		long checksum = Long.parseLong(in.readLine(), 16);
		int length = Integer.parseInt(in.readLine());

		// receive the bitmap
		byte[] compressed = new byte[length];
		in.readFully(compressed);

		// decompress
		in = new DataInputStream(new InflaterInputStream(new ByteArrayInputStream(compressed)));
		byte[] bitmap = new byte[(DESK_IMG_WIDTH * DESK_IMG_HEIGHT + 7) / 8];
		in.readFully(bitmap);

		// verify checksum
		CRC32 crc = new CRC32();
		crc.update(bitmap);
		if (crc.getValue() != checksum)
			throw new IOException("checksum error");

		// create image
		BufferedImage img = new BufferedImage(DESK_IMG_WIDTH, DESK_IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
		int colBg   = DESK_BACKGROUND.getRGB(),
			colLine = LINE_COLOR.getRGB();
		for (int y = 0, bit = 0; y < DESK_IMG_HEIGHT; ++y)
			for (int x = 0; x < DESK_IMG_WIDTH; ++x, ++bit)
				img.setRGB(x, y, (bitmap[bit / 8] & (1 << (bit % 8))) == 0 ? colBg : colLine);

		// update the desk image
		synchronized (deskImage)
		{
			double dx = deskWidth  / (double)DESK_IMG_WIDTH,
				   dy = deskHeight / (double)DESK_IMG_HEIGHT;
			AffineTransformOp op = new AffineTransformOp(AffineTransform.getScaleInstance(dx, dy),
														 AffineTransformOp.TYPE_BICUBIC);
			op.filter(img, deskImage);
			
			// also update the merged image
			resetMergedImage();
			
			// mark all merged shapes to be merged again
			synchronized (drawingShapes)
			{
				finishedShapes.addAll(mergedShapes);
				mergedShapes.clear();
			}
		}

		// repaint
		listener.run();
	}

	//------------------------------------------------------------------
	//    painting the arena
	//------------------------------------------------------------------

	public final static int PAINT_ANTIALIAS = 0x01;
	public final static int PAINT_SENSORS   = 0x02;
	public final static int PAINT_WHEELS    = 0x04;
	public final static int PAINT_IDENTIFY  = 0x08;

	public void paint(Graphics2D g, int flags)
	{
		// setup anti-aliasing
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, (flags & PAINT_ANTIALIAS) != 0
			? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF);
		
		// draw desk background
		Rectangle2D rectDesk = getDeskBounds();
		synchronized (deskImage)
		{
			// if there are any pending finished shapes, merge them into the background
			synchronized (drawingShapes)
			{
				if (!finishedShapes.isEmpty())
				{
					Graphics2D gb = mergedImage.createGraphics();
					gb.setRenderingHints(g.getRenderingHints());
					gb.setTransform(coordsTransformTopLeft);
					for (Iterator it = finishedShapes.iterator(); it.hasNext();)
					{
						DrawingShape s = (DrawingShape)it.next();
						gb.setPaint(s.color);
						gb.setStroke(s.stroke);
						gb.draw(s.shape);
					}
					mergedShapes.addAll(finishedShapes);
					finishedShapes.clear();
				}
			}
			
			// draw the background
			g.drawImage(mergedImage, (int)rectDesk.getX(), (int)rectDesk.getY(), null);
		}

		// draw outer background
		if (OUTER_BACKGROUND != null)
		{
			g.setPaint(OUTER_BACKGROUND);
			if (fullWidth > deskWidth)
			{
				g.fill(new Rectangle2D.Double(0, 0, rectDesk.getX(), fullHeight));
				g.fill(new Rectangle2D.Double(rectDesk.getMaxX(), 0, rectDesk.getX(), fullHeight));
			}
			if (fullHeight > deskHeight)
			{
				g.fill(new Rectangle2D.Double(0, 0, fullWidth, rectDesk.getY()));
				g.fill(new Rectangle2D.Double(0, rectDesk.getMaxY(), fullWidth, rectDesk.getY()));
			}
		}

		// setup transformation
		// (but take into account, that there could already be a transformation
		//  associated with the graphics, so concatenate the one of ours)
		AffineTransform oldTransform = g.getTransform();
		g.transform(coordsTransformCentered);
		AffineTransform defTransform = g.getTransform();

		// draw shapes
		synchronized (drawingShapes)
		{
			for (Iterator it = drawingShapes.values().iterator(); it.hasNext();)
			{
				DrawingShape s = (DrawingShape)it.next();
				g.setPaint(s.color);
				g.setStroke(s.stroke);
				g.draw(s.shape);
			}
		}

		// draw turtles
		Collection turtlesCopy;
		synchronized (turtles)
		{
			turtlesCopy = new ArrayList(turtles.values());
		}
		for (Iterator it = turtlesCopy.iterator(); it.hasNext();)
		{
			TurtleData turtle = (TurtleData)it.next();

			// setup transformation
			g.setTransform(defTransform);
			g.translate(turtle.xpos, turtle.ypos);
			g.rotate(-turtle.heading);

			// draw turtle body
			double d = TURTLE_DIAMETER,
			       r = d / 2; // radius
			g.setPaint(Color.black);
			g.fill(new Ellipse2D.Double(-r, -r, d, d));

			// draw heading marks
			double d1 = r / 3, // diameter of front heading mark
			       d2 = r * 3 / 5, // diameter of back heading mark
			       r1 = d1 / 2, // radius of back heading mark
			       r2 = d2 / 2; // radius of back heading mark
			g.setPaint(turtle.color == null ? Color.black : turtle.color);
			g.fill(new Ellipse2D.Double(-r1, -r1 + r/2, d1, d1));
			g.fill(new Ellipse2D.Double(-r2, -r2 - r/2, d2, d2));

			// draw sensors
			if ((flags & PAINT_SENSORS) != 0)
			{
				double sd = r  / 6,
				       sr = sd / 2;
				g.setPaint(Color.red);
				for (int i = 0; i < 6; ++i)
					g.fill(new Ellipse2D.Double(SENSOR_POS[i].getX() - sr, SENSOR_POS[i].getY() - sr, sd, sd));
			}

			// draw wheels
			if ((flags & PAINT_WHEELS) != 0)
			{
				double wd = r  / 8; // wheel width
				g.setPaint(Color.blue);
				g.fill(new RoundRectangle2D.Double(-WHEEL_DISTANCE/2 - wd/2, -WHEEL_DIAMETER/2, wd, WHEEL_DIAMETER, wd, wd));
				g.fill(new RoundRectangle2D.Double( WHEEL_DISTANCE/2 - wd/2, -WHEEL_DIAMETER/2, wd, WHEEL_DIAMETER, wd, wd));
			}

			// identify
			if ((flags & PAINT_IDENTIFY) != 0)
			{
				if (turtle.identify == null)
				{
					TextLayout txt = new TextLayout(Integer.toString(turtle.id),
					                                new Font(null, Font.BOLD, 100),
					                                new FontRenderContext(null, true, true));
					Rectangle2D rect = txt.getBounds();
					double tx = rect.getX(), ty = rect.getY(),
					       tw = rect.getWidth(), th = rect.getHeight();

					AffineTransform transform = new AffineTransform();
					double sc = d * 0.7 / th;
					transform.scale(sc, -sc);
					transform.translate(-tx - tw/2, -ty - th/2);

					turtle.identify = txt.getOutline(transform);
				}

				g.rotate(turtle.heading); // draw the id unrotated
				g.setPaint(Color.white);
				g.fill(turtle.identify);
				g.setPaint(Color.black);
				g.setStroke(new BasicStroke((float)(1 / scale)));
				g.draw(turtle.identify);
			}
		}

		// restore the original transformation
		g.setTransform(oldTransform);
	}
}
